/*
 ============================================================================
 Name        : global.h
 Author      : Oscar Contreras
 Version     :
 Copyright   : Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
 Description : First order logic theorem prover in C
 ============================================================================
 */
/****************************************************************
*
*                 global.h (global include file for Drodi project)
*
* All of the documentation and software included in the Drodi package
* is copyrighted by Oscar Contreras.
*
* Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
*
*
* This software is provided "as-is," without any express or implied warranty.
* In no event shall the author be held liable for any damages arising from the
* use, or the result of use, of this software and its documentation in terms
* of correctness, accuracy, reliability, currentness, or otherwise.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial use, and to alter and redistribute it, provided
* that the following conditions are met:
*
* 1. All redistributions of source code files must retain all copyright
*    notices that are currently in place, and this list of conditions without
*    modification.
*
* 2. All redistributions in binary form must retain all occurrences of the
*    above copyright notice that are currently in place (for example, in
*    the About boxes).
*
* 3. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software.
*
* 4. Modified versions in source or binary form must be plainly marked as
*    such, and must not be misrepresented as being the original software.
*
*
* Oscar Contreras
* rwtrainer10 AT gmail.com
*
****************************************************************/

#ifndef GLOBAL_H_
#define GLOBAL_H_
/* Include for platform independent modules */

/*--- Standard Includes ----------------------------------------*/
#include <omp.h>                        /* Include OpenMP support       */
#include <stddef.h>                     /* Standard definitions         */
#include <string.h>                     /* strcpy/memset                */
#include <ctype.h>                      /* isdigit/toupper...           */
#include <stdio.h>                      /* sprintf/fopen...             */
#include <stdlib.h>                     /* malloc/free...               */
#include <time.h>                       /* clock()                      */
#include <sys/time.h>                   /* gettimeofday()               */
#include <limits.h>                     /* CHAR_BIT                     */
#include <stdint.h>                     /* intxx_t data types           */
#include <locale.h>                     /* setlocale()                  */
#include <math.h>                       /* sqrt()                       */
#include <signal.h>                     /* signal()                     */
#include <errno.h>                      /* strerror()                   */
#include <unistd.h>                     /* getpid(), sysconf()...       */
#include <sys/types.h>                  /* getpid()                     */
#include <sys/resource.h>               /* getrlimit(), setrlimit()     */

/* Define's for compilation time code generation control. */
//#define DEBUGMEMORY                         /* Generate debug code for memory problems detected by Valgrind.*/
                                            /* See MYFREE() comments for information. */
//#define DEBUGCODE                           /* Generate additional debug code. */
//#define VERBOSE                             /* Display inferences if DEBUGCODE is defined and only one core enabled. */
//#define SATDEBUGCODE1                       /* Generate additional debug code specific to check coherence */
                                            /* of hashchcmp elements ordered sub-queue. */
//#define SATDEBUGCODE2                       /* Generate additional debug code specific to check coherence */
                                            /* of satclause elements ordered sub-queue. */
//#define VERIFYSATPROOF                      /* Verify that the SAT part of the proof is coherent. */
#define DEBUGKBDATA 0                       /* Set to 0 for normal operations. Set to 1 to force print KB data even in */
                                            /* TIMEOUT or NOMEMORY conditions. */
#define ASSUMECOMPLETE                      /* If this is defined then the algorithm is assumed to be complete. */
                                            /* Right now it is NOT complete, it reports satisfiable for a few theorem */
                                            /* and unsatisfiable problems. */
//#define MURMUR3                             /* Use Murmur3 hash function (default is to use KRM hash function). */
//#define BENCHMARKING                        /* Include code for benchmarking summary output. */
                                            /* VERY IMPORTANT: If BENCHMARKING is defined then the BENCHDIMENSION parameter */
                                            /* below must be adjusted appropriately. */
//#define LOGEXCEPTIONS                       /* If this is defined then exceptions will be logged in drodi.log file or stdout. */
#ifdef BENCHMARKING
#define BENCHTYPE 3                         /* 1 -> report solving times for problems solved. */
                                            /* 2 -> report THM, SAT, CSA, CAX or UNS for problems solved. */
                                            /* 3 -> report 1 and 2 above for problems solved. */
#define VERIFYSATFINALSTATUS 0              /* If this equals one then SAT final status is verified after a problem solving attempt. */
                                            /* Only if BENCHMARKING is defined this can be set to one, otherwise it must be set */
                                            /* to zero. See VerifySatStatus() function for details. */
#define BENCHDIMENSION 381                  /* Adjust this to the number of problems per run. */
#else
#define VERIFYSATFINALSTATUS 0              /* See comments above. */
#endif
//#define MEMORYTUNNING                       /* If this is defined some code is generated for memory limit fine tuning and the */
                                            /* SymMalloc() function is defined and called from Initialize_pgm() and Clean4Exit() */
                                            /* functions to simulate huge memory allocation. */
#ifdef MEMORYTUNNING
#define ADDTLMEMORY 4000000000              /* Additional memory to be allocated if MEMORYTUNNING is defined. */
#endif

/* Other algorithm parameter definitions. These are candidates to be adjusted via command options. Some of them */
/* can be commented out when indicated so, altering this way the code generation. */
#define MAXDIST2CONJ 3                  /* Maximum distance to conjecture. See global comments in ClauseWieght() function for details. */
#define VARWEIGHT 1.0                   /* Relative weight of variable symbols for clause weight evaluation. */
#define PREDWEIGHT 10.0                 /* Relative weight of predicate symbols for clause weight evaluation. This weight is taken */
                                        /* as a reference and should be always equal to 10.0. */
#define EQUWEIGHT 2.0                   /* Relative weight of equalities for clause weight evaluation. If this is commented out then */
                                        /* equalities are not accounted for in clause weight calculations. */
#define FUNCWEIGHT 5.0                  /* Relative weight of function symbols for clause weight evaluation. */
#define PNOTINCONJFACTOR1 1.5           /* Factor that multiplies weight of predicates not in the conjecture for first factor choice. */
#define PNOTINCONJFACTOR2 1.5           /* Same as above for second factor choice. */
#define ENOTINCONJFACTOR1 1.0           /* Factor that multiplies weight of equalities if there are no equalities in the conjecture */
                                        /* for first factor choice. */
#define ENOTINCONJFACTOR2 1.0           /* Same as above for second factor choice. */
#define FNOTINCONJFACTOR1 1.4           /* Factor that multiplies weight of functions not in the conjecture for first factor choice. */
#define FNOTINCONJFACTOR2 1.4           /* Same as above for second factor choice. */
#define POSLITERALFACTOR1 2.0           /* Factor that multiplies weight of items in positive literals of non input clauses for first */
                                        /* factor choice. */
#define POSLITERALFACTOR2 1.0           /* Same as above for second factor choice. */
#define POSLITERALFACTOR1I 1.0          /* Factor that multiplies weight of items in positive literals of input clauses for first */
                                        /* factor choice. It seems that it produces better results to assign lower values to this factor, */
                                        /* even to set it to 1.0. */
#define POSLITERALFACTOR2I 1.0          /* Same as above for second factor choice. */
#define SINGLEANY                       /* If this is defined then the SelectItems() function will select the first */
                                        /* negative literal if selection option is set to SINGLENEG or the first positive */
                                        /* literal if selection is set to SINGLEPOS. If this is NOT defined then the */
                                        /* SelectItems() function will select the MAXIMUM (if it exist) or first MAXIMAL */
                                        /* negative (if SINGLENEG) or positive (if SINGLEPOS) literal with respect to all */
                                        /* the negative literals. The results obtained is that it is better to define SINGLEANY. */
#define DEMODPOSITION 1                 /* If this equals 1 then demodulations are done first in FwdSimplify() function. */
                                        /* If this equals 2 then demodulations are done just after FirstSimplify() call. */
                                        /* If this equals 3 then demodulations are done just after SecondSimplify() call. */
                                        /* Value 1 is a bit better than the other values. */
#define KEEPSATMODEL 1                  /* If this equals 1 then each execution of SatSolver() function keeps the previous model */
                                        /* until a node is found without a forced option match. This is a slightly better option. */
                                        /* If this doesn't equal 1 then each execution of SatSolver() function keeps the previous */
                                        /* model until a node is found without a forced option match and there are still some  */
                                        /* forced options available. */
//#define PROGCLAUSESELECT                /* If this is defined then the clause selection in each round robin cycle is */
                                        /* done progressively through all the queues involved. If this is not defined */
                                        /* then the clause selection in each round robin cycle is done depleting each */
                                        /* queue one after the other. */
//#define PREFERPROCESSED                 /* If not commented out then clauses that are added to PASSIVE more than once are placed at */
                                        /* the top of the age queue. */

/* Definitions for parse functions return types. */
#define NEGATION 1                          /* Negation. */
#define UNIVERSALQ 2                        /* Valid universal quantifier. */
#define EXISTENTIALQ 3                      /* Valid existential quantifier. */
#define BRACKET 4                           /* Open parenthesis "(". */
#define VALIDATOMIC 5                       /* Valid atomic sentence. */
#define UNDECLAREDTERM 6                    /* Predicate or function not previously declared. */
#define DECLAREDFUNC 7                      /* Term has been previously declared as a function. */
#define DECLAREDPRED 8                      /* Term has been previously declared as a predicate. */
#define STDVARNAME 9                        /* Standardized variable reserved name. */
#define DECLAREDVAR 10                      /* Declared variable. */
#define VALIDVARLIST 11                     /* Valid variable list. */
#define VALIDSENTENCE 12                    /* Valid sentence. */
#define VALIDNAME 13                        /* Valid name. */

/* Define's for syntax error codes. */
#define NOMEMORY 100000100                  /* No memory. */
#define EMPTYSENTENCE 100000101             /* Empty sentence. */
#define EMPTYBLOCK 100000102                /* Empty parenthesis block. */
#define INVTPTPVARNAME 100000103            /* Invalid TPTP variable name. */
#define INVTPTPNAME 100000104               /* Invalid TPTP predicate or function name. */
#define INVALIDATOMIC 100000105             /* Invalid atomic sentence. */
#define INVALIDVARLIST 100000106            /* Invalid variable list. */
#define INVALIDNAME 100000107               /* Invalid or too long variable, constant, predicate or function name. */
#define INVALIDTERM 100000109               /* Invalid term. */
#define MISSINGARGS 100000110               /* Missing arguments. */
#define ALREADYDECLARED 100000111           /* Variable name already declared in equal scope quantifier. */
#define MISSINGVARINLIST 100000112          /* Missing variable in a variable list. */
#define INVALIDDECLAREDARGS 100000113       /* Predicate or function doesn't match previous declared number of arguments. */
#define INVALIDDECLAREDTYPE 100000114       /* Predicate or function doesn't match previous declared type. */
#define TOOMANYARGS 100000115               /* Too many arguments. */
#define RESERVEDNAME 100000116              /* Use of standard variable name for a function or predicate. */
#define SYNTAXERROR 100000117               /* Generic syntax error. */

/* Definitions for flags field of kbase structure. */
#define RETENTIONUSED 1                     /* LRS or weight limit retention processes have been used and some clauses have */
                                            /* been discarded. The KB now is not complete. */

/* Definitions for flags field of cmprefix structure and type argument of simplify functions. */
/* These flags must be compatible with SELECTED and HORN flags. The HORN flag is set in clauses */
/* if appropriate but it is not used by the moment. The flags FROMCONJECTURE and FROMMAINFILE */
/* must be compatible with flags in type field of hashchain structure. */
#define TEXTFORMULA 1                       /* Formula is of text type. */
                                            /* These formulas have their own linked queue. */
#define UNPROC 2                            /* Formula is an unprocessed binary clause. */
                                            /* These formulas have their own linked queue. */
#define PASSIVE 4                           /* Formula is a passive binary clause. */
                                            /* These formulas have their own linked queue. */
#define ACTIVE 8                            /* Formula is an active binary clause. */
                                            /* These formulas have their own linked queue. */
#define LOCKED 16                           /* Formula is a locked binary clause. */
                                            /* These formulas have their own linked queue. */
#define NOFACTORS 32                        /* Clause have no factors. */
#define INDISPOSALQUEUE 128                 /* Clause is in disposal queue for later deletion or locking. */
#define FROMPASSIVE 256                     /* Clause is coming from PASSIVE, so only subsumption must be checked */
                                            /* in FirstSimplify() function. This is only used as a flag argument in */
                                            /* some simplifying functions. This is not used in the flags field of the */
                                            /* cmprefix structure. */
#define NUMBERED 512                        /* Clause has been assigned a number. */
#define INPROOFQUEUE 1024                   /* Clause or hashchcmp component is already in proof queue. This flag must be */
                                            /* compatible with hashchcmp structure flags. */
#define ASSERTSCHAINED 2048                 /* Assertions for this clause has been already chained. Only used for */
                                            /* binary clauses. */
#define FROMCONJECTURE 4096                 /* Formula was obtained from a negated conjecture in the pre-processing. */
                                            /* Used in hashchain type field flags for symbols appearing in these formulas. */
#define SELECTEDCLAUSE 8192                 /* Clause has been selected from PASSIVE to ACTIVE. Clauses with this flag */
                                            /* that are added to PASSIVE are placed at the top of the age queue. */
                                            /* With the AVATAR architecture a clause can go from PASSIVE to ACTIVE */
                                            /* to UNPROC to PASSIVE and this cycle can be repeated several times. */
#define KEPTCLAUSE 16384                    /* Clause was kept in the relevance filtering pre-processing. */
#define FROMMAINFILE 65536                  /* Clause is coming from the main file and can be considered as a conjecture */
                                            /* to the effects of detecting contradictory axioms. This flag must be */
                                            /* compatible with NEGCONJ definition flag below. */
											/* Used in hashchain type field flags for symbols appearing in these clauses. */
#define PICKED 131072                       /* Clause has been selected as ACTIVE. Due to AVATAR architecture an active clause */
                                            /* may later be an unprocessed clause so any type of clause may have this flag. */
#define WEIGHTED 262144                     /* Clause weights have been already assigned. */
#define GOODCHOICE 524288                   /* Clause has been classified as a good choice by the learning algorithms. */

/* Definitions for inference field of cmprefix structure. */
#define CONJECTURE 0                        /* Formula was an input conjecture to the problem. */
#define INPUT 1                             /* Formula was an input axiom to the problem. */
#define NEGCONJ 2                           /* Formula is a negated conjecture. This is also used as a flag and must be */
                                            /* compatible with FROMMAINFILE above. */
#define CLAUSIFY 3                          /* Formula comes from clausifying an input formula. */
#define EQUIVFACT 4                         /* Formula is an equivalent factor of another formula. */
#define RESOLUTION 5                        /* Formula was inferred by resolution. */
#define PARAMODULATION 6                    /* Formula was inferred by paramodulation. */
#define FACTORING 7                         /* Formula was inferred by factoring. */
#define FWDEMODULATION 8                    /* Formula was inferred by forward demodulation. */
#define BKDEMODULATION 9                    /* Formula was inferred by backward demodulation. */
#define REMOVEDUPS 10                       /* Formula is the result of removing duplicate literals from another formula. */
#define TRIVEQRES 11                        /* Formula was inferred by trivial equality resolution. */
#define EQRESOLUTION 12                     /* Formula was inferred by equality resolution. */
#define DESTEQRES 13                        /* Formula was inferred by destructive equality resolution. */
#define EQFACTORING 14                      /* Formula was inferred by equality factoring. */
#define FWSUBSRESLTNS 15                    /* Formula was inferred by forward subsumption resolution. */
#define BKSUBSRESLTNS 16                    /* Formula was inferred by backward subsumption resolution. */
#define SPLIT 17                            /* Formula was inferred by clause splitting. */
#define COMPONENT 18                        /* Formula is a clause component. */
#define CONTRADICTION 19                    /* Formula was inferred by first order solver contradiction. */
#define PRENNFCONV 20                       /* Formula is the result of a pre-nnf conversion. */
#define FORMULARENAMING 21                  /* Formula is the result of renaming a subformula. */
#define PREDICATEDEF 22                     /* Formula is a predicate definition for subformula renaming. */
#define NNFCONV 23                          /* Formula is the result of a full-nnf conversion. */
#define MINISCOPING 24                      /* Formula is the result of miniscoping. */
#define SKOLEMIZATION 25                    /* Formula is the result of skolemization. */
#define TFSIMPLIFICATION 26                 /* Formula is the result of a $true and $false simplification. */

/* Definitions for byte markers in compiled clauses and discrimination trees, type field */
/* of hashchain structures and type parameter of CreateGlobalSymbol() function. These flags must be compatible */
/* with FROMCONJECTURE and FROMMAINFILE also used in type field of hashchain structures for learning process. */
#define DUMMYNODE 0                     /* This is not a flag, it is a byte marker value for tree byte markers. */
#define UNITEND 1                       /* End of clause marker. Only used for byte markers. */
                                        /* This flag must be compatible with flags for Generalize() and */
                                        /* Unify() functions and with flags in hashchcmp and satnode structures */
                                        /* and assertions signbyte (see below). */
#define EQUALITY 2                      /* Item is an equality literal. */
#define PREDICATE 4                     /* Item is a predicate literal. */
#define NEGATED 8                       /* Item is a negated literal. Only used for byte markers. */
                                        /* This flag must be compatible with flags for Generalize() and */
                                        /* Unify() functions (see below). */
#define VARIABLE 16                     /* Item is a variable. Only used for byte markers. */
#define FUNCTION 32                     /* Item is a function (may be a constant, i.e. arity zero). */
#define SELECTED 64                     /* Item is a selected literal or equality root term. */
                                        /* This flag must be compatible with flags for QueryGeneric() function. */
#define MAXIMUM 128                     /* Item is a maximum equality root term. This must be different than ITEM1SEC, STANDARD */
                                        /* and NONRECURSIVE flags. */
#define PMSELECTED 128                  /* If this is set to the same value as MAXIMUM then a positive equality will be selected */
                                        /* for paramodulating from it only if it is in a clause without negative literals. */
                                        /* If this is set to the same value as SELECTED then the above condition will not */
                                        /* be required. It works a bit better setting it to the same value as MAXIMUM. */
#define DEFINEDPRED 256                 /* Used only in "type" parameter of CreateGlobalSymbol() function */
                                        /* and type field of hashchain structure. */
#define INCONJECTURE 512                /* Symbol is used in the conjecture. Used only in "type" parameter */
                                        /* of CreateGlobalSymbol() function. */
#define RELEVANT 1024                   /* Symbol is in a relevant clause. Used only in type field of hashchain */
                                        /* structure for clause relevance filtering. */
#define SKOLEM 2048                     /* Symbol is a skolem. Used only in type field of hashchain structure for */
                                        /* learning  process. */
#define DISTINCTOBJECT 8192             /* Symbol is a "distinct object". Used only in type field of hashchain structure */
                                        /* and type parameter of CreateGlobalSymbol() function. */

/* Definitions for "type" parameter in NextItem() function. */
#define OVERSUBTERMS 1                  /* Jump to next item that is not a subterm of the given item. */
#define IMMED 2                         /* Jumpt to item immediately following the given item. */

/* Definitions for Unify() and Generalize() flags. */
/* These flags must be compatible with the UNITEND and NEGATED flags (see above). */
#define LIST 2                          /* Parts to be unified are a list of terms. */
#define ITEM1SEC 16                     /* The first Unify() function argument or the variable in a substitution */
                                        /* belongs to a secondary clause. They belong to the main clause by default. */
                                        /* Also used for flag value of CheckEqRootOrder() function so it must be */
                                        /* different than MAXIMUM, STANDARD and NONRECURSIVE values defined elsewhere. */
#define ITEM2SEC ITEM1SEC<<1            /* The second Unify() function argument or the substituting term in a */
                                        /* substitution  belongs to a secondary clause. They belong to the main */
                                        /* clause by default. This must be ITEM1SEC shifted by 1 bit. */
#define INITSUBST 64                    /* Initialize substitution to null. */
#define DEBUG 128                       /* Call BuildTxtSubstData() function to show results in text form. */

/* Definitions for status fields of kbase and mpctl structures. They must be compatible with NOMEMORY above. */
#define STOPPED 0                       /* Process is stopped. */
#define RUNNING 1                       /* Process is running a (possibly multi-threading) first order solver algorithm */
                                        /* with or without a SAT solver. */
#define RUNNINGSAT 2                    /* Process is running a single threaded SAT solver algorithm. */
#define SATISFIABLE 3                   /* Formulas in KB are satisfiable, if there are conjectures */
                                        /* they are not refutable. */
#define UNSATISFIABLE 4                 /* Formulas in KB are unsatisfiable, if there are conjectures */
                                        /* they are refutable. */
#define UNKNOWN 5                       /* It is unknown if formulas in KB are satisfiable or unsatisfiable. */
#define TIMEOUT 6                       /* Saturation algorithm timed out. This cannot be 0 or 1. */

/* Definitions for ordtype values and CompareTerms() flag. */
#define STANDARD 1                      /* Use standard KBO ordering. */
#define NONRECURSIVE 2                  /* Use non recursive KBO ordering. This flag must be compatible  with */
                                        /* MAIMUM and ITEM1SEC flags defined elsewhere. */
#define LEXICOGRAPHIC 4                 /* Use >> symbol precedence relation for atom ordering. */

/* Definitions for weight functions. */
#define UNIFORM 0                       /* Use a uniform weight function assigned a weight=1 to all functions. */
#define ARITY 1                         /* Use weight function = 1 + 10·arity(function). */

/* Definitions for ordering functions return values and for the result field of the */
/* equpnode structure. See the functions description for the meaning of each one, */
/* as it may differ from one function to another. */
#define EQUAL 1                         /* item1 = item2. */
#define GREATER 2                       /* item1 > item2. */
#define LESSER 4                        /* item2 > item1. */
#define NOGREATER 8                     /* item2 ·> item1 but neither item2 = item1 nor item2 > item1. */
#define NOLESSER 16                     /* item1 ·> item2 but neither item1 = item2 nor item1 > item2. */
#define FAILURE 32                      /* None of the above. */
#define EQUALW 64                       /* weight(item1) = weight(item2). */
#define FAILUREW 128                    /* weight(item1) is not comparable with weight(item2). */

/* Definitions for algorithm field of kbase structure. These cannot be 0 nor 1 for compatibility */
/* with other values of split field in options structure. */
#define OTTER 2                         /* Otter algorithm. */
#define DISCOUNT 3                      /* Discount algorithm. */

/* Definitions for EditDiscrTrees() function flag parameter. */
#define ADDTREES 1                      /* Add tree branches for specified clause. */
#define DELETETREES 2                   /* Delete tree branches for specified clause. */

/* Definitions for type argument of QueriUI() function. */
#define UNIFICATION 0                   /* Search is for an item that unifies with the query term. */
#define INSTANCE 1                      /* Search is for an item that is an instance of the query term. */

/* Definitions for flags of AddBinClause2KB() function. */
#define SELECT 1                        /* Appropriate clause items will be selected. */
#define SETFNUMBER 2                    /* Set formula number according to the nformulas field of KB. */

/* Definitions for flags of JumpOverBlockOrTerm() function. */
#define TOEND 0                         /* Jump to the last item in the block or term. */
#define PASTEND 1                       /* Jump to the item following the last item in the block or term. */

/* Definitions for mxwghtopt field of options structure. */
#define MXWEIGHTOFF 1                   /* Maximum weight option deactivated for algorithms without LRS enabled. */
#define MXWEIGHTLRS 2                   /* Maximum weight calculated using Limited Resource Strategy enabled */
                                        /* for all algorithms. */
#define MXWEIGHTLRSOTT 4                /* Maximum weight calculated using Limited Resource Strategy enabled for Otter */
                                        /* algorithm, disabled for Discount algorithm. */
#define MXWEIGHTLRSDSC 8                /* Maximum weight calculated using Limited Resource Strategy enabled for Discount */
                                        /* algorithm, disabled for Otter algorithm. */
#define MXWEIGHTLRSOFF 16               /* Maximum weight calculated using Limited Resource Strategy id disabled */
                                        /* for all algorithms. */
#define MXWEIGHTNUM 32                  /* Maximum weight selected by user for algorithms without LRS enabled. */

/* Definitions for SYNTAX and PROOF options. */
#define TPTP 1                          /* TPTP mode syntax. */
#define DRODI 2                         /* Drodi mode syntax. */
#define PROOFOFF 3                      /* Don't print proof. */

/* Definitions for program modes (pgmmode global variable). */
#define NONINTERACTIVE 0                /* non interactive mode, entered if some arguments are given to the program. */
#define INTERACTIVE 1                   /* Interactive mode, entered if no arguments are given to the program. */

/* Definitions for prtkbdata option. */
#define PRTWINNER 1                     /* Print winner KB data. */
#define PRTALL 2                        /* Print all KB's data. */
#define PRTOFF 3                        /* Don't print KB data. */

/* Definitions for select field of options structure. The definitions are set as if they were flags */
/* to speed up some tests, but only one of the values (flags) is allowed. */
#define MAXIMAL 1                       /* All maximal literals are selected. */
#define SINGLENEG 2                     /* If there are negative literals then the first one is selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */
#define MULTINEG 4                      /* If there are negative literals then the negative literals */
                                        /* that are maximal in the set of negative literals are selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */
#define SINGLEPOS 8                     /* If there are positive literals then the first one is selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */
#define MULTIPOS 16                     /* If there are positive literals then the positive literals */
                                        /* that are maximal in the set of positive literals are selected */
                                        /* and the other literals are not selected. Otherwise all */
                                        /* maximal literals are selected. */

/* Definitions for flags field in hashchcmp and satnode structures and assertions signbyte */
/* (see AddAssertions() function comments). These definitions must be compatible with UNITEND */
/* that is also used in assertions signbyte and with INPROOFQUEUE cmprefix structure flag */
/* that is also used in hashchcmp flags. */
#define POSITIVE 2                      /* The component is true in SAT interpretation model or in assertion_data. */
#define NEGATIVE 4                      /* The component is false in SAT interpretation model or in assertion_data. */
#define UNDEFINED 8                     /* The component is undefined in SAT interpretation model. This is only used */
                                        /* in flags field of hashchcmp structure. */
#define GROUND 16                       /* The component is ground. This is only used in flags field of hashchcmp structure. */
                                        /* This flag must be compatible with flags for pbmflags global variable and serves */
                                        /* to identify ground problems. */
#define PRPOSITIVE 32                   /* The component was true in previous model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define PRNEGATIVE 64                   /* The component was false in previous model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define PRUNDEFINED 128                 /* The component was undefined in previous model of SAT interpretation. This is only used */
                                        /* in flags field of hashchcmp structure. */
#define LASTOPTION 256                  /* The assignment of node is the last possibility available, no other possibility */
                                        /* (true or false) is left or valid to be explored.*/
#define ORDERED 512                     /* The hashchcmp or satclause structure element has been linked to the ordered */
                                        /* ordered sub-queue. */
#define FREENODE 2048                   /* The node is allocated but it is not in use and its data is no longer necessary. */
                                        /* The node can be reused. This way unnecessary memory allocation and freeing is */
                                        /* avoided. This is only used in flags field of satnode structure. */
#define MODELPOSITIVE 4096              /* The component is true in current model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define MODELNEGATIVE 8192              /* The component is false in current model of SAT interpretation or in assertion_data. */
                                        /* This is only used in flags field of hashchcmp structure. */
#define MODELUNDEFINED 16384            /* The component is undefined in current model of SAT interpretation. This is only used */
                                        /* in flags field of hashchcmp structure. */
#define POSCOMPADDED 32768              /* The component clause C <- [C] has been added to first order solver. This is */
                                        /* only used in flags field of hashchcmp structure. */
#define NEGCOMPADDED 65536              /* The component clause ~C <- ~[C] has been added to first order solver. This is */
                                        /* only used in flags field of hashchcmp structure. */

/* Definitions for flags in pbmflags global variable. The GROUND flag defined above is also used for pbmflags. */
#define EXISTCONJ 1                     /* A conjecture exist for the current problem. */
#define EXISTNEGCONJ 2                  /* A negated conjecture exist for the current problem. */
#define WITHHORNCLAUSES 4               /* Problem is not horn but there are horn clauses. */
#define UNITEQUALITY 8                  /* Problem has some positive unit equalities. */
#define NONUNITEQUALITY 32              /* Problem has some equalities none of which are positive unit equalities. */
#define UNITCLAUSE 64                   /* All clauses in problem are unit clauses. */
#define PROOFWITHNEGCONJ 128            /* The proof includes some formulas with NEGCONJ inference. */
#define SATREFUTATION 256               /* There was a Sat refutation. */
#define HORN 32768                      /* Problem is horn. This flag must be compatible with flags field */
                                        /* of cmprefix structure. */
#define CNFFORMAT 65536                 /* Problem was read in CNF TPTP format. */
#define EXISTAXIOMFILE 131072           /* Problem included an axiom file. */

/* Definitions for type field in proofnode structure. */
#define SYMBOLDEFNTN 1                  /* Symbol definition. */
#define SATCLAUSE 2                     /* Sat clause. */
#define ACLAUSE 3                       /* A-Clause. */

/* Definitions for flags field in lkedfitem structure. */
#define OPENBRACKET 1                   /* Opening bracket type item. */
#define CLOSEBRACKET 2                  /* Closing bracket type item. */
#define EQUPRED 4                       /* Equality predicate type item. */
#define NAMEDPRED 8                     /* Normal predicate type item. */
#define NAMEDFUNC 16                    /* Function type item. */
#define NAMEDVAR 32                     /* Variable type item. */
#define UQUANTIFIER 64                  /* Universal quantifier type item. */
#define EQUANTIFIER 128                 /* Existential quantifier type item. */
#define ANDOPER 256                     /* AND connective type item. */
#define OROPER 512                      /* OR connective type item. */
#define IMPOPER 1024                    /* Implication connective type item. */
#define REVIMPOPER 2048                 /* Reverse implication connective type item. */
#define EQUIVOPER 4096                  /* Equivalence connective type item. */
#define XOROPER 8192                    /* XOR connective type item. */
#define NOROPER 16384                   /* NOR connective type item. */
#define NANDOPER 32768                  /* NAND connective type item. */
#define NEGITEM 65536                   /* Negated item. */
#define POSPOL 131072                   /* Positive polarity item. */
#define NEGPOL 262144                   /* Negative polarity item. */
#define NEUTRALPOL 524288               /* Neutral polarity item. */
#define RENAMED 1048576                 /* Variable renamed in a quantifier or variable item. */
#define FREEVAR 2097152                 /* The quantifier variable is a free variable in the renamed block. */
#define TRUEPRED 4194304                /* The predicate is a $true predefined predicate. */
#define FALSEPRED 8388608               /* The predicate is a $false predefined predicate. */

/* Definitions for clause relevance filtering program option (prproflags global variable). */
#define CLFILTERINGRESIGN 1             /* Clause relevance filtering is not performed performed in case of timeout condition. */
#define CLFILTERINGOFF 2                /* No clause relevance filtering is performed. */
#define CLFILTERINGON 4                 /* Clause relevance filtering is always performed. */
#define CLFILTERINGNONEQU 8             /* Clause relevance filtering is performed in problems without equalities only. */
                                        /* Only one of the CLFILTERINGOFF, CLFILTERINGON or CLFILTERINGNONEQU flags can be set. */
                                        /* However CLFILTERINGRESIGN flag is compatible with the other three flags. */
#define CLFILTERINGITERLIMIT 16         /* The threshold in ppthrshld global variable is the maximum number of iterations */
                                        /* allowed in the relevance filtering algorithm. Otherwise the threshold in ppthrshld */
                                        /* global variable is the maximum time in seconds allowed for the algorithm. */
#define CLFILTERINGLIMITREACHED 32      /* A timeout condition occurred or maximum number of iterations reached during */
                                        /* relevance filtering pre-processing before */
                                        /* the filtering process was completed. */
#define CLFILTERINGPRUNNED 64           /* The relevance filtering deleted one or more clauses. */

/* Definitions of masks for each type of adjustable parameter. */
#define ALGMASK 1                       /* ALGORITHM parameter mask. */
#define SELECTMASK 0b1000000000110      /* SELECT parameter mask. */
#define WFACTSETMASK 0b1000             /* Weight factor index parameter mask. */
#define EQFACTMASK 0b10000              /* EQUIVFACT parameter mask. */
#define FACTORINGMASK 0b100000          /* FACTORING parameter mask. */
#define TERMORDRMASK 0b1000000          /* TERMORDER parameter mask. */
#define LITORDRMASK 0b110000000         /* LITORDER parameter mask. */
#define FUNCWMASK 0b1000000000          /* FUNCTIONW parameter mask. */
#define MAXWMASK 0b110000000000         /* MAXWEIGHT parameter mask. */

/* Definitions of values for each type of adjustable parameter. */
#define OTTER_P 0                       /* Otter algorithm. */
#define DISCOUNT_P 1                    /* Discount algorithm. */
#define SINGLENEG_P 0                   /* SINGLENEG SELECT option. */
#define MULTINEG_P 0b10                 /* MULTINEG SELECT option. */
#define MAXIMAL_P 0b100                 /* MAXIMAL SELECT option. */
#define SINGLEPOS_P 0b110               /* SINGLEPOS SELECT option. */
#define MULTIPOS_P 0b1000000000000      /* MULTIPOS SELECT option. */
#define WFACT1_P 0                      /* Weight factor index 1. */
#define WFACT2_P 0b1000                 /* Weight factor index 2. */
#define EQFACTON_P 0                    /* OFF EQUIVFACT option. */
#define EQFACTOFF_P 0b10000             /* ON EQUIVFACT option. */
#define FACTON_P 0                      /* OFF FACTORING option. */
#define FACTOFF_P 0b100000              /* ON FACTORING option. */
#define TRMORDSTD_P 0                   /* STANDARD TERMORDER option. */
#define TRMORDNR_P 0b1000000            /* NONRECURSIVE TERMORDER option. */
#define LITORDSTD_P 0                   /* STANDARD LITORDER option. */
#define LITORDNR_P 0b10000000           /* NONRECURSIVE LITORDER option. */
#define LITORDLEX_P 0b100000000         /* LEXICOGRAPHIC LITORDER option. */
#define UNIFORM_P 0                     /* UNIFORM FUNCTIONW option. */
#define ARITY_P 0b1000000000            /* ARITY FUNCTIONW option. */
#define LRSOTT_P 0                      /* LRSOTTER MAXWEIGHT option. */
#define LRSOFF_P 0b10000000000          /* LRSOFF MAXWEIGHT option. */
#define LRSALL_P 0b100000000000         /* LRS MAXWEIGHT option. */
#define LRSDSC_P 0b110000000000         /* LRSDISCOUNT MAXWEIGHT option. */
#define LEARN_P 0b10000000000000        /* Use learning option. This doen't need an adjustable parameter mask. */

/* Definitions to be used if VERIFYSATSOLVER is not zero. They are always defined anyway. */
#define SATOK 0                         /* SAT solver status is OK. */
#define SATSAT 1                        /* SAT solver has reported a refutation but all SAT clauses are satisfied. */
#define SATUNSAT 2                      /* SAT solver has reported a model but not all SAT clauses are satisfied. */
#define SATUNPR1 4                      /* There are unprocessed clauses that must be locked. */
#define SATUNPR2 8                      /* There are clauses in unprocessed queue not flagged as unprocessed. */
#define SATUNPR12 16                    /* Conditions SATUNPR1 and SATUNPR2. */
#define SATPASSIVE1 32                  /* There are passive clauses that must be locked. */
#define SATPASSIVE2 64                  /* There are clauses in passive queue not flagged as passive. */
#define SATPASSIVE12 128                /* Conditions SATPASSIVE1 and SATPASSIVE2. */
#define SATACTIVE1 256                  /* There are active clauses that must be locked. */
#define SATACTIVE2 512                  /* There are clauses in active queue not flagged as active. */
#define SATACTIVE12 1024                /* Conditions SATACTIVE1 and SATACTIVE2. */
#define SATLOCKED1 2048                 /* There are active clauses that must be locked. */
#define SATLOCKED2 4096                 /* There are clauses in active queue not flagged as active. */
#define SATLOCKED12 8192                /* Conditions SATLOCKED1 and SATLOCKED2. */
#define SATUNK 16384                    /* SAT solver status is unknown because the problem was not reported as solved. */

/* Definitions for the automatic learning process. */
#define VECTORLENGTH 1558               /* Number of bytes in a clause vector used for learning. */
#define CONJECTURECOMPONENT 10.0        /* Value of feature if a conjecture exist. */
#define AXIOMFILECOMPONENT 10.0         /* Value of feature if an axiom file exist. */
#define PROBLEMTYPECOMPONENT 10.0       /* Value of problem type features. */
#define PENALTYPARAMETER 0.1            /* Penalty parameter for learning process. */
#define LEARNPRECISION 0.005            /* Learning result vector relative error threshold. */
#define TOLERANCE 26.0                  /* Duality gap tolerance before starting to check learning result vector relative error */
                                        /* in dual DCDescentMethod() function. */
#define BADCLAUSEWEIGHT 100             /* Weight penalty if the clause is not classified as good by the learning algorithm. */
#define WEIGHT0FACTOR 1.0               /* Factor that multiplies clause clweight[1] to get clause clweight[0]. */
#define SHRINKING                       /* If defined then a shrinking DCDescentMethod() function will be used. */
                                        /* Otherwise a DCDescentMethod() function without shrinking will be used. */

/* Miscellanea definitions. */
#define VERSION "Drodi V3.1.5"          /* Program version. */
#define SPREAD 0                        /* Use spread core affinity bind option. */
#define CLOSE 1                         /* Use close core affinity bind option. */
#define LRSTHRESHOLD 200                /* Number of passive clauses moved to active per LRS cycle. */
#define LRSCOEF 1.1                     /* Coefficient to apply to number of LRS reachable clauses. */
#define TIMEOUTCCLDSP 10                /* Number of cycles for timeout check in clause disposal. */
#define INPUT_CHUNK_SIZE 1024           /* Chunk size in bytes for command input. */
#define VAR_CHUNK_SIZE 100              /* Chunk size in bytes for text formula names. */
#define STR_CHUNK_SIZE 200              /* Chunk size in bytes for text buffers. */
                                        /* This must be bigger or equal to 10. */
#define TREEPTR_CHUNK_SIZE 4            /* Pointer set chunk size in tree pointer elements. */
#define STACK_CHUNK_SIZE 40             /* Stack elements chunk size in number of elements. */
#define TREE_CHUNK_SIZE 100             /* Tree chunk size in bytes. */
#define SUBST_CHUNK_SIZE 1000           /* Chunk size in bytes for substitution buffer. */
#define CLAUSE_CHUNK_SIZE 200           /* Compiled auxiliary clauses chunk size in bytes. */
#define PNDARGS_CHUNK_SIZE 10           /* Pending arguments elements chunk size in number of elements. */
#define VARRENUM_CHUNK_SIZE 50          /* Variable renumbering buffer size in number of elements. */
#define ARGSMASK 0x1fffffff             /* Mask for number of arguments. */
#define MAXRECURSIVE 60000              /* Maximum recursive calls in QueryGeneric() and QueryUI() functions. */
#define ABSMAXWEIGHT 0xffffffff         /* Absolute maximum clause weight. This is to prevent exhausting the stack space by */
                                        /* too many calls to QueryGeneric() for too long clauses. It is rare but it may happen. */
#define MAXWEIGHT 1000                  /* Maximum number of weight queue pointers hold in each weight queue. */
#define WEIGHTQUEUES 5                  /* Number of weight queues used for clause selection. */
#define SYMBOLDIM1 3                    /* Number of different symbol weights used to compute the different clause weights. */
                                        /* This is the dimension of the weight set field of hashchain structure. */
#define SYMBOLDIM2 2                    /* Number of different total symbol weights in a clause. This is the dimension of the */
                                        /* symbolw set field of binprefix structure. */
#define MAXPOSITIONS 400                /* Maximum number of hash values for item positions. */
#define MAXSTRATEGIES 2880              /* Number of predefined strategies, this depends on the number of parameters that */
                                        /* are automatically adjusted depending on the problem type. See InitStrategies() */
                                        /* function for details. */
#define BKSTRATEGIESPERTHREAD 4         /* Number of backup strategies per thread to cover early returns due to incompleteness. */
#define HASHVALUES 0x10000              /* Number of different hash values for symbol names and clause variants. */
#define MAXMINISCOPETIME 1.0            /* Miniscope time limit in seconds. */
#ifdef WINTYPE
#define WPATHSEPARATOR L'\\'            /* Windows type wide char file path separator. */
#else
#define WPATHSEPARATOR L'/'             /* Linux & AIX type wide char file path separator. */
#endif

/* Useful macros */
#ifdef DEBUGMEMORY
#define MYFREE(X) myfree(X,__FILE__,__LINE__)
#define DBGMEMELEMS 2000
#else
#define MYFREE(X) myfree(X)
#define DBGMEMELEMS 0
#endif
#define ABS(X) ((X)>=0?X:-(X))
#define MAX(X,Y) ((X)>=(Y)?X:Y)
#define MIN(X,Y) ((X)<=(Y)?X:Y)
#define NUMELEMS(X) sizeof(X)/sizeof(X[0])
#define MATCHFLGS11(X,Y) (((X)&ITEM1SEC)==((Y)&ITEM1SEC))
#define MATCHFLGS12(X,Y) (((X)&ITEM1SEC)==(((Y)&ITEM2SEC)>>1))
#define SPACE_AFTER_HASHCHAIN (6*(num_wkbs+1))
#define SHORTSYMBOLSIZE sizeof(symbol)-(2*sizeof(void *))
#ifdef MURMUR3
#define ROT32(x, y) ((x << y) | (x >> (32 - y)))
#endif

/* STRUCTURES */
/* Structure for chaining names buffers for told sentences. */
/* See AddNametoChain() function for full format description of names buffer. */
typedef struct snameschain {
	struct snameschain *nxtnameschain;      /* Pointer to next nameschain structure, NULL if last. */
	struct slkedfitem *quantifier;          /* Pointer to quantifier linked list item corresponding to a variable item. */
	char *names;                            /* Pointer to names list buffer. */
	int32_t buffsize;                       /* Size of buffer. */
} nameschain;

/* Structure for formula common prefix. */
typedef struct scmprefix {
	struct scmprefix *next;             /* Pointer to next formula for formulas of the same kind (TEXT, ACTIVE, PASSIVE, */
	                                    /* UNPROC) in the KB. NULL if last formula of its kind. */
	struct scmprefix *prev;             /* Pointer to previous formula for formulas of the same kind (TEXT, ACTIVE, PASSIVE, */
	                                    /* UNPROC) in the KB. NULL if first formula of its kind. */
	                                    /* This is intended to free formula memory. */
	struct scmprefix *parent1;          /* First parent formula pointer. */
	struct scmprefix *parent2;          /* Second parent formula pointer, NULL if there is no second parent. */
	                                    /* For predicate definition formulas this pointer is used to chain */
	                                    /* the set of predicate definition formulas used in a single formula */
	                                    /* renaming inference. */
	struct scmprefix *prevdisp;         /* Pointer to previous formula in disposal queue. NULL if last formula */
	                                    /* or formula not in disposal queue. This queue interconnects clauses */
	                                    /* from other queues.*/
	union {
		char *text;                     /* Pointer to formula in plain text form. Used for text type formulas. */
		struct sbinprefix *bin;         /* Pointer to binary clause prefix structure. Used for compiled clauses. */
	} part2;
	uint32_t number;                    /* Number assigned to formula. */
	int32_t inference;                  /* Inference from where the formula comes from (see definitions above). */
	uint32_t flags;                     /* See flags above. This must be the last field of this structure. */
} cmprefix;

/* Structure for binary clause prefix. */
typedef struct sbinprefix {
	cmprefix *clause;                   /* Pointer to cmprefix structure of the formula itself. */
	cmprefix *nextweight[WEIGHTQUEUES]; /* Set of pointers to next formula in each weight queue. NULL if last formula. */
	cmprefix *prevweight[WEIGHTQUEUES]; /* Set of pointers to previous formula in each weight queue. NULL if last formula. */
	uint8_t *asserts;                   /* Pointer to buffer with clause assertions. */
	uint8_t *lockasserts;               /* Pointer to buffer with clause locking assertions. */
	uint64_t signature;                 /* Signature of last inference process. */
	uint64_t locksignature;             /* Signature for lock and unlock processes. */
	uint32_t literals;                  /* Number of literals in clause. */
	float symbolw[SYMBOLDIM2];          /* Set of symbol weights used to compute clause weight for the different weight queues. */
	                                    /* Index 0 is the pure symbol weight. */
	                                    /* Index 1 is the symbol weight with a factor applied if the symbol doesn't appear */
	                                    /* in the conjecture. */
	uint32_t dist2conj;                 /* Distance to conjecture. See global comments in ClauseWieght() function for details. */
	uint32_t clweight[WEIGHTQUEUES];    /* Clause weight for each weight queue. */
	int32_t depth;                      /* Max nesting depth of predicate/terms. */
	int32_t size;                       /* Size in bytes of formula including ending UNITEND byte. */
	int32_t asize;                      /* Size in bytes of assertions including ending UNITEND byte. */
	int32_t lasize;                     /* Size in bytes of locking assertions assertions including ending UNITEND byte. */
	int32_t maxvarnb;                   /* Maximum variable number, -1 if no variables. This is useful for */
	                                    /* memory allocations in ordering functions. */
	int32_t equflg;                     /* 0 if equality ordering data has not been set in equalities. */
	                                    /* 1 if equality ordering data has been set in SELECTED equalities */
	                                    /* by calls from SelectItems() function. */
	                                    /* 2 if equality ordering data has been set in the only equality */
	                                    /* of a clause used for backward demodulation. */
	                                    /* See comments in FreeEquData() function for indications on how */
	                                    /* equality ordering data is allocated and freed with the help of this flag. */
	uint8_t formula[1];                 /* Dummy for start of formula items. This must be the last field of this structure. */
} binprefix;

/* Structure for SAT clauses. */
typedef struct ssatclause {
	struct ssatclause *glblnext;        /* Pointer to next SAT clause of the same global queue, NULL if last. */
	struct ssatclause *glblprev;        /* Pointer to previous SAT clause of the same global queue, NULL if last. */
	struct ssatclause *next;            /* Pointer to next SAT clause of the SAT solver ordered clauses queue, NULL if last. */
	struct ssatclause *prev;            /* Pointer to previous SAT clause of the SAT solver ordered clauses queue, NULL if last. */
	struct scmprefix *parent;           /* Pointer to parent A-clause. */
	int32_t number;                     /* Number assigned to the SAT clause. */
	int32_t inference;                  /* Inference from which the SAT clause was obtained. It can be SPLIT or CONTRADICTION. */
	                                    /* See definitions above. */
	int32_t literals;                   /* Number of literals in clause. */
	int32_t satlits;                    /* Number of satisfied literals in clause. */
	int32_t undeflits;                  /* Number of undefined literals in clause. */
	int32_t inproof;                    /* Set to 1 if clause is in the unsatisfiable core but is not the contradiction clause */
	                                    /* that caused the SAT refutation, set to 2 if clause is in the unsatisfiable core and */
	                                    /* it is the contradiction clause that caused the SAT refutation, zero otherwise. */
	uint8_t formula[1];                 /* Dummy for start of formula items. This must be the last field of this structure. */
} satclause;

/* Structure for statistic information. */
/* If new fields are added to this structure the following changes must be made */
/* to the involved functions: */
/* - Search for an existing field and add the appropriate equivalent code */
/*   for the new field. */
/* - Update totals calculation in PrintStatData() function if appropriate. */
/* - Add code for printing proof lines in PrintProof() function if appropriate. */
typedef struct sstats {
	uint64_t tautologies;               /* Number of tautologies detected. */
	uint64_t fwdsubsum;                 /* Number of forward subsumptions performed. */
	uint64_t bcksubsum;                 /* Number of backward subsumptions performed. */
	uint64_t equivfact;                 /* Number of equivalent factor inferences. */
	uint64_t factors;                   /* Number of factoring inferences. */
	uint64_t removedups;                /* Number of duplicate literal removing simplifications. */
	uint64_t fwdemodulations;           /* Number of forward demodulation inferences. */
	uint64_t bkdemodulations;           /* Number of backward demodulation inferences. */
	uint64_t fwsubsresltns;             /* Number of forward subsumtion resolution inferences. */
	uint64_t bksubsresltns;             /* Number of forward subsumtion resolution inferences. */
	uint64_t triveqres;                 /* Number of trivial equality resolution inferences. */
	uint64_t desteqres;                 /* Number of destructive equality resolution inferences. */
	uint64_t resolutions;               /* Number of resolution inferences. */
	uint64_t paramodulations;           /* Number of paramodulation inferences. */
	uint64_t equresolutions;            /* Number of equality resolution inferences. */
	uint64_t equfactorings;             /* Number of equality factoring inferences. */
	uint64_t splits;                    /* Number of split inferences. */
	uint64_t prennfxform;               /* Number of pre-nnf inferences. */
	uint64_t nnfxform;                  /* Number of nnf inferences. */
	uint64_t miniscoping;               /* Number of miniscoping inferences. */
	uint64_t skolemize;                 /* Number of skolemization inferences. */
	uint64_t predicatedef;              /* Number of predicate definitions for sub-formula renaming. */
	uint64_t formularenaming;           /* Number of formula renaming inferences. */
	uint64_t cnfconversion;             /* Number of CNF conversion inferences. */
	uint64_t tfsimplif;                 /* Number of $true and $false simplifications. */
	uint64_t sattautologies;            /* Number of tautologies detected in SAT clauses. */
	uint64_t satremovedups;             /* Number of duplicate literal removing simplifications in SAT clauses. */
	uint64_t satsubsum;                 /* Number of subsumptions performed for SAT clauses. */
	uint64_t nbactive;                  /* Number of active clauses. */
	uint64_t nbpassive;                 /* Number of passive clauses. */
	uint64_t nbunproc;                  /* Number of unprocessed clauses. */
	uint64_t numitems;                  /* Number of statistics items accumulated. */
	float memory;                       /* Memory used. */
	double elapsed_time;                /* Elapsed time in seconds. */
} stats;

/* Structure for program options. */
typedef struct soptions {
	int32_t fweight;                    /* Type of function weight (UNIFORM or ARITY). */
	int32_t termord;                    /* Terms ordering type (STANDARD or NONRECURSIVE). */
	int32_t litord;                     /* Literals ordering type (STANDARD, NONRECURSIVE or LEXICOGRAPHIC). */
	int32_t select;                     /* Type of literal selection (MAXIMAL, SINGLENEG or MULTINEG). */
	int32_t equivfact;                  /* Equivalent factoring is enabled if this is not zero. */
	int32_t factoring;                  /* Factoring is enabled if this is not zero. */
	int32_t algorithm;                  /* Selected algorithm (OTTER or DISCOUNT). */
	int32_t mxwghtopt;                  /* MXWEIGHTxxx (see definitions above). */
	uint32_t maxweight;                 /* Maximum weight allowed. */
	int32_t factoridx;                  /* Index to weight factors to be used in clause weight calculations. */
	                                    /* See PNOTINCONJFACTOR, ENOTINCONJFACTOR, FNOTINCONJFACTOR and POSLITERALFACTOR definitions. */
	int32_t split;                      /* Splitting and interface between first order prover and */
	                                    /* SAT solver is enabled if this is not zero. */
	float timeout;                      /* Timeout setting. */
} options;

/* Structure for hash separate symbol chaining. */
typedef struct shashchain {
	struct shashchain *nextelem;        /* Pointer to next hashchain element or NULL if last. This pointer is */
	                                    /* used for freeing memory allocated to hashchain elements. */
	struct shashchain *nextsymbol;      /* Pointer to next hashchain element of the same hash value */
	                                    /* or NULL if last. */
	uint8_t **discr1;                   /* Address of set of pointers to discrimination trees for active clauses. */
	                                    /* Used for selected positive literals and maximal root terms of positive */
	                                    /* selected equalities. */
	                                    /* Used by Resolution(), PM2Clause(), BckSubsuming(), FwSubsResltn() */
	                                    /* and BkSubsResltn() functions. */
	                                    /* There is one pointer in the set for each KB in the sets corresponding */
	                                    /* to this and to the following fields in this structure. */
	                                    /* WARNING - WARNING - WARNING: If additional fields discrx or similar */
	                                    /* fields that are pointers to sets that are placed just after the structure */
	                                    /* are added then the macro SPACE_AFTER_HASHCHAIN must be modified with the */
	                                    /* number of fields of this type. */
	uint8_t **discr2;                   /* Address of set of pointers to discrimination trees for active clauses. */
                                        /* Used for selected negative literals, terms in selected literals that */
	                                    /* are not positive or negative equalities and terms in maximal equality */
	                                    /* root terms of positive or negative selected equalities. */
	                                    /* Used by PMFromClause(), BckSubsuming(), FwSubsResltn() and BkSubsResltn(). functions. */
	uint8_t **discr3;                   /* Address of set of pointers to discrimination trees for active clauses. */
	                                    /* Used for non selected positive literals and all terms. */
	                                    /* Used by CheckFwSubsuming(), BckSubsuming(), BckDemodulation(), FwSubsResltn() */
	                                    /* and BkSubsResltn() functions. */
	uint8_t **discr4;                   /* Address of set of pointers to discrimination trees for active clauses. */
	                                    /* Used for non selected negative literals and maximal root terms of positive */
	                                    /* equalities that are suitable for performing demodulation from them. */
                                        /* Used by CheckFwSubsuming(), BckSubsuming(), BckDemodulation(), FwSubsResltn() */
                                        /* and BkSubsResltn() functions. */
	uint8_t **discr5;                   /* Address of set of pointers to discrimination trees for passive clauses. */
	                                    /* Used for all positive literals and maximal root terms of positive */
	                                    /* equalities that are suitable for performing demodulation from them. */
                                        /* Used by CheckFwSubsuming(), BckSubsuming(), BckDemodulation(), FwSubsResltn() */
                                        /* and BkSubsResltn() functions. */
	uint8_t **discr6;                   /* Address of set of pointers to discrimination trees for passive clauses. */
	                                    /* Used for all negative literals and all terms. */
                                        /* Used by CheckFwSubsuming(), BckSubsuming(), BckDemodulation(), FwSubsResltn() */
                                        /* and BkSubsResltn() functions. */
	int32_t arity;                      /* Arity of symbol. */
	int32_t type;                       /* Type flags. Valid flags are FUNCTION, PREDICATE, EQUALITY and DEFINEDPRED */
	float weight[SYMBOLDIM1];           /* Symbol weights for weight queues for evaluating the clause weight used to check */
	                                    /* clause retention. See SYMBOLWEIGHT definition above. */
	                                    /* Index 0 refers to pure symbol weight, depending on its kind (variable,predicate or */
	                                    /* function, see VARWEIGHT, PREDWEIGHT and FUNCWEIGHT definitions). */
	                                    /* Indices 1 and 2 refers to the weight as defined above multiplied by appropriate factors */
	                                    /* if the symbol doesn't appear in the conjecture for two different weight parameterization. */
	int32_t precedence;                 /* Symbol precedence value for ordering purposes. */
	uint32_t hash;                      /* Symbol 32 bits hash value corresponding to symbol name. */
} hashchain;

/* Structure for hash component symbol chaining. */
typedef struct shashchcmp {
	struct shashchcmp *nextelem;        /* Pointer to next hashchcmp element or NULL if last. This pointer is */
	                                    /* used for freeing memory allocated to hashchcmp elements. */
	struct shashchcmp *nextsymbol;      /* Pointer to next hashchcmp element of the same hash value */
	                                    /* or NULL if last. */
	struct shashchcmp *nextundef;       /* Pointer to next undefined hashchcmp element of the same component. */
	struct shashchcmp *prevundef;       /* Pointer to previous undefined hashchcmp element of the same component. */
	struct sasymbol *frstposassrt;      /* Pointer to first positive asymbol in assertions of first order prover. */
	struct sasymbol *frstnegassrt;      /* Pointer to first negative asymbol in assertions of first order prover. */
	struct sasymbol *firstlock;         /* Pointer to first asymbol in locks of first order prover. */
	struct sasymbol *frstsat;           /* Pointer to first asymbol in SAT solver. */
	struct ssatnode *satnode;           /* Pointer to satnode that assigned a value to this component. */
	uint8_t *formula;                   /* Pointer to clause binary formula of component clause associated */
	                                    /* with the P-predicate. */
	int32_t size;                       /* Size in bytes of formula including ending UNITEND byte. */
	int32_t literals;                   /* Number of literals in formula. */
	int32_t number;                     /* Number assigned to symbol name, the xxx in spl_xxx name. */
	int32_t defnumber;                  /* Number assigned to the symbol definition in the proof. The component */
	                                    /* clause is assigned the number defnumber+1. */
	uint32_t hash;                      /* Symbol 32 bits hash value corresponding to clause component. */
	uint32_t flags;                     /* Flags for this component, see definitions above. */
	uint32_t npositive;                 /* Number of clauses in SAT solver with a positive literal for this component. */
	uint32_t nnegative;                 /* Number of clauses in SAT solver with a negative literal for this component. */
	uint32_t absdif;                    /* Absolute value of npositive field minus nnegative field. */
} hashchcmp;

/* Structure for coefficients of weight ordering polynomials. */
typedef struct scoeff {
	int32_t alpha0;                     /* First coefficient of weight polynomial. */
	int32_t *alpha;                     /* Pointer to second and subsequent coefficients of weight polynomial. */
	struct scoeff *eqterm1;             /* Pointer to coeff structure of first root term in an equality. */
	                                    /* NULL if current structure is not for an equality literal. */
	struct scoeff *eqterm2;             /* Pointer to coeff structure of second root term in an equality. */
	                                    /* NULL if current structure is not for an equality literal. */
} coeff;

/* Structure for coefficients of weight ordering polynomials, compact format. */
/* These are used for equality data saved in long term storage. */
typedef struct scoeffcpt {
	int32_t coefnb;                     /* Coefficient number. Number 0 corresponds to symbols coefficient. */
	                                    /* Number n corresponds to variable n-1 for n>1. A negative value */
	                                    /* indicates the end of the list. */
	int32_t coef;                       /* Coefficient value. Not used in the last list element. */
} coeffcpt;

/* Structure for coefficients of weight ordering polynomials, compact format and linked list. */
/* These are used for polynomials saved in short term storage. The linked list is ordered */
/* for increasing values of the coefnb field. */
typedef struct slkedcoef {
	int32_t coefnb;                     /* Coefficient number. Number 0 corresponds to symbols coefficient. */
	                                    /* Number n corresponds to variable n-1 for n>1. */
	int32_t coef;                       /* Coefficient value. */
	struct slkedcoef *next;             /* Pointer to next polynomial term, NULL if last term. */
} lkedcoef;

/* Structure for linear polynomial node of tree structure for evaluating */
/* ordering constraints of equality root terms. */
typedef struct sequpnode {
	coeffcpt *coefs;                    /* Pointer to linear coefficients of term in compact form. */
	int32_t result;                     /* Indicates the result (EQUAL, GREATER, LESSER or FAILURE) of an execution */
	                                    /* of CompareEquTerms(), that is, with no substitution applied. It also */
	                                    /* indicates the result of the call to CompareItemWeight2() done */
	                                    /* from CompareEquTerms() with  EQUALW, NOGREATER, NOLESSER */
	                                    /* or FAILUREW flags. */
	struct sequpnode *subterms[1];      /* Set of pointers to child (subterms) nodes. This must be the last of this structure. */
} equpnode;

/* Structure for symbols (literals and functions) items data. */
typedef struct ssymbol {
	int32_t offset;                     /* Offset from beginning of formula field buffer to start of current item. */
	int32_t nextoff;                    /* Offset from beginning of formula field buffer to next item */
	                                    /* that is not a subterm of current item. */
	hashchain *symbol;                  /* Pointer to hashchain structure for the item symbol. */
	equpnode *leftnode;                 /* Pointer to left node equpnode structure. */
	equpnode *rightnode;                /* Pointer to right node equpnode structure. */
} symbol;

/* Structure for assertion symbols (P-predicates, component names). */
typedef struct sasymbol {
	hashchcmp *symbol;                  /* Pointer to hashchain structure for the item symbol. */
	struct sasymbol *next;              /* Pointer to next asymbol structure for the same component of the same class. */
	                                    /* (assertion, lock or SAT clause literal). */
	struct sasymbol *prev;              /* Pointer to previous asymbol structure for the same component of the same class. */
	union {
		binprefix *binp;                /* Pointer to A-clause binary formula to which the asymbol belongs if asymbol */
		                                /* belongs to assertions or lock assertions of an A-clause. */
		satclause *satcl;               /* Pointer to SAT clause binary formula to which the asymbol belongs if asymbol */
		                                /* belongs to a SAT clause. */
	} part2;
} asymbol;

/* Structure for variable nodes in discrimination trees. */
typedef struct sdsvarnode {
	uint8_t *next;                      /* Pointer to next child of the same parent node. */
	uint16_t clsbrkts;                  /* Number of brackets closed just before the term of this node. */
	                                    /* This must be the last field of this structure. */
} dsvarnode;

/* Structure for function nodes in discrimination trees. */
typedef struct sdsfncnode {
	uint8_t *next;                      /* Pointer to next child of the same parent node. */
	hashchain *symbol;                  /* Pointer to hashchain structure for the function symbol. */
	uint16_t clsbrkts;                  /* Number of brackets closed just before the term of this node. */
	                                    /* This must be the last field of this structure. */
} dsfncnode;

/* Structure for end nodes in discrimination trees. */
typedef struct sdendcnode {
	int32_t size;                       /* Number of total symbol pointers allocated. */
	int32_t used;                       /* Number of used symbol pointers. */
	uint8_t **symbol;                   /* Pointer to set of pointers to item in the clause that matches the tree leave. */
} dsendnode;

/* Structure for SAT solver node. */
typedef struct ssatnode {
	struct ssatnode *nextnode;          /* Pointer to next SAT solver node, NULL if last.*/
	struct ssatnode *prevnode;          /* Pointer to previous SAT solver node, NULL if first.*/
	hashchcmp *symbol;                  /* SAT symbol that was assigned a value true or false in this node. */
	int32_t flags;                      /* Flags for this node: POSITIVE or NEGATIVE and possibly LASTOPTION. */
	satclause *forcedcl;                /* Pointer to unit clause that forced the value of the node symbol field, */
	                                    /* NULL if the node was not forced by a unit clause. */
	uint32_t number;                    /* Sequence number for this node. */
} satnode;

/* Structure for proof node. */
typedef struct sproofnode {
	struct sproofnode *nextnode;        /* Pointer to next SAT solver node, NULL if last.*/
	struct sproofnode *prevnode;        /* Pointer to previous SAT solver node, NULL if first.*/
	int32_t type;                       /* Type of node, see definitions above. */
	union {
		hashchcmp *symbol;              /* Pointer to component symbol. */
		satclause *satclause;           /* Pointer to SAT clause. */
		cmprefix *Aclause;              /* Pointer to A-clause. */
	} ptr;
} proofnode;

/* Structure for vertical features scan node. */
typedef struct svertnode {
	struct svertnode *nextnode;         /* Pointer to next node, NULL if last.*/
	struct svertnode *prevnode;         /* Pointer to previous node, NULL if first.*/
	uint8_t *item;                      /* Pointer to clause item. */
	uint8_t *nextitem;                  /* Next item jumping over subterms of current item. */
	int32_t depth;                      /* Node depth, starting with zero. */
} vertnode;

/* Structure for learning record. */
typedef struct slearnrecord {
	uint64_t header;                    /* Record header with flags indicating the relevant part of the strategy */
	                                    /* to which the learn record applies. The relevant strategy flags are DISCOUNT_P, */
	                                    /* MULTINEG_P, MAXIMAL_P, TRMORDNR_P, LITORDNR_P, LITORDLEX_P, ARITY_P */
	                                    /* and GOODCHOICE. */
	double vector[VECTORLENGTH];        /* Record features vector. See GetClauseVector() function for a description */
	                                    /* of each feature in vector. */
} learnrecord;

/* Structure for learning feature vector. */
typedef struct slearnvector {
	double vector[VECTORLENGTH];        /* Features vector. See GetClauseVector() function for a description */
	                                    /* of each feature in vector. */
} learnvector;

/* Structure for KB definition. */
typedef struct skbase {
	cmprefix *firsttxt;                 /* Pointer to first text formula, NULL if no text formulas added yet. */
	                                    /* Older formulas are placed first and this pointer points to the oldest */
	                                    /* text formula. */
	cmprefix *lasttxt;                  /* Pointer to last text formula, NULL if no text formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* text formula. */
	cmprefix *frstpassive;              /* Pointer to first passive formula, NULL if no passive formulas added yet. */
                                        /* Older formulas are placed first and this pointer points to the oldest */
                                        /* passive formula. */
	cmprefix *lstpassive;               /* Pointer to last passive formula, NULL if no passive formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* passive formula. */
	cmprefix *frstactive;               /* Pointer to first active formula, NULL if no active formulas added yet. */
                                        /* Older formulas are placed first and this pointer points to the oldest */
                                        /* active formula. */
	cmprefix *lstactive;                /* Pointer to last active formula, NULL if no active formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* active formula. */
	cmprefix *frstunproc;               /* Pointer to first unprocessed formula, NULL if no unprocessed formulas added yet. */
                                        /* Older formulas are placed first and this pointer points to the oldest */
                                        /* unprocessed formula. */
	cmprefix *lstunproc;                /* Pointer to last unprocessed formula, NULL if no unprocessed formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* unprocessed formula. */
	cmprefix *frstlocked;               /* Pointer to first locked formula, NULL if no locked formulas added yet. */
                                        /* Older formulas are placed first and this pointer points to the oldest */
                                        /* locked formula. */
	cmprefix *lstlocked;                /* Pointer to last locked formula, NULL if no locked formulas added yet. */
                                        /* Newer formulas are placed last and this pointer points to the newest */
                                        /* locked formula. */
	cmprefix *frstglweight[WEIGHTQUEUES];  /* Pointers to first formula in weight queue for each weight queue, */
	                                    /* each pointer is NULL if no formulas of that weight added yet. */
	                                    /* Only passive clauses are linked to weight queue. */
	cmprefix *frstweight[WEIGHTQUEUES][MAXWEIGHT+1];  /* Pointers to first formula in weight queue for each weight queue and for */
	                                    /* each weight, each pointer is NULL if no formulas of that weight added yet. */
	                                    /* Only passive clauses are linked to weight queue. */
	cmprefix *lstweight[WEIGHTQUEUES][MAXWEIGHT+1];  /* Pointers to last formula in weight queue for each weight queue and for */
	                                    /* each weight, each pointer is NULL if no formulas of that weight added yet. */
	                                    /* Only passive clauses are linked to weight queue. */
	cmprefix *lastdisp;                 /* Pointer to last formula in disposal queue, NULL if no formulas */
	                                    /* added yet. This queue interconnects clauses from other queues. */
	proofnode *frstprfnode;             /* Pointer to first proof node. */
	satclause *firstsatq;               /* Pointer to first SAT clause in the SAT queue. */
	satclause *lastsatq;                /* Pointer to last SAT clause in the SAT queue. */
	satclause *frstsatglbl;             /* Pointer to first SAT clause in the SAT solver global queue. */
	satclause *lastsatglbl;             /* Pointer to last SAT clause in the SAT solver global queue. */
	satclause *frstunsats;              /* Pointer to first SAT clause in the SAT solver ordered clause queue. */
	satclause *lstunsats;               /* Pointer to last SAT clause in the SAT solver ordered clause queue. */
	satclause *frstunsats2;             /* Pointer to start of the part of SAT solver ordered clause queue with clauses */
	                                    /* that are unsatisfied and have two or more undefined literals. */
	hashchcmp *frsthshundf;             /* Pointer to first undefined hashchcmp element for this KB in the hashchcmp ordered queue. */
	hashchcmp *frsthunbp;               /* Pointer to first undefined hashchcmp element with clauses with negative and */
	                                    /* positive literals (both polarities) for this KB in the hashchcmp ordered queue. */
	hashchcmp *lasthshundf;             /* Pointer to last undefined hashchcmp element for this KB in the hashchcmp ordered queue. */
	satnode *satrootnode;               /* SAT solver root node pointer. */
	double *lrnvector;                  /* Pointer to learning vector or NULL if no learning vector isavailable. */
	dsendnode vartrmidx;                /* Indexes of positive equalities selected for paramodulating from them in active clauses */
	                                    /* that have maximal root terms that are variables. */
	uint64_t signature;                 /* Last signature used for inference process. */
	uint64_t locksignature;             /* Last signature used for lock and unlock processes. */
	int32_t flags;                      /* KB flags, see definitions above. */
	int32_t nformulas;                  /* Number of formulas in KB so far. */
	options opts;                       /* Options for KB. */
	int32_t selectcnt;                  /* Number of extractions done so far in the current round-robin clause extraction cycle. */
	int32_t lrscnt;                     /* Count of clauses removed from the passive queue in the current LRS cycle. */
	uint32_t splsymbols;                /* Number of P-predicates for P-predicate name assignment. */
	int32_t status;                     /* KB status (see flags above). */
	struct timeval time;                /* Start time of current saturation algorithm. */
	struct timeval endtime;             /* End time of current saturation algorithm. */
	stats prstats;                      /* Thread statistics for current proof. */
} kbase;

/* Structure for substitutions. See Unify() function for substitution format. */
typedef struct ssubst {
	uint8_t *buffer;                    /* Pointer to buffer with substitution. May be NULL. */
	                                    /* See Unify() function for format description. */
	uint32_t buffsize;                  /* Total buffer size in bytes. */
	uint32_t substsize;                 /* Size in bytes of substitutions including the UNITEND byte. */
} subst;

/* Structure to store subsuming related data for each literal in a subsuming or subsumed formula. */
typedef struct ssubsumdata {
	int32_t lits1;                      /* Number of literals in candidate subsumed formula. */
	int32_t lits2;                      /* Number of literals in candidate subsuming clause. */
	uint8_t ***pinst;                   /* Address of set of pointers, one pointer for each literal in the candidate */
	                                    /* subsuming clause. Each pointer is the address of the set of pointers to */
	                                    /* instance literals in candidate subsumed formula for each candidate subsuming */
	                                    /* literal. */
	int32_t **instnb;                   /* Address of set of pointers, one pointer for each literal in the candidate */
                                        /* subsuming clause. Each pointer is the address of the set of instance literal */
	                                    /* numbers in candidate subsumed formula for each candidate subsuming literal. */
	int32_t *nbofinst;                  /* Address of set containing the number of instance literals for each literal */
	                                    /* in the subsuming clause. The number of instance literals is set to -1 if */
	                                    /* there is an instance literal identical to the literal in the subsuming clause. */
} subsumdata;

/* Structure for discrimination tree stack element. */
typedef struct sstkelem {
	uint8_t *node;                      /* Pointer to corresponding node in stack. */
	uint8_t *subterm;                   /* Pointer to corresponding subterm in query term. */
	union {
		int32_t index;                  /* Index of dsendnode structure symbol field (only for end nodes). */
		int32_t pndbrkts;               /* Number of brackets that must be closed after the node term corresponding */
		                                /* to the stack element in order to terminate the traversing of a tree term */
		                                /* corresponding to a variable in the query term (for nodes other than end nodes). */
	} data;
} stkelem;

/* Structure for discrimination tree search stack to allow backtracking. */
typedef struct sdscstack {
	int32_t size;                       /* Number of total stack elements allocated. */
	int32_t used;                       /* Number of used stack elements. */
	stkelem element[1];                 /* First stack element. This must be the last field of this structure. */
} dscstack;

/* Structure for linked list formula format. */
typedef struct slkedfitem {
	struct slkedfitem *next;            /* Pointer to next linked list item. */
	struct slkedfitem *prev;            /* Pointer to previous linked list item. */
	union {
		struct slkedfitem *mtchbrckt;   /* Pointer to matching bracket item. */
		nameschain *vardata;            /* Pointer to nameschain structure of a quantifier or variable item. */
		hashchain *symbol;              /* Pointer to hashchain structure of a predicate or function item. */
	} ptr;
	union {
		hashchain *sksymbol;            /* Pointer to hashchain skolem symbol in existential quantifier or variable items. */
		int16_t varnumber;              /* Number assigned to standard name of variable in a quantifier or variable item. */
		struct {
			int32_t posclauses;         /* Lower bound of number of clauses generated for a block for the */
			                            /* open bracket or literal starting the block. */
			int32_t negclauses;         /* Lower bound of number of clauses generated for the negation of a block*/
			                            /* for the open bracket or literal starting the block. */
		} clauses;
	} data;
	struct slkedfitem *dupq;            /* Auxiliary pointer to duplicate quantifier, used when duplicating blocks. */
	int32_t flags;                      /* Item flags. */
} lkedfitem;

/* Structure for multi-threading control. */
typedef struct smpctl {
	clock_t start;                      /* CPU ticks at start of saturation process. */
	clock_t end;                        /* CPU ticks at end of saturation process. */
	int32_t status;                     /* Process status. See definitions above. */
	int32_t solvekb;                    /* First KB / thread to get SATISFIABLE or UNSATISFIABLE condition. */
	omp_lock_t omp_lockc1;              /* Lock for process status updates. */
	omp_lock_t omp_lockc2;              /* Lock for KB re-initialization in saturation process. */
} mpctl;

/*----------------------------------------------------------------
 *
 *  Global variables shared with other modules that are initialized here.
 *  WARNING: Don't use "static" as then a different variable will be created in
 *  each module with the same name.
 *
 *--------------------------------------------------------------*/
#ifndef VARTYPE
#define VARTYPE extern
extern double tot_cpu_time; /* Total CPU time used in seconds. */
extern int32_t tot_pbms; /* Total number of problems examined. */
extern int32_t solved_pbms; /* Total number of problems solved. */
extern int32_t queuerts[WEIGHTQUEUES+1]; /* Ratio numbers for weight queues and age queue with learning disabled. */
extern int32_t queuerts2[WEIGHTQUEUES+1]; /* Ratio numbers for weight queues and age queue with learning queue. */
extern int32_t *slctordr; /* Pointer to clause selection order queue set with learning disabled. The I-th clause selection */
                          /* is picking the clause from the slctordr[I] weight queue if slctordr[I]<WEIGHTQUEUES, */
                          /* otherwise the clause is picked from the age queue. */
extern int32_t *slctordr2; /* Pointer to clause selection order queue set with learning enabled. */
extern int32_t wqdeplord[WEIGHTQUEUES]; /* Weight queues depletion order with learning disabled. If wqdelord[I]=J then */
                                        /* the weight queue with J index is depleted I-th. */
extern int32_t wqdeplord2[WEIGHTQUEUES]; /* Weight queues depletion order with learning enabled. */
extern hashchain *kb_hash; /* First chained hashchain element. */
extern hashchcmp **kb_hashcmp; /* Address of set of pointers to first chained hashchcmp element for each thread. */
extern hashchcmp ***hashsplit; /* Address of set of pointers to thread hash keys indexed by hash values of clause variants. */
extern int32_t syntaxmode; /* Syntax type: TPTP or DRODI. */
extern int32_t proofmode; /* Proof printing type: TPTP or DRODI. */
extern int32_t prtkbdata; /* PRINTKBDATA type: PRTWINNER, PRTALL or PRTOFF. */
extern int32_t verbose; /* Set to 1 if VERBOSE option is ON, zero otherwise. */
extern char *importfile; /* Pointer to IMPORT file name or NULL. */
extern char *learnto; /* Pointer to LEARNTO file name or NULL. */
extern char *learnfrom; /* Pointer to LEARNFROM file name or NULL. */
extern int32_t learnnew; /* If zero then learn-to output will be appended to learn-to file. Otherwise a new file */
                         /* will be created. */
extern int32_t loadredundant; /* If not zero then add formulas with lemma, theorem and corollary roles, */
                       /* otherwise these formulas are not loaded. */
extern int32_t timeslices; /* Number of time slices. */
extern double ppthrshld; /* Threshold for the preprocessing clause relevance filtering. */
                  /* See GlobalPreprocess() global comments for details. */
extern int32_t prproflags; /* Preprocessing clause filtering option flags. This can be CLFILTERINGOFF, CLFILTERINGON, */
                    /* CLFILTERINGNONEQU and/or CLFILTERINGRESIGN (see definitions above). */
extern int16_t pbmtypestrats[2][8][54]; /* Strategies to use for each type of problem. */
                                     /* See InitStrategies() for details. */
extern int16_t nbpbmtypestr[2][8]; /* Each element contains the number of MIN and MED strategies in pbmtypestrats[][] */
                                   /* global variable. See InitStrategies() for details. */
extern uint32_t usrparam; /* User parameter options. */
extern uint32_t usrparam_mask; /* User parameter options mask. */
extern volatile sig_atomic_t signalflag; /* Not zero if a program error caught signal process is in progress. */
extern volatile sig_atomic_t signalflag2; /* Not zero if a SIGALRM or SIGXCPU signal has been detected. */
extern uint64_t allcdmemory; /* Currently allocated memory. */
#if defined(DEBUGCODE) && defined(VERBOSE)
extern int32_t verbfrom; /* Initial clause number for verbose tracking. */
extern int32_t verbto; /* Final clause number for verbose tracking. */
#endif
#else
double tot_cpu_time=0.0;
int32_t tot_pbms=0;
int32_t solved_pbms=0;
int32_t queuerts[WEIGHTQUEUES+1]={0,1,2,2,0,1};
int32_t queuerts2[WEIGHTQUEUES+1]={6,1,2,2,0,1};
int32_t *slctordr=NULL;
int32_t *slctordr2=NULL;
int32_t wqdeplord[WEIGHTQUEUES]={0,1,2,3,4};
int32_t wqdeplord2[WEIGHTQUEUES]={0,1,2,3,4};
hashchain *kb_hash=NULL;
hashchcmp **kb_hashcmp=NULL;
hashchcmp ***hashsplit=NULL;
int32_t loadredundant=1;
int32_t syntaxmode=TPTP;
int32_t proofmode=TPTP;
int32_t prtkbdata=PRTOFF;
int32_t verbose=0;
char *importfile=NULL;
char *learnto=NULL;
int32_t learnnew=0;
char *learnfrom=NULL;
int32_t timeslices=3;
double ppthrshld=0.1;
int32_t prproflags=CLFILTERINGNONEQU;
int16_t pbmtypestrats[2][8][54]=
	{{{185,40,857,913,531,149,1275,322,764,274,715,1104,741,623,505,600,669,233,1159,390,717,1138,75,433,678,515,300,1033,1387,1240,1337,
			240,471,383,795,998,313,658,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{204,235,101,428,924,362,585,370,185,1330,537,825,130,22,166,60,678,348,1372,210,687,676,864,658,1258,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
			0,0,0,0,0,0,0,0,0,0,0,0,0},
	{451,687,678,57,649,676,217,667,679,658,647,666,669,537,656,659,377,657,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
			0,0,0,0,0},
	{62,385,979,679,1083,688,667,140,60,955,396,1058,78,57,1153,144,647,162,248,669,1069,347,666,464,668,998,422,70,649,115,227,0,0,0,0,0,
			0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{252,1262,1255,905,714,622,653,591,635,44,620,141,1035,1256,704,444,773,669,1111,70,273,953,301,1145,63,115,667,785,293,465,33,217,35,
			511,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{1161,555,31,803,5,57,301,75,1011,785,463,537,225,669,1390,871,679,471,990,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
			0,0,0,0,0,0},
	{143,697,956,622,625,679,377,317,261,654,908,263,677,142,185,755,301,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
			0,0,0,0,0},
	{541,225,315,952,354,940,663,678,21,941,71,505,275,217,921,771,461,421,75,661,285,623,793,383,1275,669,901,605,1338,245,536,0,0,0,0,0,
			0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}},
	{{2175,2153,2403,1721,2001,105,2173,2705,1454,2701,2025,2314,1930,2355,2121,669,1580,1551,1574,1926,2418,1761,1622,1745,2021,1942,1751,
			2271,2416,1851,1561,1977,233,2373,1909,658,1573,1476,1387,2195,471,1592,1957,2039,2593,1562,2557,383,433,678,795,390,1553,313},
	{204,2075,1817,2579,1821,2412,1643,1505,2470,2591,1475,2331,2205,2415,1511,1812,2446,1713,2131,60,2013,2095,2124,1818,1543,657,1623,1473,
			1867,2381,2139,1464,428,1907,1754,235,537,864,1258,1808,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{1655,697,1583,678,1555,1661,669,1635,2381,2265,1645,676,2221,667,1775,1867,1919,1783,687,1497,1561,1847,658,1883,377,647,1461,656,649,0,
			0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{1802,1805,2423,1509,1945,1673,1495,2598,2361,1686,1893,1511,2695,1923,1058,2576,1496,2180,2261,1461,1459,2743,1497,2645,1839,1778,1935,
			1557,1808,385,1645,1812,2413,1677,2607,1593,1643,1806,955,647,2690,2129,1553,2137,1591,1533,1943,2311,2096,1920,70,2304,1759,0},
	{2141,252,2557,905,1532,714,1451,1689,622,2254,2630,1742,2153,620,2481,1502,2440,1555,1821,2532,1787,2180,1945,1542,1702,1595,2075,2111,
			669,1581,33,2163,667,1995,2061,1797,785,1871,2181,2381,1715,217,1980,0,0,0,0,0,0,0,0,0,0,0},
	{1741,2235,2141,331,2185,1841,2191,1765,1977,1843,57,31,1817,1987,1595,2137,1943,537,669,1390,871,679,471,990,0,0,0,0,0,0,0,0,0,0,0,0,0,
			0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{1823,2151,1497,2781,2593,2130,2528,2379,2061,1950,2387,1735,2721,1483,1939,2179,2129,2730,1651,2031,1962,2416,2140,1977,2181,2720,1969,
			2091,1695,2735,2080,1716,2159,2750,1935,2041,2301,2343,755,2119,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
	{2273,2054,354,2382,1960,2243,2137,1861,1981,2211,1715,71,1813,1621,771,217,1643,2774,1817,2343,1920,285,2263,1989,1822,75,1675,1275,1633,
			669,1338,536,245,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}}};
uint16_t nbpbmtypestr[2][8]={{38, 25, 18, 31, 34, 19, 17, 31},{54, 40, 29, 53, 43, 24, 40, 33}};
uint32_t usrparam=0;
uint32_t usrparam_mask=0;
volatile sig_atomic_t signalflag=0;
volatile sig_atomic_t signalflag2=0;
uint64_t allcdmemory=0;
#if defined(DEBUGCODE) && defined(VERBOSE)
int32_t verbfrom=0;
int32_t verbto=0x7fffffff;
#endif
#endif

/*----------------------------------------------------------------
 *
 *  Global variables shared with other modules that are not initialized here.
 *  WARNING: Don't use "static" as then a different variable will be created in
 *  each module with the same name.
 *
 *--------------------------------------------------------------*/
VARTYPE char *oldenv; /* Original environment on program entry. */
VARTYPE kbase **kbset; /* Address of pointers to allocated working KB's. */
VARTYPE kbase *mainkb; /* Pointer to main KB with axioms (theory) being analyzed. */
VARTYPE int num_wkbs;  /* Number of working KB's, equals the number of cores */
                       /* enabled. See Create_KB() function for details. */
VARTYPE int last_num_wkbs;  /* Value of num_wkbs before changing the number of cores. */
VARTYPE int num_cores;  /* Number of physical cores. */
VARTYPE int proc_affinity; /* Core affinity option. */
VARTYPE uint64_t numskolem; /* Number of skolem functions defined in main KB so far. */
VARTYPE int32_t skolemid; /* Number to identify unique skolem prefix. */
VARTYPE char skfprefix[15]; /* Prefix for unique skolem function names. */
VARTYPE uint64_t numpdefs; /* Number of predicate definitions in main KB so far. */
VARTYPE int32_t pdefid; /* Number to identify unique predicate definition prefix. */
VARTYPE char prdprefix[15]; /* Prefix for unique predicate definition names. */
VARTYPE int32_t splid; /* Number to identify unique P-predicate prefix. */
VARTYPE char splprefix[15]; /* Prefix for unique P-predicates. */
VARTYPE uint32_t numsymbols; /* Number of symbols for symbol weight assignment. */
VARTYPE hashchain *hashkeys[HASHVALUES]; /* Hash keys indexed by hash values of symbol names. */
VARTYPE uint32_t hashpos[MAXPOSITIONS]; /* Hash values for item positions. */
VARTYPE hashchain equkey; /* Key corresponding to equality literal. */
VARTYPE uint8_t **equptrs; /* Address of sets of pointers for equality hashchain element. */
VARTYPE int32_t binpsize; /* Effective size of binprefix structure. */
VARTYPE int32_t satclsize; /* Effective size of satclause structure. */
VARTYPE int32_t vnodesize; /* Effective size of dsvarnode structure. */
VARTYPE int32_t fnodesize; /* Effective size of dsfncnode structure. */
VARTYPE mpctl procctl; /* Multithreading control data. */
VARTYPE double cpu_time; /* CPU time used in seconds. */
VARTYPE stats glblstats; /* GLobal statistics. */
VARTYPE options glblopts; /* Global options. */
VARTYPE int32_t pbmflags; /* Flags for current problem. See definitions above. */
VARTYPE int32_t pbmtype; /* Problem type number. */
VARTYPE int32_t pgmmode; /* Problem mode: INTERACTIVE or NONINTERACTIVE. */
VARTYPE int32_t slctordrsize; /* Number of elements in slctordr set. */
VARTYPE int32_t slctordrsize2; /* Number of elements in slctordr2 set. */
VARTYPE uint16_t strategies[MAXSTRATEGIES]; /* Strategies selected for problem being solved. */
VARTYPE int32_t nextstrat; /* Index to next free strategy in the strategies[] set. */
VARTYPE int32_t nbstrats; /* Total number of strategies in the strategies[] set. */
VARTYPE learnrecord *lrnrecords; /* Learning records with headers and vectors from learning file. */
VARTYPE int32_t numlrnrecs; /* Number of record elements in set pointed by learnrecord. */
VARTYPE uint64_t avmemory; /* Available memory at program start in bytes as obtained from sysconf(). */
VARTYPE uint64_t maxmemory; /* Memory limit to be used except for ProcessLearnData() function. */
VARTYPE uint64_t rlim_cur; /* Soft memory limit at program start. */
#ifdef DEBUGMEMORY
VARTYPE void *memptrs[DBGMEMELEMS]; /* Pointers to memory allocations that must be tracked when DEBUGMEMORY is defined. */
                            /* The actual size of this set must be adjusted for each case. */
VARTYPE int32_t memtrack[DBGMEMELEMS]; /* Status of each entry in memptrs[]. The actual size of this set must be equal to the */
                               /* size of memptrs[]. The content is: */
                               /* 0 -> Entry not used. */
                               /* 1 -> Entry for memory allocated but not yet freed. */
                               /* 2 -> Entry for memory allocated and freed. */
VARTYPE int32_t ndemods;               /* Number of entries used in the above sets. */
VARTYPE int32_t auxcounter;            /* Auxiliary counter. */
#endif
#ifdef DEBUGCODE
VARTYPE int32_t dbgnb; /* For setting appropriate breakpoint in Otter and Discount functions. */
#endif
#ifdef BENCHMARKING
VARTYPE double bresults[3][BENCHDIMENSION]; /* Benchmarking results. */
VARTYPE int32_t bncounter; /* Benchmarking counter. */
VARTYPE char *benchfname; /* Benchmarking output file name. */
#endif

/*----------------------------------------------------------------
 *
 *  Platform independent functions common to all platform independent modules.
 *
 *--------------------------------------------------------------*/
/* Drodi module functions. */
int main(int argc,char *argv[]); /* Entry point. */
int32_t Initialize_pgm(void); /* Initialize program */
int32_t CheckExit(void); /* Checks before exit */
void Clean4Exit(void); /* Performs the needed cleaning before exit */
int32_t CheckSaved(char *,char *); /* Checks before exit */
int32_t Create_KB(int32_t); /* Create a new KB */
int32_t Initialize_KB(int32_t); /* Initialize a KB */
void Delete_KB(int nkb); /* Delete a KB */
void Free_Formula(int32_t); /* Free formula and tree memory and initialize queue pointers */
void FreeBinFormulas(int32_t); /* Free binary formulas and tree memory and initialize queue pointers */
void Initialize_hash(void); /* Initialize symbol hash tables for names in input formulas */
int32_t Alloc_hashcmp(void); /* Allocate symbols and hash tables for component names */
void Free_hashcmp(void); /*Free symbols and hash tables for component names */
void Initialize_hashcmp(int32_t); /* Initialize symbol hash tables for component names */
int32_t Tell(char *,int32_t); /* Add formula to main KB */
int32_t InputSentence(char **,int32_t *); /* Input formula lines */
int32_t CheckSyntax(char **,int32_t *,char **,nameschain *,int32_t); /* Check sentence syntax */
char *SkipBlanks(char *); /* Skips space characters */
int32_t ParseSentenceStart(char *,char **,nameschain *,int32_t); /* Get formula start type */
int32_t ParseVariableList(char *string,char **,nameschain *); /* Variable list parser */
int32_t AddNametoChain(char *,nameschain *,int32_t); /* Add name to text formulas names chain */
int32_t ParseAtomic(char *,char **,nameschain *,int32_t); /* Atomic formula parser */
int32_t ParseName(char *,char **); /* Parses a variable, constant, function or predicate name */
int32_t ParseTerm(char *,char **,char **,nameschain *,int32_t *,int32_t); /* Parse term */
int32_t CheckDeclared(char *,nameschain *,int32_t,int32_t *); /* Check if name is locally declared */
int32_t Ask(char *); /* Include a negated conjecture in main KB */
void AddTxtFormula2KB(cmprefix *,int32_t,int32_t,int32_t); /* Add text formula to a KB */
void ResetStats(stats *); /* Reset statistics structure */
void AddInferenceStats(stats *,stats *); /* Add inference statistics */
int32_t TPTPQuantifiers(char **,int32_t *); /* Set scope of quantifiers */
float FormatMemory(float,char **); /* Format memory in KB, MB and GB */
char *JumpOverQuotes(char *); /* Jump over a quoted block */
void CatchSignal1(int); /* Catch program exception signals. */
void CatchSignal2(int); /* Catch program interruption signals. */
char *strdupmn(char *); /* Memory tracking version of strdup() function */
void *mallocmn(size_t); /* Allocate and track total memory */
void *reallocmn(void *,size_t); /* Reallocate and track total memory */
uint64_t GetAvailableMemory(void); /* Get available memory */
#ifndef DEBUGMEMORY
void myfree(void *); /* Free and track total memory */
#else
void myfree(void *,char *,int); /* Performs memory debug tasks and calls free() function */
#endif
#ifdef MEMORYTUNNING
int32_t SymMalloc(uint64_t); /* Allocate additional memory for MAXMEMORY tuning */
#endif

/* clausify module functions. */
cmprefix *PreprocessFormula(cmprefix *,char **,int32_t *,int32_t); /* Text formula pre-processing */
int32_t PutBrackets(lkedfitem **ptopitem); /* Insert parenthesis according to precedence */
int32_t SafeAppend(char **,int32_t *,char *,int32_t); /* Append a string safely */
cmprefix *Convert2PreNNF(lkedfitem **,cmprefix *,char **,int32_t *); /* Perform pre-nnf conversion */
cmprefix *Convert2NNF(lkedfitem **,cmprefix *,char **,int32_t *); /* Perform nnf conversion */
void DropImplication(lkedfitem *); /* Translate an implication */
void DropRevImplication(lkedfitem *); /* Translate a reverse implication */
lkedfitem *DropEquivalence(lkedfitem *,lkedfitem **); /* Translate an equivalence */
lkedfitem *DropXOR(lkedfitem *,lkedfitem **); /* Translate a XOR connective */
void DropNAND(lkedfitem *); /* Translate a NAND */
void DropNOR(lkedfitem *); /* Translate a NOR */
int32_t IndentNegations(lkedfitem *); /* Move negations inwards */
void DropQuantifiers(lkedfitem **); /* Delete quantifiers in formula */
void RemoveLinkedListChain(lkedfitem **,lkedfitem *,lkedfitem *); /* Remove a chain of linked list items */
int32_t Convert2CNF(char *,char **,char **,int32_t *); /* Convert sentence to CNF */
int32_t CnfOrCnf(char **,int32_t *,char *); /* Merge two CNF's separated by an OR operator */
lkedfitem *Text2LinkedList(char *); /* Convert formula text to linked list format */
void FreeLkedLstFormula(lkedfitem *); /* Free formula storage in linked list format */
void RemoveRedundantBrackets(lkedfitem **); /* Remove redundant brackets */
lkedfitem *RemoveAdjBrackets(lkedfitem *,lkedfitem **); /* Remove set of adjacent redundant brackets */
void SetPolarity(lkedfitem *); /* Set polarity of linked list items */
void AddItem2LkdLstF(lkedfitem **,lkedfitem **,lkedfitem **); /* Add item to formula in linked list format */
int32_t AddOpenBracket(lkedfitem **,lkedfitem **,lkedfitem **); /* Add open bracket to formula in linked list format */
int32_t AddQuantifier(char *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain **); /* Add quantifier to formula in linked list format */
int32_t AddName(char *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add name to formula in linked list format */
int32_t AddPredOrFunc(hashchain *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add a predicate or function in linked list format */
int32_t AddEquality(char *,char **,lkedfitem **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add an (in)equality in linked list format */
int32_t AddEquRootTerm(char *,char **,lkedfitem **,lkedfitem **,
		nameschain *); /* Add an (in)equality root term in linked list format */
int32_t DuplicateBlock(lkedfitem *,lkedfitem *,lkedfitem **,lkedfitem **); /* Duplicate a block of linked list items */
void InsertBlock(lkedfitem *block,lkedfitem *,lkedfitem *); /* Insert a block of linked list items */
lkedfitem *JumpOverBlockOrTerm(lkedfitem *,int32_t); /* Jump over a block or term */
cmprefix *MiniscopeFormula(lkedfitem **,cmprefix *,char **,int32_t *); /* Miniscope a formula */
int32_t MiniscopeQuantifier(lkedfitem *,lkedfitem **,struct timeval *); /* Miniscope a quantifier */
int32_t MiniscopeSegment(lkedfitem *,lkedfitem *,lkedfitem **,struct timeval *); /* Miniscope a formula segment */
int32_t MiniscopeBlock(lkedfitem *,lkedfitem *,lkedfitem *,lkedfitem **,struct timeval *); /* Miniscope a formula block */
int32_t IsVarInBlock(lkedfitem *,lkedfitem *); /* Check if a variable exists in a block */
void RelinkVars(lkedfitem *,lkedfitem *,lkedfitem *); /* Rebuild links of variables in a block */
lkedfitem *DuplicateQuantifier(lkedfitem *,lkedfitem **,lkedfitem **,nameschain *); /* Duplicate and insert a quantifier */
cmprefix *Skolemize(lkedfitem **,cmprefix *,char **,int32_t *); /* Skolemize a formula */
cmprefix *TFSimplify(lkedfitem **,cmprefix *,char **,int32_t *); /* Simplify $true and $false predicates */
void RenameVariables(lkedfitem *); /* Rename variables in a formula */
cmprefix * AddLkdLstFormula2KB(lkedfitem *,cmprefix *,cmprefix *,int32_t,char **,
		int32_t *); /* Add linked list formula to main KB in text form */
cmprefix *FormulaRenaming(lkedfitem **,cmprefix *,char **,int32_t *); /* Formula renaming */
void GetClauseNumberBounds(lkedfitem *,int32_t *,int32_t *); /* Get lower bounds of number of clauses in a segment */
void CalcSegmentBounds(int32_t *,int32_t *,int32_t,int32_t,int32_t,int32_t,
		int32_t); /* Calculate lower bounds of number of clauses in a segment */
int32_t SegmentRenaming(lkedfitem **,lkedfitem *,int32_t,int32_t,cmprefix **,
		cmprefix *,nameschain *); /* Rename blocks within a segment */
void GetaABCoefs(int32_t,int32_t,lkedfitem *,lkedfitem *,int32_t,int32_t *,
		int32_t *); /* Get "a" and "b" coefficients of a block */
int32_t RenameBlock(lkedfitem **,lkedfitem *,cmprefix **,cmprefix *,nameschain *); /* Rename a block in a formula */

/* compile module functions. */
int32_t CompileCNF(char *formula,cmprefix *); /* Compile a CNF text formula */
int32_t CompileClause(char *,cmprefix **,cmprefix *); /* Compile a clause */
void CompileLiteral(char *,int32_t *,cmprefix *,int32_t *); /* Literal compilation */
void CompileTerm(char *,int32_t *,cmprefix *,int32_t *,int32_t); /* Term compilation */
void CompileSymbol(char *,int32_t *,cmprefix *,int32_t *,int32_t); /* Symbol compilation */
int32_t BytesRequired(char *); /* Compute the number of bytes needed for a compiled clause */
int32_t AddBinClause2KB(cmprefix *,int32_t,int32_t); /* Add binary clause to a KB. */
int32_t RenumberVars(binprefix *); /* Renumber variables in a binary formula */
int32_t SafeAppendF(char **,int32_t *,char *,int32_t *); /* Append a string fast and safely */

/* indexing module functions. */
int32_t AddGlobalSymbols(nameschain *); /* Add locally defined symbols to glvoid InsertBlock(lkedfitem *block,lkedfitem *insertpt)obal symbol list */
hashchain *CreateGlobalSymbol(char *,int32_t,int32_t); /* Create global symbol entry */
#ifdef MURMUR3
uint32_t Murmur3(char *); /* Murmur3 hash function */
#else
uint32_t KRMultHash(char *); /* Kernighan and Ritchie hash function */
#endif
hashchain *CheckGlblDeclared(char *); /* Check if a name is globally declared */
int32_t EditDiscrTrees(cmprefix *,int32_t,int32_t); /* Edit discrimination trees for clause */
int32_t DsTermIndexing(uint8_t *,uint8_t **); /* Add item to a discrimination tree */
int32_t DeleteBranch(uint8_t *,uint8_t **); /* Delete item in a discrimination tree */
void FreeDiscrBranch(uint8_t *); /* Free discrimination tree branch memory */
int32_t QueryGeneric(uint8_t *,uint8_t **,dscstack **,
		int32_t,uint8_t *); /* Get next generalization candidate item */
int32_t QueryUI(uint8_t *,uint8_t **,dscstack **,int32_t,
		uint8_t *,int32_t); /* Get next unification or instance candidate item */
int32_t Unlink(cmprefix *,int32_t); /* Unlink binary clause from KB */
void FreeTreeMemory(int32_t); /* Free KB tree memory */
uint32_t HashClause(uint8_t *); /* Calculate 32 bits hash value for a clause key */

/* infere module functions. */
uint8_t *NextItem(uint8_t *,int32_t); /* Jump to next item in clause */
int32_t IsEqual(uint8_t *,uint8_t *,int32_t); /* Check if two items are identical */
int32_t IsEqual2(uint8_t *,uint8_t *); /* Check if two items in different clauses are identical */
int32_t Unify(uint8_t *,uint8_t *,subst *,int32_t); /* Perform literal or term unification */
int32_t UnifyVar(uint8_t *,uint8_t *,subst *,int32_t); /* Unify a variable with a term */
int32_t Generalize(uint8_t *,uint8_t *,subst *,int32_t); /* Perform literal, term or list generalization */
int32_t GeneralizeVar(uint8_t *,uint8_t *,subst *,int32_t); /* Generalize a variable with a term */
int32_t IsCyclicSubst(subst *); /* Cyclic substitution check */
int32_t CheckClosedLoop(uint8_t *,uint8_t *,int16_t *,int32_t *,int16_t *,
		int32_t *); /* Check if a variable is in a closed loop */
uint32_t CheckVarInTerm(uint8_t *,uint8_t *,subst *,
		int32_t); /* Check if a variable occurs in a term with substitution */
uint32_t CheckVarInTerm2(uint8_t *,uint8_t *); /* Check if a variable occurs in a term */
int32_t CopyClause(cmprefix *,int32_t *,cmprefix *); /* Copy a clause */
void DeleteLiteral(binprefix *,uint8_t *); /* Deletes a literal in a clause */
uint8_t *GetLiteralPtr(cmprefix *,int32_t); /* Gets a pointer to literal given its order number */
int32_t GetLiteralNbr(binprefix *,uint8_t *); /* Gets a literal number given a literal pointer */
int32_t ApplySubst2Clause(cmprefix *,int32_t *,subst *,int32_t); /* Apply substitution to a clause */
void GetSubstTermData(subst *,uint8_t **,int32_t); /* Initialize set of pointers to terms in a substitution */
int32_t ApplySubst2Item(uint8_t **,binprefix **,int32_t *,uint8_t *,
		int32_t,uint8_t **); /* Apply substitution to an item */
void UpdateParams(binprefix *); /* Update binary clause parameters */
void UpdateParams2(uint8_t *); /* Update binary formula parameters */
uint8_t *UpdateSymbolParms(binprefix *,uint8_t *,int32_t,int32_t,
		int32_t); /* Update binprefix and symbol parameters in an item */
uint8_t *UpdateSymbolParms2(uint8_t *,uint8_t *); /* Update symbol parameters in an item */
void UpdateClauseSymbolWeights(binprefix *); /* Update input clause symbolw field */
void Infere(cmprefix *,int32_t); /* Perform saturation inferences */
void Resolution(cmprefix *,int32_t); /* Perform resolution inference */
int32_t BuildResolvent(cmprefix *,cmprefix *,uint8_t *,uint8_t *,
		cmprefix **,subst *,int32_t,int32_t); /* Build clause inferred by resolution */
void PM2Clause(cmprefix *,int32_t); /* Perform Paramodulation to a clause */
void PMFromClause(cmprefix *,int32_t); /* Perform Paramodulation from a clause */
int32_t BuildPMClause(cmprefix *,cmprefix *,uint8_t *,uint8_t *,uint8_t *,uint8_t *,
		cmprefix **,subst *); /* Build clause inferred by paramodulation */
void EqualityResolution(cmprefix *,int32_t); /* Perform equality resolution inference */
void EqualityFactoring(cmprefix *,int32_t); /* Perform equality factoring inference */
void Factor(cmprefix *,int32_t); /* Infer factors of a clause */

/* simplify module functions. */
int32_t IsTautology(cmprefix *); /* Check if clause is a tautology. */
int32_t RemoveDuplicates(cmprefix *,cmprefix **,int32_t,int32_t); /* Remove literal duplicates from clause */
int32_t TrivEqResolution(cmprefix *,cmprefix **); /* Trivial equality resolution inference */
int32_t DestrEqResolution(cmprefix *,cmprefix **); /* Destructive equality resolution inference */
int32_t EquivalentFactor(cmprefix *,cmprefix **,int32_t); /* Get an equivalent factor of a clause */
int32_t CheckFwSubsuming(cmprefix *,int32_t,int32_t); /* Check if clause is subsumed by a KB */
int32_t BckSubsuming(cmprefix *,int32_t,int32_t); /* Perform backward subsumption inference */
int32_t IsSubsumedByClause(uint8_t *,int32_t,binprefix *,
		subst *); /* Check if clause is subsumed by another clause */
int32_t IsSubsumedByClause2(binprefix *,binprefix *,
		subst *); /* IsSubsumedByClause() version for subsumption resolution */
int32_t LiteralSubsuming(uint8_t *,uint8_t *,int32_t,subsumdata *,subst *,
		struct timeval *); /* Check literal subsuming */
int32_t FwdSimplify(cmprefix **,int32_t,int32_t); /* Forward and backward simplifications */
int32_t FirstSimplify(cmprefix **,int32_t,int32_t); /* First clause simplification */
int32_t SecondSimplify(cmprefix **,int32_t,int32_t); /* Forward clause simplification */
int32_t BckSimplify(cmprefix *,int32_t,int32_t); /* Backward simplification */
int32_t ResolFactor(binprefix *,uint8_t *,binprefix **,uint8_t **,
		subst *,int32_t); /* Factoring for resolution inferences */
int32_t FwdDemodulation(cmprefix **,int32_t,int32_t); /* Perform forward demodulation */
int32_t BckDemodulation(cmprefix *,int32_t,int32_t); /* Perform backward demodulation */
int32_t Demodulate(cmprefix *,uint8_t *,binprefix *,uint8_t *,
		cmprefix **,subst *,int32_t); /* Demodulate a clause */
int32_t FwSubsResltn(cmprefix *,cmprefix **,int32_t,int32_t); /* Perform forward subsumption resolution */
int32_t BkSubsResltn(cmprefix *,int32_t,int32_t); /* Perform backward subsumption resolution */
int32_t SubsResltn(cmprefix *,uint8_t *,binprefix *,uint8_t *,
		cmprefix **); /* Perform subsumtion resolution inference */

/* ordering module functions. */
int32_t CompareLiterals(uint8_t *,uint8_t *,coeff *,coeff *,int32_t,int32_t); /* Literals comparison */
int32_t CompareEqualities(uint8_t *,uint8_t *,coeff *,coeff *,int32_t,int32_t); /* Equality literals comparison */
int32_t ComparePredicates(uint8_t *,uint8_t *,coeff *,coeff *,int32_t,int32_t); /* Predicates comparison */
int32_t CompareItems(uint8_t *,uint8_t *,coeff *,coeff *,int32_t,int32_t); /* Terms and predicates comparison */
int32_t CompareEquTerms(binprefix *,uint8_t *,uint8_t *,equpnode **,
		equpnode **,int32_t,int32_t); /* Equality root terms comparison */
int32_t CompSubstEquTrms(binprefix *,uint8_t *,uint8_t *,equpnode **,equpnode **,subst *,
		int32_t,int32_t,int32_t); /* Equality root terms comparison with substitution */
int32_t CompareSubstItems(uint8_t *,uint8_t *,int32_t,subst *,int32_t,
		int32_t,int32_t); /* Terms comparison with substitution */
int32_t SecondKBOCheck(uint8_t *,uint8_t *,int32_t,int32_t); /* Perform second KBO check */
int32_t SecondKBOCheck2(binprefix *,uint8_t *,uint8_t *,equpnode *,
		equpnode *,int32_t,int32_t); /* Perform second KBO check for polynomial trees */
int32_t SecondKBOCheck3(binprefix *,uint8_t *,uint8_t *,equpnode *,
		equpnode *,subst *,int32_t,int32_t,int32_t); /* Perform second KBO check with substitution */
int32_t CompareLexic(binprefix *,uint8_t *,uint8_t *,equpnode *,equpnode *,subst *,int32_t,
		int32_t,int32_t); /* Perform lexicographical comparison with substitution */
int32_t CompareItemWeight(uint8_t *,uint8_t *,coeff *,coeff *,int32_t); /* Terms and predicates weight comparison */
int32_t CompareItemWeight2(coeffcpt *,coeffcpt *); /* Equality subterms weight comparison */
int32_t CompItmWeightSubst(equpnode *,equpnode *,subst *,int32_t,
		int32_t);/* Equality subterms weight comparison with substitution. */
int32_t ApplySubst2Pol(lkedcoef **,subst *,int32_t,int32_t); /* Apply substitution to polynomial. */
int32_t ApplySbst2TwoItm(uint8_t *,uint8_t *,uint8_t **,uint8_t **,cmprefix **,
		subst *,int32_t,int32_t); /* Apply substitution to two terms */
int32_t CalcCoeffs(uint8_t *,coeff *,int32_t,int32_t); /* Calculate weight polynomial coefficients */
int32_t SelectItems(cmprefix *,int32_t); /* Select items in a clause */
int32_t CheckEqRootOrder(binprefix *,uint8_t *,uint8_t *,uint8_t *,subst *,int32_t,
		int32_t,int32_t); /* Check ordering constraints in equality root terms */
uint8_t *UpdateOffsets(binprefix *,uint8_t *); /* Update item offsets */
int32_t SetEquData(binprefix *,uint8_t *,uint8_t *,uint8_t *,
		int32_t); /* Initialize equality data for ordering constraints. */
int32_t SetEquChildNode(binprefix *,uint8_t *,equpnode **,
		int32_t); /* Initialize and set equality tree node for ordering constraints. */
int32_t SetEqTrmPol(binprefix *,uint8_t *,coeffcpt **,int32_t); /* Set equality subterm compact polynomial. */
int32_t SetLftRhtPols(binprefix *,coeffcpt **,coeffcpt **); /* Convert compact polynomials to left and right format. */
void FreeEquData(binprefix *); /* Free equality ordering data storage allocated for a compiled clause */
void FreeEquNode(uint8_t *,equpnode *); /* Free equality ordering node and its data. */
void FreeEquNodeData(uint8_t *,equpnode *); /* Free data in an equality ordering node. */

/* multith module functions. */
void Saturate(void); /* Start saturation algorithms */
void SetMTParams(int32_t,int32_t); /* Set strategy specific KB parameters */
void InferenceThread(void); /* Main inference thread process */
void Otter(int32_t); /* Otter saturation algorithm */
void Discount(int32_t); /* Discount saturation algorithm */
cmprefix *SelectClause(int32_t); /* Select current clause from passive clauses */
int32_t Retained(cmprefix *,int32_t); /* Check if clause is retained */
void ClauseWeight(cmprefix *,int32_t); /* Compute clause weight */
int32_t LRSClean(int32_t,int32_t); /* Delete unreachable passive clauses */
int32_t SetClSelectOrder(void); /* Set select order for clause selection queues */
int32_t Preprocess(int32_t); /* Perform problem pre-processing */
uint64_t GlobalPreprocess(void); /* Perform problem global pre-processing */
void GoType(void); /* Print problem type flags */
void Clausify(void); /* Convert problem formulas to CNF and print them in TPTP format */
int32_t InitStrategies(void); /* Select strategies for current problem */
int32_t GetPbmTypeNumber(void); /* Calculate problem type number */
uint16_t StratNb2StratFlags(int32_t); /* Convert a strategy number to strategy flags */

/* filemgmt module functions. */
char *Decompile(binprefix *); /* Translate a binary clause formula to text format */
char *Decompile2(uint8_t *); /* Translate a binary hashchcmp formula to text format */
char *DecompileSat(satclause *); /* Translate a SAT clause formula to text format */
char *DecompileAsserts(uint8_t *,int32_t); /* Translate an A-clause assertions to text format */
int32_t AddTermText(uint8_t *,char **,int32_t *,int32_t *); /* Translate a binary term to text format */
#ifdef DEBUGCODE
void BuildTxtSubstData(uint8_t *,uint8_t *,subst *,int32_t); /* Build substitution data in text form */
#endif
void PrintStats(int32_t); /* Display statistics */
void PrintStatData(stats *,int32_t); /* Display statistics data */
void PrintThreadSettings(int32_t); /* Print thread option settings */
int32_t BuildProof(void); /* Print theorem proof */
int32_t AddSatCl2Proof(void); /* Add SAT clauses to proof queue */
int32_t AddAclause2Proof(proofnode *); /* Add A-clause to proof queue */
void LinkProofNode(proofnode *,proofnode *); /* Link node to proof queue */
int32_t PrintProof(void); /* Print theorem proof */
int32_t PrintModel(void); /* Print model from a GOSAT run */
int32_t Import(char *,char *,int32_t,int32_t); /* Import a TPTP problem file */
int32_t ReadLine(char **,int32_t *,int32_t,FILE *); /* Read a line from a text file */
int32_t Cnf2Fof(char **); /* Convert a CNF formula to FOF */
int32_t FileExist(char *); /* Check if a file exist */
void PrintKBData(int32_t); /* Print current data of a knowledge base */
int32_t PrintAClauses(int32_t); /* Print A-clauses of a knowledge base */
int32_t PrintSingleAClause(cmprefix *); /* Print a single A-clause */
int32_t PrintCompSymbols(int32_t); /* Print the component symbols of a knowledge base */
int32_t PrintSATClauses(int32_t); /* Print SAT clauses of a knowledge base */
int32_t PrintSingleSATClause(satclause *); /* Print a single SAT clause */
int32_t LinkedList2Text(lkedfitem *,char **,int32_t *,int32_t); /* Convert formula in linked list format to text */
int32_t Equ2Text(lkedfitem **,char **,int32_t *,int32_t *); /* Convert (in)equality in linked list formula to text */
int32_t PredOrFunc2Text(lkedfitem **,char **,int32_t *,
		int32_t *); /* Convert predicate or function in linked list formula to text */
int32_t Term2Text(lkedfitem **,char **,int32_t *,int32_t *); /* Convert term in linked list formula to text */
int32_t Quantifier2Text(lkedfitem **,char **,int32_t *,int32_t *); /* Convert a quantifier in linked list formula to text */
int32_t PrintClausifiedFormulas(void); /* Print main KB formulas in TPTP CNF format */
char *ParseFilename(char *,char **); /* Parse and build a filename */
void Help(char*); /* Display general help or a help topic */

/* satsolver module functions. */
int32_t SplitClause(cmprefix *,int32_t,int32_t); /* Get maximal split of an A-clause */
int32_t AddAssertions(binprefix *,binprefix *,binprefix *); /* Add assertions from two clauses to a third clause */
int32_t GetLocksAsserts(binprefix *,binprefix *,binprefix *); /* Get simplification assertions and locking assertions */
void ChainAsserts(uint8_t *); /* Chain assertions to appropriate queue */
void ChainLocks(uint8_t *); /* Chain locking assertions to appropriate queue */
void UnchainAsserts(uint8_t *); /* Unchain assertions */
void UnchainLocks(uint8_t *); /* Unchain locking assertions */
int32_t GetSplitSymbol(uint8_t *,int32_t,int32_t,hashchcmp **,
		int32_t,int32_t); /* Get a P-predicate symbol for a clause */
int32_t IsVariant(uint8_t *,int32_t,hashchcmp *); /* Check if clause is a variant of another clause */
int32_t LiteralVariant(uint8_t *,hashchcmp *,int8_t *,int32_t,subst *,struct timeval *); /* Literal variant check */
int32_t ItemVariant(uint8_t *,uint8_t *,subst *); /* Item variant check */
int32_t SatQ2SatSolver(int32_t); /* Send SAT queue to SAT solver */
int32_t SatSolver(int32_t); /* SAT solver algorithm */
void ChainOrdComponent(hashchcmp *,int32_t); /* Chain a hashchcmp element to the ordered sub-queue */
void UnchainOrdComponent(hashchcmp *,int32_t); /* Remove a hashchcmp element from the ordered sub-queue */
void ChainSatClause(satclause *,int32_t); /* Chain a SAT solver clause to ordered sub-queue */
void UnchainSatClause(satclause *,int32_t); /* Remove a satclause element from the ordered sub-queue */
int32_t DoSymbolAssgnmnt(satnode *,satclause **,int32_t); /* Perform symbol assignment tasks */
void UndoSymbolAssgnmnt(satnode *,int32_t); /* Undo symbol assignment */
void GoSat(void); /* Call SAT solver without first order prover */
int32_t ApplyModelChanges(int32_t); /* Apply changes of current interpretation */
int32_t RemoveLocks(hashchcmp *,int32_t); /* Remove locks unlocked by model change */
int32_t DoLocks(hashchcmp *,int32_t); /* Do new locks produced by model change */
int32_t AddCompClause(hashchcmp *,int32_t); /* Add non existing component clauses */
int32_t BuildContrClause(cmprefix *,int32_t); /* Add contradiction clause to SAT queue */
int32_t SatSimplify(satclause **,int32_t); /* Check tautologies and remove duplicates in SAT clause */
int32_t SatSubsumption(satclause *,int32_t); /* Perform by and to Sat clause subsumptions */
int32_t IsSatClSubsumed(satclause *,satclause *,uint8_t *); /* Check if a Sat clause subsumes another Sat clause */
int32_t Backtrack(satclause *,satnode **,int32_t); /* Backtrack from a contradiction */
#ifdef VERIFYSATPROOF
void VerifySatProof(void); /* Verify SAT part of the proof */
#endif
#if VERIFYSATFINALSTATUS == 1
void VerifySatStatus(void); /* Verify SAT final status after a problem solving run */
#endif

/* satsolver module functions. */
int32_t ProcessLearnData(char *,char *,int32_t); /* Build learn data from output learn file */
double *DCDescentMethod(learnvector *,uint8_t *,int32_t); /* Dual Coordinate Descent Method */
double *GetClauseVector(cmprefix *); /* Build a clause vector */
int32_t VerticalFeatures(cmprefix *,double *); /* Compute vertical features of a clause vector */
void HorizontalFeatures(cmprefix *,double *); /* Compute horizontal features of a clause vector */
void StaticFeatures(cmprefix *clause,double *); /* Compute static features of a clause vector */
void LearnFromProblem(); /* Extract learning information from solved problem */
int32_t ReadLearnRecords(void); /* Read learn records generated by PROCLEARN command */

/* parsing module functions. */
void CommandProc(void); /* Command processor */
void ParseArgs(int,char **); /* Parse program arguments */
int32_t ChkConvertOption(char *); /* Check and convert option to interactive command */
int32_t ParseCommand(char *,char **); /* Command parser */
#endif /* GLOBAL_H_ */
