/*
 ============================================================================
 Name        : simplify.c
 Author      : Oscar Contreras
 Version     :
 Copyright   : Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
 Description : First order logic theorem prover in C
 ============================================================================
 */

/****************************************************************
*
*               simplify (Simplification inference functions module)
*
* All of the documentation and software included in the Drodi package
* is copyrighted by Oscar Contreras.
*
* Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
*
*
* This software is provided "as-is," without any express or implied warranty.
* In no event shall the author be held liable for any damages arising from the
* use, or the result of use, of this software and its documentation in terms
* of correctness, accuracy, reliability, currentness, or otherwise.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial use, and to alter and redistribute it, provided
* that the following conditions are met:
*
* 1. All redistributions of source code files must retain all copyright
*    notices that are currently in place, and this list of conditions without
*    modification.
*
* 2. All redistributions in binary form must retain all occurrences of the
*    above copyright notice that are currently in place (for example, in
*    the About boxes).
*
* 3. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software.
*
* 4. Modified versions in source or binary form must be plainly marked as
*    such, and must not be misrepresented as being the original software.
*
*
* Oscar Contreras
* rwtrainer10 AT gmail.com
*
****************************************************************/

/*-------------------------------------------------------------
 *
 *  DESCRIPTION:
 *
 *    This source contains the simplification inferences of the Drodi package.
 *
 *    Drodi is a multi-platform Open Source Artificial Intelligence application.
 *    See Drodi.c for additional information.
 *
 *
 *  IMPLEMENTATION:
 *
 *    Multi-platform
 *
 *  NOTES:
 *
 *
 *--------------------------------------------------------------*/

/* Include for all C sources */
#include "global.h"

/** Global variables for this module: only those that need initialization in their definition. */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check if a clause is a tautology)  OCJ
 *
 *    This function checks if a clause in internal compiled format
 *    is a tautology.
 *
 *    Two checks are performed to check for a tautology:
 *
 *    The first check is positive if the clause contains a literal
 *    and its negation, that is, for clauses of the form "C|P|~P"
 *    where C is a sub clause and P is an atom. The literals must
 *    be one exactly the negation of the other, not just unify with
 *    each other. For instance, consider the following clause:
 *      A(B,x)|!A(y,C)
 *    where x and y are universally quantified variables. This
 *    clause is not a tautology because let A(B,D) be false and
 *    A(E,C) be true. Then the clause is false, so it is not
 *    necessarily true.
 *    Another example, the clause:
 *      A(x)|!A(B)
 *    where x is an universally quantified variable. This is not
 *    a tautology because let A(C) be false and A(B) be true.
 *    Then the clause is false.
 *
 *    The second check is for equational tautologies, that is,
 *    for clauses of the form C|t=t or C|"a"="a" or C|"a"!="b" where C
 *    is a sub clause, t is a term and "a" and "b" are "distinct objects.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: pointer to clause in binary format.
 *
 *  RETURNS:
 *
 *    0 -> Clause is not a tautology.
 *    1 -> Clause is a tautology.
 *
 *
 *--------------------------------------------------------------*/
int32_t IsTautology(cmprefix *clause){
	auto hashchain *rtterm1s,*rtterm2s;                  /* Pointers to root term symbols in equalities */
	auto uint8_t *ptr1,*ptr2,*ptr3;                      /* Auxiliary pointers */

	/* Loop through literals in clause to detect equalities */
	/* of identical terms or "distinct objects" tautologies. */
	if ((UNITEQUALITY|NONUNITEQUALITY)&pbmflags){
		for (ptr1=&clause->part2.bin->formula[0];(*ptr1)!=UNITEND;
				ptr1=NextItem(ptr1,OVERSUBTERMS)){
			if (EQUALITY&ptr1[0]){
				ptr2=NextItem(ptr1,IMMED);
				ptr3=NextItem(ptr2,OVERSUBTERMS);
				if (0==(NEGATED&ptr1[0])){
					if (IsEqual(ptr2,ptr3,0)){
						return(1);
					}
				} else {
					if ((FUNCTION&ptr2[0])&&(FUNCTION&ptr3[0])){
						rtterm1s=((symbol *)&ptr2[1])->symbol;
						rtterm2s=((symbol *)&ptr3[1])->symbol;
						if ((rtterm1s!=rtterm2s)&&(DISTINCTOBJECT&rtterm1s->type)
								&&(DISTINCTOBJECT&rtterm2s->type)){
							return(1);
						}
					}
				}
			}
		}
	}

	/* Loop through pairs of literals in clause. */
	for (ptr1=&clause->part2.bin->formula[0];(*ptr1)!=UNITEND;
			ptr1=NextItem(ptr1,OVERSUBTERMS)){
		for (ptr2=NextItem(ptr1,OVERSUBTERMS);(*ptr2)!=UNITEND;
				ptr2=NextItem(ptr2,OVERSUBTERMS)){
			if (IsEqual(ptr1,ptr2,NEGATED)){
				return(1);
			}
		}
	}

	return(0);
} /* IsTautology */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Remove literal duplicates from clause)  OCJ
 *
 *    This function removes literal duplicates from a clause in
 *    internal compiled binary format. The simplified clause
 *    keeps the original assertions.
 *
 *    If the clause is successfully simplified then it is placed in
 *    new allocated storage and flagged as PASSIVE. If buildtxt
 *    parameter is not zero then the original clause is converted
 *    to text form and also placed in a new allocated storage that
 *    is linked to the original cmprefix structure. No other modifications
 *    are done to the cmprefix structure. The original binprefix storage
 *    is freed.
 *
 *    It is the caller responsibility to link and/or unlink the clauses
 *    and set the other prefix fields.
 *
 *    See EquivalentFactor() function for comments regarding program
 *    performance.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause to be simplified
 *            by factoring.
 *    pnewcl: Address of pointer to cmprefix structure of new clause
 *            to be allocated if there are literal removals.
 *    nkb: KB index of KB that owns the process.
 *    buildtxt: If not zero and a simplification results then the
 *              original clause is converted to text form and placed
 *              in a new allocated storage that is linked to the original
 *              cmprefix structure.
 *
 *  RETURNS:
 *
 *    0 -> process was OK and no duplicate literals were found.
 *    1 -> Some duplicate literals were removed.
 *    NOMEMORY -> not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t RemoveDuplicates(cmprefix *clause,cmprefix **pnewcl,int32_t nkb,int32_t buildtxt){
	auto cmprefix *auxcl;                               /* Pointer to auxiliary clause */
	auto int32_t size;                                  /* Size of auxiliary clause binary buffer in bytes */
	auto uint8_t *ptr1,*ptr2;                           /* Auxiliary pointers */
	auto char *ptr10;                                   /* Buffer for clause in text form. */
	auto binprefix *ptr4;                               /* Auxiliary pointer */
	auto int32_t ii,pp;                                 /* Auxiliary */

	#ifdef DEBUGCODE
	/* Debug. */
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Allocate storage for auxiliary clause and copy */
	/* clause into the first auxiliary clause. */
	size=clause->part2.bin->size+binpsize;
	if ((auxcl=mallocmn(sizeof(cmprefix)))==NULL){
		return(NOMEMORY);
	}
	if ((auxcl->part2.bin=mallocmn(size))==NULL){
		MYFREE(auxcl);
		return(NOMEMORY);
	}
	memcpy(auxcl->part2.bin,clause->part2.bin,size);
	auxcl->part2.bin->clause=auxcl;

	/* Loop through literals in auxiliary clause. */
	for (ptr1=&auxcl->part2.bin->formula[0],pp=0;(*ptr1)!=UNITEND;
			ptr1=NextItem(ptr1,OVERSUBTERMS)){

		/* Loop through pairs of literals in auxiliary clause. */
		for (ptr2=NextItem(ptr1,OVERSUBTERMS);(*ptr2)!=UNITEND;
				ptr2=(ii?NextItem(ptr2,OVERSUBTERMS):ptr2)){
			ii=1;

			/* Identical literals. Delete the ptr2 literal. */
			if (IsEqual(ptr1,ptr2,0)){
					DeleteLiteral(auxcl->part2.bin,ptr2);
					pp=1; /* Indicate there was a removal. */
					ii=0; /* Keep in the same literal position in next iteration for ptr2. */
			}
		}
	}

	/* There is a removal. */
	if (pp){

		/* Reallocate simplification result to exact size */
		/* and set clause pointer. The original assertions */
		/* are kept. but their asymbol's must be adjusted */
		/* to the new clause. */
		if (NULL==(ptr4=reallocmn(auxcl->part2.bin,auxcl->part2.bin->size+binpsize))){
			MYFREE(auxcl->part2.bin);
			MYFREE(auxcl);
			return(NOMEMORY);
		}
		auxcl->part2.bin=ptr4;
		auxcl->part2.bin->equflg=0;
		auxcl->inference=REMOVEDUPS;
		auxcl->parent1=clause;
		auxcl->parent2=NULL;
		auxcl->flags=UNPROC;
		if (auxcl->part2.bin->asserts!=NULL){
			for (ptr1=&auxcl->part2.bin->asserts[0];ptr1[0]!=UNITEND;ptr1+=(1+sizeof(asymbol))){
				((asymbol *)&ptr1[1])->part2.binp=auxcl->part2.bin;
			}
		}
		*pnewcl=auxcl;

		/* Convert original clause to text form if needed. */
		if (buildtxt){
			if (NULL==(ptr10=Decompile(clause->part2.bin))){
				MYFREE(auxcl->part2.bin);
				MYFREE(auxcl);
				return(NOMEMORY);
			}
			MYFREE(clause->part2.bin);
			clause->part2.text=ptr10;
		}

	/* If there was not a removal then free auxiliary clause storage. */
	} else {
		MYFREE(auxcl->part2.bin);
		MYFREE(auxcl);
	}

	return(pp);
} /* RemoveDuplicates */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Trivial equality resolution inference)  OCJ
 *
 *    This function performs trivial equality resolution inference
 *    simplification of a clause in internal compiled binary format.
 *    The simplified clause keeps the original assertions.
 *
 *    If the clause is successfully simplified then it is placed in
 *    new allocated storage and flagged as PASSIVE. The original clause
 *    is converted to text form and also placed in a new allocated
 *    storage that is linked to the original cmprefix structure.
 *    No other modifications are done to the cmprefix structure.
 *    The original binprefix storage is freed.
 *
 *    It is the caller responsibility to link and/or unlink the clauses
 *    and set the other prefix fields.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause to be simplified
 *            by equivalent factoring.
 *    pnewcl: Address of pointer to cmprefix structure of new clause
 *            to be allocated if a simplification is performed.
 *
 *  RETURNS:
 *
 *    0 -> process was OK and no simplification was performed.
 *    1 -> Clause has been successfully simplified.
 *    NOMEMORY -> not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t TrivEqResolution(cmprefix *clause,cmprefix **pnewcl){
	auto cmprefix *auxcl;                               /* Pointer to auxiliary clause */
	auto int32_t size;                                  /* Size of auxiliary clause binary buffer in bytes */
	auto hashchain *rtterm1s,*rtterm2s;                 /* Pointers to root term symbols in equalities */
	auto uint8_t *ptr1,*ptr2,*ptr3;                     /* Auxiliary pointers */
	auto char *ptr10;                                   /* Buffer for clause in text form. */
	auto binprefix *ptr4;                               /* Auxiliary pointer */
	auto int32_t ii,pp;                                 /* Auxiliary */

	#ifdef DEBUGCODE
	/* Debug. */
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Allocate storage for auxiliary clause and copy */
	/* clause into the first auxiliary clause. */
	size=clause->part2.bin->size+binpsize;
	if ((auxcl=mallocmn(sizeof(cmprefix)))==NULL){
		return(NOMEMORY);
	}
	if ((auxcl->part2.bin=mallocmn(size))==NULL){
		MYFREE(auxcl);
		return(NOMEMORY);
	}
	memcpy(auxcl->part2.bin,clause->part2.bin,size);
	auxcl->part2.bin->clause=auxcl;

	/* Loop through literals in auxiliary clause. */
	for (ptr1=&auxcl->part2.bin->formula[0],pp=0;(*ptr1)!=UNITEND;
			ptr1=(ii?NextItem(ptr1,OVERSUBTERMS):ptr1)){
		ii=1;

		/* Literal is an equality. */
		if (EQUALITY&ptr1[0]){
			ptr2=NextItem(ptr1,IMMED);
			ptr3=NextItem(ptr2,OVERSUBTERMS);

			/* Negated equality and root terms are equal. */
			if (NEGATED&ptr1[0]){
				if (IsEqual(ptr2,ptr3,0)){
					DeleteLiteral(auxcl->part2.bin,ptr1);
					pp=1; /* Indicate there was a simplification. */
					ii=0; /* Keep in the same literal position in next iteration for ptr2. */
				}

			/* Positive equality and root terms are unequal "distinct objects". */
			} else if ((FUNCTION&ptr2[0])&&(FUNCTION&ptr3[0])){
				rtterm1s=((symbol *)&ptr2[1])->symbol;
				rtterm2s=((symbol *)&ptr3[1])->symbol;
				if ((rtterm1s!=rtterm2s)&&(DISTINCTOBJECT&rtterm1s->type)
									&&(DISTINCTOBJECT&rtterm2s->type)){
					DeleteLiteral(auxcl->part2.bin,ptr1);
					pp=1; /* Indicate there was a simplification. */
					ii=0; /* Keep in the same literal position in next iteration for ptr2. */
				}
			}
		}
	}

	/* If there is a simplification then set clause pointer */
	/* and convert original clause to text form. The original */
	/* assertions are kept but their asymbol's must be */
	/* adjusted to the new clause. */
	if (pp){
		if (NULL==(ptr4=reallocmn(auxcl->part2.bin,auxcl->part2.bin->size+binpsize))){
			MYFREE(auxcl->part2.bin);
			MYFREE(auxcl);
			return(NOMEMORY);
		}
		auxcl->part2.bin=ptr4;
		if (NULL==(ptr10=Decompile(clause->part2.bin))){
			MYFREE(auxcl->part2.bin);
			MYFREE(auxcl);
			return(NOMEMORY);
		}
		auxcl->inference=TRIVEQRES;
		auxcl->parent1=clause;
		auxcl->parent2=NULL;
		auxcl->flags=UNPROC;
		auxcl->part2.bin->equflg=0;
		if (auxcl->part2.bin->asserts!=NULL){
			for (ptr1=&auxcl->part2.bin->asserts[0];ptr1[0]!=UNITEND;ptr1+=(1+sizeof(asymbol))){
				((asymbol *)&ptr1[1])->part2.binp=auxcl->part2.bin;
			}
		}
		*pnewcl=auxcl;
		FreeEquData(clause->part2.bin);
		MYFREE(clause->part2.bin);
		clause->part2.text=ptr10;

	/* If there was not a removal then free auxiliary clause storage. */
	} else {
		MYFREE(auxcl->part2.bin);
		MYFREE(auxcl);
	}

	return(pp);
} /* TrivEqResolution */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Destructive equality resolution inference)  OCJ
 *
 *    This function performs destructive equality resolution inference
 *    simplification of a clause in internal compiled binary format.
 *    The simplified clause keeps the original assertions.
 *
 *    If the clause is successfully simplified then it is placed in
 *    new allocated storage and flagged as PASSIVE. The original clause
 *    is converted to text form and also placed in a new allocated
 *    storage that is linked to the original cmprefix structure.
 *    No other modifications are done to the cmprefix structure.
 *    The original binprefix storage is freed.
 *
 *    It is the caller responsibility to link and/or unlink the clauses
 *    and set the other prefix fields.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause to be simplified
 *            by equivalent factoring.
 *    pnewcl: Address of pointer to cmprefix structure of new clause
 *            to be allocated if a simplification is performed.
 *
 *  RETURNS:
 *
 *    0 -> process was OK and no simplification was performed.
 *    1 -> Clause has been successfully simplified.
 *    NOMEMORY -> not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t DestrEqResolution(cmprefix *clause,cmprefix **pnewcl){
	auto cmprefix *auxcl;                               /* Pointer to auxiliary clause */
	auto int32_t size;                                  /* Size of auxiliary clause binary buffer in bytes */
	auto uint8_t *ptr1,*ptr2,*ptr5;                     /* Auxiliary pointers */
	auto char *ptr10;                                   /* Buffer for clause in text form. */
	auto binprefix *ptr4;                               /* Auxiliary pointer */
	auto subst Subst;                                   /* Auxiliary substitution */
	auto int32_t ii,jj,pp;                              /* Auxiliary */

	#ifdef DEBUGCODE
	/* Debug. */
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Allocate storage for auxiliary clause and copy */
	/* clause into the first auxiliary clause. */
	size=clause->part2.bin->size+binpsize;
	if ((auxcl=mallocmn(sizeof(cmprefix)))==NULL){
		return(NOMEMORY);
	}
	if ((auxcl->part2.bin=mallocmn(size))==NULL){
		MYFREE(auxcl);
		return(NOMEMORY);
	}
	memcpy(auxcl->part2.bin,clause->part2.bin,size);
	auxcl->part2.bin->clause=auxcl;

	/* Initialize substitution. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Loop through literals in auxiliary clause. */
	for (ptr1=&auxcl->part2.bin->formula[0],pp=jj=0;(*ptr1)!=UNITEND;
			ptr1=(ii?NextItem(ptr1,OVERSUBTERMS):ptr1),jj=(ii?jj+1:jj)){
		ii=1;

		/* Literal is a negated equality. */
		if ((NEGATED|EQUALITY)==((NEGATED|EQUALITY)&ptr1[0])){

			/* Empty substitution. */
			if (Subst.buffer!=NULL){
				Subst.substsize=1;
				Subst.buffer[0]=UNITEND;
			}

			/* Check if one of the root equality terms is a variable. */
			ptr2=NextItem(ptr1,IMMED);
			if (VARIABLE&ptr2[0]){
				ptr5=NextItem(ptr2,OVERSUBTERMS);
			} else {
				ptr2=NextItem(ptr5=ptr2,OVERSUBTERMS);
				if (0==(VARIABLE&ptr2[0])){
					ptr5=NULL;
				}
			}

			/* One of the root equality terms is a variable. */
			if (ptr5!=NULL){

				/* Check unification and build substitution. */
				switch (UnifyVar(ptr2,ptr5,&Subst,0)){

					/* Not enough memory. */
					case NOMEMORY:
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						if (auxcl->part2.bin->asserts!=NULL){
							MYFREE(auxcl->part2.bin->asserts);
						}
						MYFREE(auxcl->part2.bin);
						MYFREE(auxcl);
						return(NOMEMORY);
						break;

					/* Unification is possible and we have a substitution. */
					case 1:

						/* Delete equality. */
						DeleteLiteral(auxcl->part2.bin,ptr1);

						/* Apply substitution. */
						if (NOMEMORY==ApplySubst2Clause(auxcl,&size,&Subst,0)){
							if (Subst.buffer!=NULL){
								MYFREE(Subst.buffer);
							}
							if (auxcl->part2.bin->asserts!=NULL){
								MYFREE(auxcl->part2.bin->asserts);
							}
							MYFREE(auxcl->part2.bin);
							MYFREE(auxcl);
							return(NOMEMORY);
						}

						/* Update ptr1. This is necessary because auxiliary clause */
						/* binprefix may have been reallocated by ApplySubst2Clause() */
						/* and/or the positions of the literals may have changed by */
						/* substitution application. */
						ptr1=GetLiteralPtr(auxcl,jj);

						/* Indicate there was a simplification and keep in the */
						/* same literal position in next iteration for ptr2. */
						pp=1;
						ii=0;
						break;
				}
			}
		}
	}

	/* Free substitution storage. */
	if (Subst.buffer!=NULL){
		MYFREE(Subst.buffer);
	}

	/* There is a simplification. */
	if (pp){

		/* Set clause pointer and convert original clause to text form. */
		/* The original assertions are kept but their asymbol's must be */
		/* adjusted to the new clause. */
		if (NULL==(ptr4=reallocmn(auxcl->part2.bin,auxcl->part2.bin->size+binpsize))){
			MYFREE(auxcl->part2.bin);
			MYFREE(auxcl);
			return(NOMEMORY);
		}
		auxcl->part2.bin=ptr4;
		ptr4->equflg=0;
		if (NULL==(ptr10=Decompile(clause->part2.bin))){
			MYFREE(auxcl->part2.bin);
			MYFREE(auxcl);
			return(NOMEMORY);
		}
		auxcl->inference=DESTEQRES;
		auxcl->parent1=clause;
		auxcl->parent2=NULL;
		auxcl->flags=UNPROC;
		if (auxcl->part2.bin->asserts!=NULL){
			for (ptr1=&auxcl->part2.bin->asserts[0];ptr1[0]!=UNITEND;ptr1+=(1+sizeof(asymbol))){
				((asymbol *)&ptr1[1])->part2.binp=auxcl->part2.bin;
			}
		}
		*pnewcl=auxcl;
		MYFREE(clause->part2.bin);
		clause->part2.text=ptr10;

		/* Debug. */
		#ifdef DEBUGCODE
		ptr10=Decompile(auxcl->part2.bin);
		#ifdef VERBOSE
		if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
			printf("Destructive equality resolution simplification result [%d]:\n%s\n",kbset[0]->nformulas,ptr10);
		}
		#endif
		MYFREE(ptr10);
		#endif

	/* If there was not a removal then free auxiliary clause storage. */
	} else {
		MYFREE(auxcl->part2.bin);
		MYFREE(auxcl);
	}

	return(pp);
} /* DestrEqResolution */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Get an equivalent factor of a clause)  OCJ
 *
 *    This function factors a clause in internal compiled binary
 *    format. This is an alternative to removing literal duplicates.
 *    The simplified clause keeps the original assertions.
 *
 *    The advantage is that the resulting clause is more simple and
 *    is able to subsume more clauses than just removing literal
 *    duplicates. For instance, the clause A(B)|A(x) is not simplified
 *    at all with duplicates removal but equivalent factoring simplifies
 *    it to the equivalent clause A(B). Moreover, A(B) subsumes A(B)|C
 *    however A(B)|A(x) doesn't subsumes it with the algorithms
 *    currently implemented.
 *
 *    Factoring is only performed if an exactly equivalent clause
 *    is achieved. To be sure that an exactly equivalent clause is
 *    achieved the following conditions must be met:
 *      1.- The substitution obtained from the unification must be empty
 *      or the variables being substituted must all belong to only one
 *      of the literals being unified. That is, the unification must be
 *      a generalization.
 *      2.- The substitution cannot have cyclic atom substitutions. For
 *      instance, the literals in the clause ~less(X0,X1)|~less(X1,X0)
 *      each is a generalization of the other but none of them is an
 *      equivalent factor because the substitution {X0/X1, X1/X0}
 *      is cyclic.
 *      3.- The factored clause must subsume the non factored clause.
 *
 *    If the clause is successfully simplified then it is placed in
 *    new allocated storage and flagged as PASSIVE. The original clause
 *    is converted to text form and also placed in a new allocated
 *    storage that is linked to the original cmprefix structure.
 *    No other modifications are done to the cmprefix structure.
 *    The original assertions are kept in the factored clause.
 *    The original binprefix storage is freed.
 *
 *    It is the caller responsibility to link and/or unlink the clauses
 *    and set the other prefix fields.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause to be simplified
 *            by equivalent factoring.
 *    pnewcl: Address of pointer to cmprefix structure of new clause
 *            to be allocated if equivalent factoring is successful.
 *    nkb: KB index of KB that owns the process.
 *
 *  RETURNS:
 *
 *    0 -> There was no equivalent factor.
 *    1 -> Clause has been successfully factored and simplified.
 *    TIMEOUT -> Time has expired or other thread has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t EquivalentFactor(cmprefix *clause,cmprefix **pnewcl,int32_t nkb){
	auto cmprefix *auxcl1,*auxcl2,*auxcl3;              /* Pointers to auxiliary clauses */
	auto int32_t size1,size2,size3;                     /* Size of auxiliary clause binary buffers in bytes */
 	auto double time;                                   /* Elapsed time in seconds */
 	auto struct timeval timestr;                        /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr5,*ptr6;               /* Auxiliary pointers */
	auto char *ptr10;                                   /* Buffer for clause in text form. */
	auto binprefix *ptr4;                               /* Auxiliary pointer */
	auto subst Subst;                                   /* Substitution */
	auto int32_t lit1,lit2;                             /* Literal numbers */
	auto int32_t ii,jj,kk,mm,pp;                        /* Auxiliary */

	#ifdef DEBUGCODE
	/* Debug. */
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Allocate storage for auxiliary clauses and auxiliary factors */
	/* and copy clause into the first auxiliary clause. */
	size1=size2=size3=clause->part2.bin->size+binpsize;
	if ((auxcl1=mallocmn(sizeof(cmprefix)))==NULL){
		return(NOMEMORY);
	}
	if ((auxcl1->part2.bin=mallocmn(size1))==NULL){
		MYFREE(auxcl1);
		return(NOMEMORY);
	}
	if ((auxcl2=mallocmn(sizeof(cmprefix)))==NULL){
		MYFREE(auxcl1->part2.bin);
		MYFREE(auxcl1);
		return(NOMEMORY);
	}
	if ((auxcl2->part2.bin=mallocmn(size2))==NULL){
		MYFREE(auxcl1->part2.bin);
		MYFREE(auxcl1);
		MYFREE(auxcl2);
		return(NOMEMORY);
	}
	if ((auxcl3=mallocmn(sizeof(cmprefix)))==NULL){
		MYFREE(auxcl1->part2.bin);
		MYFREE(auxcl1);
		MYFREE(auxcl2->part2.bin);
		MYFREE(auxcl2);
		return(NOMEMORY);
	}
	if ((auxcl3->part2.bin=mallocmn(size3))==NULL){
		MYFREE(auxcl1->part2.bin);
		MYFREE(auxcl1);
		MYFREE(auxcl2->part2.bin);
		MYFREE(auxcl2);
		MYFREE(auxcl3);
		return(NOMEMORY);
	}
	memcpy(auxcl1->part2.bin,clause->part2.bin,size1);
	auxcl1->part2.bin->clause=auxcl1;

	/* Initialize substitution. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Loop through literals in first auxiliary clause. */
	for (ptr1=&auxcl1->part2.bin->formula[0],pp=lit1=0;(*ptr1)!=UNITEND;
			ptr1=(jj?NextItem(ptr1,OVERSUBTERMS):ptr1),lit1=(jj?lit1+1:lit1)){
		jj=1;

 		/* Check timeout and solution by other thread. */
 		if (nkb!=num_wkbs){
 			gettimeofday(&timestr,NULL);
 			time=timestr.tv_sec-kbset[nkb]->time.tv_sec
 					+(timestr.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0;
 			if ((time>=kbset[nkb]->opts.timeout)||(procctl.status==UNSATISFIABLE)
 					||(procctl.status==SATISFIABLE)){
				if (Subst.buffer!=NULL){
					MYFREE(Subst.buffer);
				}
				MYFREE(auxcl1->part2.bin);
				MYFREE(auxcl1);
				MYFREE(auxcl2->part2.bin);
				MYFREE(auxcl2);
				MYFREE(auxcl3->part2.bin);
				MYFREE(auxcl3);
 				return(TIMEOUT);
 			}
 			if ((procctl.status==NOMEMORY)||(procctl.status==TIMEOUT)){
				if (Subst.buffer!=NULL){
					MYFREE(Subst.buffer);
				}
				MYFREE(auxcl1->part2.bin);
				MYFREE(auxcl1);
				MYFREE(auxcl2->part2.bin);
				MYFREE(auxcl2);
				MYFREE(auxcl3->part2.bin);
				MYFREE(auxcl3);
				return(procctl.status);
			}
 		}

 		/* Loop through pairs of literals in first auxiliary clause. */
		for (ptr2=NextItem(ptr1,OVERSUBTERMS),lit2=lit1+1;(jj&&((*ptr2)!=UNITEND));
				ptr2=(kk&&jj?NextItem(ptr2,OVERSUBTERMS):ptr2),lit2=(kk&&jj?lit2+1:lit2)){
			kk=1;

 			#ifdef DEBUGCODE
			/* Debug. */
			ptr10=Decompile(auxcl1->part2.bin);
			MYFREE(ptr10);
			#endif

			/* Detection of literals generalization. */
			if (NOMEMORY==(ii=Generalize(ptr1,ptr2,&Subst,INITSUBST))){
				if (Subst.buffer!=NULL){
					MYFREE(Subst.buffer);
				}
				MYFREE(auxcl1->part2.bin);
				MYFREE(auxcl1);
				MYFREE(auxcl2->part2.bin);
				MYFREE(auxcl2);
				MYFREE(auxcl3->part2.bin);
				MYFREE(auxcl3);
				return(NOMEMORY);
			} else if (ii==0){
	 			if (NOMEMORY==(ii=Generalize(ptr2,ptr1,&Subst,INITSUBST))){
	 				if (Subst.buffer!=NULL){
	 					MYFREE(Subst.buffer);
	 				}
	 				MYFREE(auxcl1->part2.bin);
	 				MYFREE(auxcl1);
	 				MYFREE(auxcl2->part2.bin);
	 				MYFREE(auxcl2);
	 				MYFREE(auxcl3->part2.bin);
	 				MYFREE(auxcl3);
	 				return(NOMEMORY);
	 			}
	 			if (ii){
	 				ii=2;
	 			}
			}

			/* Valid generalization. */
			if (ii){

				/* Identical literals. Delete the ptr2 literal. */
				if ((Subst.buffer==NULL)||(Subst.substsize==1)){
					DeleteLiteral(auxcl1->part2.bin,ptr2);
					pp=1; /* Indicate there was a factor. */
					kk=0; /* Keep in the same literal position in next iteration for ptr2. */

				/* Literals are not identical. */
				} else {

					/* Copy first auxiliary clause to second auxiliary clause, */
					/* delete the more general literal and apply substitution. */
					CopyClause(auxcl2,&size2,auxcl1);
					switch (ii){
						case 1:
							DeleteLiteral(auxcl2->part2.bin,
									&auxcl2->part2.bin->formula[0]+
									(ptr1-&auxcl1->part2.bin->formula[0]));
							break;
						case 2:
							DeleteLiteral(auxcl2->part2.bin,
									&auxcl2->part2.bin->formula[0]+
									(ptr2-&auxcl1->part2.bin->formula[0]));
							break;
					}
					if (NOMEMORY==ApplySubst2Clause(auxcl2,&size2,&Subst,0)){
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						MYFREE(auxcl1->part2.bin);
						MYFREE(auxcl1);
						MYFREE(auxcl2->part2.bin);
						MYFREE(auxcl2);
						MYFREE(auxcl3->part2.bin);
						MYFREE(auxcl3);
						return(NOMEMORY);
					}

					/* Copy second auxiliary clause to third auxiliary clause and delete */
					/* duplicate literals that may have arisen by applying the substitution */
					/* above. This is necessary to ensure that IsSubsumedByClause() function */
					/* works properly (see comments in IsSubsumedByClause() function description). */
					CopyClause(auxcl3,&size3,auxcl2);
					for (ptr5=&auxcl3->part2.bin->formula[0];(*ptr5)!=UNITEND;
							ptr5=NextItem(ptr5,OVERSUBTERMS)){
						for (ptr6=NextItem(ptr5,OVERSUBTERMS);(*ptr6)!=UNITEND;
								ptr6=(mm?NextItem(ptr6,OVERSUBTERMS):ptr6)){
							if (IsEqual(ptr5,ptr6,0)){
								DeleteLiteral(auxcl3->part2.bin,ptr6);
								mm=0; /* Keep in the same literal position in next iteration for ptr6. */
							} else {
								mm=1;
							}
						}
					}

					/* If third auxiliary clause subsumes first auxiliary clause then */
					/* copy second auxiliary clause to first auxiliary clause and adjust */
					/* loop pointers. */
					switch (IsSubsumedByClause(&auxcl1->part2.bin->formula[0],
							auxcl1->part2.bin->literals,auxcl3->part2.bin,&Subst)){
						case NOMEMORY:
							if (Subst.buffer!=NULL){
								MYFREE(Subst.buffer);
							}
							MYFREE(auxcl1->part2.bin);
							MYFREE(auxcl1);
							MYFREE(auxcl2->part2.bin);
							MYFREE(auxcl2);
							MYFREE(auxcl3->part2.bin);
							MYFREE(auxcl3);
							return(NOMEMORY);
							break;
						case 1:
							CopyClause(auxcl1,&size1,auxcl2);
							pp=1; /* Indicate there was a factor. */

							/* Loop control and loop pointer adjustment. */
							ptr1=GetLiteralPtr(auxcl1,lit1);
							switch (ii){

								/* If ptr1 literal generalized ptr2 then terminate ptr2 loop and */
								/* keep in the same literal position in next iteration for ptr1. */
								case 1:
									jj=0;
									ptr2=NULL;
									break;

								/* If ptr2 literal generalized ptr1 then keep in the same literal */
								/* position in next iteration for ptr2. */
								case 2:
									kk=0;
									ptr2=GetLiteralPtr(auxcl1,lit2);
									break;
							}
							break;
					}
				}
			}
		}
	}

	/* Free second and third auxiliary clause storage. */
	if (Subst.buffer!=NULL){
		MYFREE(Subst.buffer);
	}
	MYFREE(auxcl2->part2.bin);
	MYFREE(auxcl2);
	MYFREE(auxcl3->part2.bin);
	MYFREE(auxcl3);

	/* There was an equivalent factor. Set clause pointer and */
	/* convert original clause to text form. The original */
	/* assertions are kept but their asymbol's must be */
	/* adjusted to the new clause. */
	if (pp){
		if (NULL==(ptr4=reallocmn(auxcl1->part2.bin,auxcl1->part2.bin->size+binpsize))){
			MYFREE(auxcl1->part2.bin);
			MYFREE(auxcl1);
			return(NOMEMORY);
		}
		auxcl1->part2.bin=ptr4;
		ptr4->equflg=0;
		if (NULL==(ptr10=Decompile(clause->part2.bin))){
			MYFREE(auxcl1->part2.bin);
			MYFREE(auxcl1);
			return(NOMEMORY);
		}
		auxcl1->inference=EQUIVFACT;
		auxcl1->parent1=clause;
		auxcl1->parent2=NULL;
		auxcl1->flags=UNPROC;
		if (auxcl1->part2.bin->asserts!=NULL){
			for (ptr1=&auxcl1->part2.bin->asserts[0];ptr1[0]!=UNITEND;ptr1+=(1+sizeof(asymbol))){
				((asymbol *)&ptr1[1])->part2.binp=auxcl1->part2.bin;
			}
		}
		*pnewcl=auxcl1;
		MYFREE(clause->part2.bin);
		clause->part2.text=ptr10;

 	/* If there was not an equivalent factor then free */
	/* first auxiliary clause storage. */
	} else {
		MYFREE(auxcl1->part2.bin);
		MYFREE(auxcl1);
	}

	return(pp);
} /* EquivalentFactor */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check if clause is subsumed by a KB)  OCJ
 *
 *    This function checks if the given clause in compiled or binary
 *    format is subsumed by some clause that exists in the indicated KB
 *    by calling IsSubsumedByClause() function for every candidate clause
 *    in the requested KB. It must also be specified the type of subsuming
 *    clauses that will be considered, either active or passive clauses.
 *
 *    Candidate clauses are selected by inspecting every literal in the
 *    argument clause and selecting clauses with a generalization literal.
 *    To prevent selection of the same candidate subsuming clauses by two
 *    different literals once a clause is selected it is signed with an
 *    8 bytes integer. If a clause is signed then it is not selected.
 *
 *    Once all candidate clauses corresponding to a given literal have
 *    been unsuccessfully tested then that literal is excluded from the
 *    IsSubsumedByClause() function call to speed the algorithm.
 *
 *    If the clause is subsumed then its locking assertions are set and
 *    the clause is locked if appropriate. However the clause is not
 *    deleted even if it is not locked. This must be done by the calling
 *    function.
 *
 *    See IsSubsumedByClause() function for details of clause subsumption
 *    condition.
 *
 *    This function must only be called for a working KB. It must NOT
 *    be called for main KB.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in compiled binary form.
 *    nkb: KB index of KB that will be searched for a subsuming clause.
 *         This must be a working KB. It must NOT be the main KB.
 *    type: PASSIVE if passive clauses are to be checked as subsuming
 *          clauses. Active clauses are always checked. Any other
 *          flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Clause is not subsumed by any clause of the indicated type
 *         in the indicated KB.
 *    1 -> Clause is subsumed by a clause of the indicated type in the
 *         indicated KB.
 *    TIMEOUT -> Time has expired or other thread has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t CheckFwSubsuming(cmprefix *clause,int32_t nkb,int32_t type){
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto dscstack *stack;                                /* Pointer to stack for sequential calls to QueryGeneric() */
	auto subst Subst;                                    /* Substitution for calls to IsSubsumedByClause() */
	auto uint64_t signature;                             /* KB signature. */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr3;                      /* Auxiliary pointers */
	auto int32_t ii,jj,kk,mm;                            /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Remember KB signature and update it. */
	signature=kbset[nkb]->signature;
	kbset[nkb]->signature++;

	/* Initialize substitution for IsSubsumedByClause() calls. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Loop through clause literals. */
	stack=NULL;
	for (ptr1=&clause->part2.bin->formula[0],mm=0;ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS),mm++){

		/* Loop through ACTIVE and PASSIVE clauses. ACTIVE clauses have */
		/* selected and non selected literals. */
		for (kk=ACTIVE;kk!=-1;
				kk=(kk==ACTIVE?ACTIVE|SELECTED:(kk==PASSIVE?-1:(type&PASSIVE?PASSIVE:-1)))){

			/* Get pointer to top tree node. */
			switch (kk){
				case ACTIVE:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr4[nkb];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr3[nkb];
					}
					break;
				case PASSIVE:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr6[nkb];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr5[nkb];
					}
					break;
				default:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr2[nkb];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr1[nkb];
					}
					break;
			}

			/* Loop through generalization candidates. */
			if (stack!=NULL){
				stack->used=0;
			}
			do {

				/* Check timeout and solution by other thread. */
				if (nkb!=num_wkbs){
					gettimeofday(&timestr,NULL);
					time=timestr.tv_sec-kbset[nkb]->time.tv_sec
							+(timestr.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0;
		 			if ((time>=kbset[nkb]->opts.timeout)||(procctl.status==UNSATISFIABLE)
		 					||(procctl.status==SATISFIABLE)){
						if (stack!=NULL){
							MYFREE(stack);
						}
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						return(TIMEOUT);
					}
		 			if ((procctl.status==NOMEMORY)||(procctl.status==TIMEOUT)){
						if (stack!=NULL){
							MYFREE(stack);
						}
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						return(procctl.status);
					}
				}

				/* Get a candidate subsuming literal. */
				switch (ii=QueryGeneric(NextItem(ptr1,IMMED),&ptr2,&stack,0,ptr3)){

					/* Not enough memory. */
					case NOMEMORY:
						if (stack!=NULL){
							MYFREE(stack);
						}
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						return(NOMEMORY);
						break;

					/* We have a valid candidate. */
					case 1:

						/* Get pointer to candidate binprefix structure. */
						binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

						/* Debug. */
						#ifdef DEBUGCODE
						ptr10=Decompile(binp);
						MYFREE(ptr10);
						#endif

						/* Check if candidate clause has been already processed. */
						if (binp->signature!=signature){

							/* Check subsumption. */
							switch (jj=IsSubsumedByClause(ptr1,clause->part2.bin->literals-mm,
									binp,&Subst)){

								/* No memory. */
								case NOMEMORY:
									if (stack!=NULL){
										MYFREE(stack);
									}
									if (Subst.buffer!=NULL){
										MYFREE(Subst.buffer);
									}
									return(NOMEMORY);
									break;

								/* Clause is not subsumed. Sign candidate to prevent checking */
								/* it again. */
								case 0:
									binp->signature=signature;
									break;

								/* Clause is subsumed. */
								default:

									/* Get locking assertions. */
									clause->flags&=~(TEXTFORMULA|UNPROC|PASSIVE|ACTIVE|LOCKED);
									if (NOMEMORY==GetLocksAsserts(binp,clause->part2.bin,NULL)){
										if (stack!=NULL){
											MYFREE(stack);
										}
										if (Subst.buffer!=NULL){
											MYFREE(Subst.buffer);
										}
										return(NOMEMORY);
									}

									/* Free memory and return. */
									if (stack!=NULL){
										MYFREE(stack);
									}
									if (Subst.buffer!=NULL){
										MYFREE(Subst.buffer);
									}
									return(jj);
									break;
							}
						}
						break;
				}
			} while(ii);
		}
	}

	/* Free memory and return clause not subsumed. */
	if (stack!=NULL){
		MYFREE(stack);
	}
	if (Subst.buffer!=NULL){
		MYFREE(Subst.buffer);
	}
	return(0);
} /* CheckFwSubsuming */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform backward subsumption inference)  OCJ
 *
 *    This function checks if the given clause in compiled or binary
 *    format subsumes one or more clauses that exist in the indicated KB
 *    by calling IsSubsumedByClause() function for every candidate clause
 *    in the requested KB. It must also be specified the type of subsuming
 *    clauses that will be considered, either active or passive clauses.
 *
 *    Candidate clauses are selected by inspecting the first literal in the
 *    argument clause and selecting clauses with an instantiation literal.
 *
 *    Subsumed clauses are either locked and have their locking assertions
 *    set or they are added to the disposal queue but not removed from
 *    their current queue.
 *
 *    See IsSubsumedByClause() function for details of clause subsumption
 *    condition.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in compiled binary form.
 *    nkb: KB index of KB that will be searched for subsumed clauses
 *    type: PASSIVE if passive clauses are to be checked as subsumed
 *          clauses. Active clauses are always checked. Any other
 *          flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Process was ok.
 *    1 -> Process performed and an empty clause without assertions was
 *         generated. This function never returns 1 but this is
 *         documented here as a possible return code from functions
 *         called by BckSimplify() function.
 *    TIMEOUT -> Time has expired or other thread has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t BckSubsuming(cmprefix *clause,int32_t nkb,int32_t type){
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto dscstack *stack;                                /* Pointer to stack for sequential calls to QueryGeneric() */
	auto subst Subst;                                    /* Substitution for calls to IsSubsumedByClause() */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr3;                      /* Auxiliary pointers */
	auto int32_t ii,jj,kk;                               /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Initialize substitution for IsSubsumedByClause() calls. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Loop through ACTIVE and PASSIVE clauses. ACTIVE clauses have */
	/* selected and non selected literals. */
	stack=NULL;
	ptr1=&clause->part2.bin->formula[0];
	for (kk=ACTIVE;kk!=-1;
			kk=(kk==ACTIVE?ACTIVE|SELECTED:(kk==PASSIVE?-1:(type&PASSIVE?PASSIVE:-1)))){

		/* Get pointer to top tree node. */
		switch (kk){
			case ACTIVE:
				if (ptr1[0]&NEGATED){
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr4[nkb];
				} else {
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr3[nkb];
				}
				break;
			case PASSIVE:
				if (ptr1[0]&NEGATED){
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr6[nkb];
				} else {
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr5[nkb];
				}
				break;
			default:
				if (ptr1[0]&NEGATED){
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr2[nkb];
				} else {
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr1[nkb];
				}
				break;
		}

		/* Loop through instantiation candidates. */
		if (stack!=NULL){
			stack->used=0;
		}
		do {

			/* Check timeout and solution by other thread. */
			if (nkb!=num_wkbs){
				gettimeofday(&timestr,NULL);
				time=timestr.tv_sec-kbset[nkb]->time.tv_sec
						+(timestr.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0;
	 			if ((time>=kbset[nkb]->opts.timeout)||(procctl.status==UNSATISFIABLE)
	 					||(procctl.status==SATISFIABLE)){
					if (stack!=NULL){
						MYFREE(stack);
					}
					if (Subst.buffer!=NULL){
						MYFREE(Subst.buffer);
					}
					return(TIMEOUT);
				}
	 			if ((procctl.status==NOMEMORY)||(procctl.status==TIMEOUT)){
					if (stack!=NULL){
						MYFREE(stack);
					}
					if (Subst.buffer!=NULL){
						MYFREE(Subst.buffer);
					}
					return(procctl.status);
				}
			}

			/* Get a candidate subsumed literal. */
			switch (ii=QueryUI(NextItem(ptr1,IMMED),&ptr2,&stack,0,ptr3,INSTANCE)){

				/* Not enough memory. */
				case NOMEMORY:
					if (stack!=NULL){
						MYFREE(stack);
					}
					if (Subst.buffer!=NULL){
						MYFREE(Subst.buffer);
					}
					return(NOMEMORY);
					break;

				/* We have a valid candidate. */
				case 1:

					/* Get pointer to candidate binprefix structure. */
					binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

					/* Debug. */
					#ifdef DEBUGCODE
					ptr10=Decompile(binp);
					MYFREE(ptr10);
					#endif

					/* Clause is not in disposal queue therefore */
					/* it has not been already simplified. */
					if (0==(binp->clause->flags&INDISPOSALQUEUE)){

						/* Check subsumption. */
						switch (jj=IsSubsumedByClause(&binp->formula[0],binp->literals,
								clause->part2.bin,&Subst)){

							/* Not enough memory. */
							case NOMEMORY:
								if (stack!=NULL){
									MYFREE(stack);
								}
								if (Subst.buffer!=NULL){
									MYFREE(Subst.buffer);
								}
								return(NOMEMORY);
								break;

							/* Candidate clause is subsumed. */
							case 1:

								/* Get locking assertions. */
								if (NOMEMORY==GetLocksAsserts(clause->part2.bin,binp,NULL)){
									if (stack!=NULL){
										MYFREE(stack);
									}
									if (Subst.buffer!=NULL){
										MYFREE(Subst.buffer);
									}
									return(NOMEMORY);
								}

								/* Add subsumed clause to disposal queue for later deletion or */
								/* locking. This queue interconnects clauses from other queues. */
								/* The clause must not be removed from any queue to which it is */
								/* linked. */
								binp->clause->flags|=INDISPOSALQUEUE;
								binp->clause->prevdisp=kbset[nkb]->lastdisp;
								kbset[nkb]->lastdisp=binp->clause;
								kbset[nkb]->prstats.bcksubsum++;
								kbset[nkb]->lrscnt++;

								/* Debug. */
								#ifdef DEBUGCODE
								ptr10=Decompile(binp);
								#ifdef VERBOSE
								if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
									printf("Clause backward subsumed:\n%s\n",ptr10);
								}
								#endif
								MYFREE(ptr10);
								#endif
								break;
						}
					}
					break;
			}
		} while(ii);
	}

	/* Free memory and return. */
	if (stack!=NULL){
		MYFREE(stack);
	}
	if (Subst.buffer!=NULL){
		MYFREE(Subst.buffer);
	}
	return(0);
} /* BckSubsuming */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check if clause is subsumed by another clause)  OCJ
 *
 *    This function checks if clause of formula1 in compiled binary format
 *    is subsumed by clause in formula2 also in compiled binary format.
 *    The clause in formula1 is subsumed by the clause in binp2 if there
 *    exist a substitution such that applied to binp2 and adding 0 or more
 *    literals then formula1 is obtained. This means that clause in binp2 is
 *    both more general and simpler than formula1 because it is true whenever
 *    formula in binp2 is true and it has the same or fewer literals. The
 *    substitution is only applied to clause in binp2 so the variables substituted
 *    must all belong to clause in binp2.
 *
 *    The substitution may be a complex one formed by a set
 *    of single substitutions.
 *
 *    This function doesn't consider assertions at all. This must be done
 *    by the calling function when applicable.
 *
 *    The process is as follows:
 *    - This function computes data related to which literals of the candidate
 *      subsumed clause are instances of each literal of the candidate subsuming
 *      clause.
 *    - Any of the following conditions produce an early return as
 *      with result of no subsuming:
 *      - There are more literals in formula2 than in formula1.
 *      - Two different literals in binp2 have the same single instance literal
 *        in formula1.
 *      - There is a literal in binp2 without any assignable candidate
 *        instance literals.
 *      - There are more literals binp2 than instance literals in formula1.
 *    - The function calls LiteralSubsuming() function and returns the value
 *      obtained.
 *
 *    This algorithm is potentially exponential in execution time. To avoid
 *    hang ups a timeout condition checking is introduced. In case of timeout
 *    it is assumed that the clause is not subsumed. This will affect a little
 *    bit the program performance but it will prevent hang ups and I think that
 *    it preserves completeness. This timeout condition must not be confused
 *    with the global process timeout condition.
 *
 *    IMPORTANT: To ensure that this function works properly the candidate
 *    subsuming clause must not have duplicate literals because if there is a
 *    single instance literal for both duplicate literals this function will be
 *    unable to assign it to both duplicate literals. This is not a problem
 *    if the clause is in the KB as it has been properly simplified already.
 *    However, for cases like equivalent factor simplifications the removal of
 *    duplicate literals must be done in the calling function. For instance,
 *    when equivalent factoring is applied to the clause:
 *      A(x,y)|A(u,v)|B(C,D)|B(x,y)|B(u,v)
 *    where x, y, u, v are variables then a candidate equivalent factor is
 *      A(u,v)|B(C,D)|B(u,v)|B(u,v)
 *    This function cannot assign both duplicate literals B(u,v) to the
 *    instance literal B(x,y) or B(u,v), therefore failing to identify
 *    a subsumption.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    formula1: Pointer to formula in compiled binary form as it is in the
 *              formula field of binprefix structure. This formula will be
 *              checked for being subsumed by binp2. It points to a literal
 *              in the clause that is being checked for being subsumed, but
 *              not necessarily to the first literal.
 *    lits1: Number of literals starting at formula1 pointer.
 *    binp2: Pointer to binprefix structure of clause in compiled binary. This
 *           clause will be checked for subsuming formula1.
 *    Subst: Pointer to a substitution structure. This structure is
 *           kept between calls to this function in order to save
 *           allocation and freeing operations of substitution buffer.
 *
 *  RETURNS:
 *
 *    0 -> Clause is not subsumed or a function timeout condition occurred.
 *    1 -> Clause is subsumed.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t IsSubsumedByClause(uint8_t *formula1,int32_t lits1,binprefix *binp2,subst *Subst){
	auto subsumdata ssdata;                              /* Subsuming data structure. */
	struct timeval sttime;                               /* Function starting time */
	auto uint8_t *ptr1,*ptr2;                            /* Auxiliary pointers */
	auto int32_t ii,jj;                                  /* Auxiliary */

	/* Check the number of literals. This is not strictly necessary */
	/* unless factoring is on, as then the factor would always be */
	/* subsumed by the parent. However it helps simplify clauses */
	/* and speeds up the program. */
	if (binp2->literals>lits1){
		return(0);
	}

	/* Allocate subsumdata buffers. */
	if (NULL==(ssdata.pinst=mallocmn(binp2->literals*((sizeof(int32_t)*(1+lits1))
			+(sizeof(void *)*(lits1+2)))))){
		return(NOMEMORY);
	}

	/* Get start time. */
	gettimeofday(&sttime,NULL);

	/* Initialize basic data in subsumdata structure. */
	ssdata.lits1=lits1;
	ssdata.lits2=binp2->literals;
	ssdata.pinst[0]=(uint8_t **)&ssdata.pinst[binp2->literals];
	for (ii=1;ii<binp2->literals;ii++){
		ssdata.pinst[ii]=&ssdata.pinst[ii-1][lits1];
	}
	ssdata.instnb=(int32_t **)&ssdata.pinst[binp2->literals-1][lits1];
	ssdata.instnb[0]=(int32_t *)&ssdata.instnb[binp2->literals];
	for (ii=1;ii<binp2->literals;ii++){
		ssdata.instnb[ii]=&ssdata.instnb[ii-1][lits1];
	}
	ssdata.nbofinst=(int32_t *)&ssdata.instnb[binp2->literals-1][lits1];

	/* Fill data in buffers of subsumdata structure. */
	/* Loop through literals of candidate subsuming clause. */
	for (ptr1=&binp2->formula[0],ii=0;ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS),ii++){

		/* Initialize number of instance literals for this candidate subsuming literal. */
		ssdata.nbofinst[ii]=0;

		/* Loop through literals of candidate subsumed formula. */
		for (ptr2=formula1,jj=0;ptr2[0]!=UNITEND;ptr2=NextItem(ptr2,OVERSUBTERMS),jj++){

			/* Check if ptr1 is a generalization of ptr2. */
			switch (Generalize(ptr1,ptr2,Subst,ITEM2SEC|INITSUBST)){

				/* Not enough memory. */
				case NOMEMORY:
					MYFREE(ssdata.pinst);
					return(NOMEMORY);
					break;

				/* Valid generalization. */
				case 1:

					/* Substitution is empty therefore literals are equal. */
					/* Indicate it and leave the loop of literals in  candidate */
					/* subsumed formula. */
					if ((Subst->buffer==NULL)||(Subst->substsize==1)){
						ssdata.nbofinst[ii]=-1;
						break;
					}

					/* Literals are not equal. Set subsuming data. */
					ssdata.pinst[ii][ssdata.nbofinst[ii]]=ptr2;
					ssdata.instnb[ii][ssdata.nbofinst[ii]]=jj;
					ssdata.nbofinst[ii]++;
					break;
			}

			/* If literals are equal leave the loop of literals in candidate */
			/* subsumed formula. */
			if (ssdata.nbofinst[ii]==-1){
				break;
			}
		}

		/* Return no subsumption if there are no candidate instance literals */
		/* for this literal. */
		if (ssdata.nbofinst[ii]==0){
			MYFREE(ssdata.pinst);
			return(0);
		}
	}

	/* Initialize substitution if necessary. */
	if (Subst->buffer!=NULL){
		Subst->substsize=1;
		Subst->buffer[0]=UNITEND;
	} else {
		Subst->substsize=0;
	}

	/* Call LiteralSubsuming() function. */
	ii=LiteralSubsuming(&binp2->formula[0],&binp2->formula[0],0,&ssdata,Subst,&sttime);
	if (ii==TIMEOUT){
		ii=0;
	}

	/* Free memory and return. */
	MYFREE(ssdata.pinst);
	return(ii);
} /* IsSubsumedByClause */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (IsSubsumedByClause() version for subsumption resolution)  OCJ
 *
 *    This function is the specific version of IsSubsumedByClause() function
 *    for subsumption resolution simplifications. The main difference is that
 *    it takes a (possibly non empty) substitution as argument and that
 *    substitution must be compatible with the final subsumption substitution.
 *
 *    The substitution must have binp1 items as secondary.
 *
 *    This function doesn't consider assertions at all. This must be done
 *    by the calling function when applicable.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    binp1: Pointer to binprefix structure of clause in compiled binary. This
 *           clause will be checked for being subsumed by binp2.
 *    binp2: Pointer to binprefix structure of clause in compiled binary. This
 *           clause will be checked for subsuming binp1.
 *    Subst: Pointer to a substitution structure. This substitution must be
 *           compatible with the final subsumption substitution.
 *
 *  RETURNS:
 *
 *    0 -> Clause is not subsumed or a function timeout condition occurred.
 *    1 -> Clause is subsumed.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t IsSubsumedByClause2(binprefix *binp1,binprefix *binp2,subst *Subst){
	auto subsumdata ssdata;                              /* Subsuming data structure. */
	struct timeval sttime;                               /* Function starting time */
	auto uint8_t *ptr1,*ptr2;                            /* Auxiliary pointers */
	auto int32_t ii,jj,kk;                               /* Auxiliary */

	/* Check the number of literals. This is not strictly necessary. */
	/* However it helps simplify clauses and speeds up the program. */
	if (binp2->literals>binp1->literals){
		return(0);
	}

	/* If candidate subsuming clause has no literals then return */
	/* successful subsumption. */
	if (binp2->literals==0){
		return(1);
	}

	/* Allocate subsumdata buffers. */
	if (NULL==(ssdata.pinst=mallocmn(binp2->literals*((sizeof(int32_t)*(1+binp1->literals))
			+(sizeof(void *)*(binp1->literals+2)))))){
		return(NOMEMORY);
	}

	/* Remember substitution status. */
	if (Subst->buffer!=NULL){
		kk=Subst->substsize;
	} else {
		kk=-1;
	}

	/* Get start time. */
	gettimeofday(&sttime,NULL);

	/* Initialize basic data in subsumdata structure. */
	ssdata.lits1=binp1->literals;
	ssdata.lits2=binp2->literals;
	ssdata.pinst[0]=(uint8_t **)&ssdata.pinst[binp2->literals];
	for (ii=1;ii<binp2->literals;ii++){
		ssdata.pinst[ii]=&ssdata.pinst[ii-1][binp1->literals];
	}
	ssdata.instnb=(int32_t **)&ssdata.pinst[binp2->literals-1][binp1->literals];
	ssdata.instnb[0]=(int32_t *)&ssdata.instnb[binp2->literals];
	for (ii=1;ii<binp2->literals;ii++){
		ssdata.instnb[ii]=&ssdata.instnb[ii-1][binp1->literals];
	}
	ssdata.nbofinst=(int32_t *)&ssdata.instnb[binp2->literals-1][binp1->literals];

	/* Fill data in buffers of subsumdata structure. */
	/* Loop through literals of candidate subsuming clause. */
	for (ptr1=&binp2->formula[0],ii=0;ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS),ii++){

		/* Initialize number of instance literals for this candidate subsuming literal. */
		ssdata.nbofinst[ii]=0;

		/* Loop through literals of candidate subsumed formula. */
		for (ptr2=&binp1->formula[0],jj=0;ptr2[0]!=UNITEND;ptr2=NextItem(ptr2,OVERSUBTERMS),jj++){

			/* Reset substitution to its previous state. */
			if (kk>0){
				Subst->substsize=kk;
				Subst->buffer[kk-1]=UNITEND;
			} else if (Subst->buffer!=NULL){
				Subst->substsize=1;
				Subst->buffer[0]=UNITEND;
			} else {
				Subst->substsize=0;
			}

			/* Check if ptr1 is a generalization of ptr2. */
			switch (Generalize(ptr1,ptr2,Subst,ITEM2SEC)){

				/* Not enough memory. */
				case NOMEMORY:
					MYFREE(ssdata.pinst);
					return(NOMEMORY);
					break;

				/* Valid generalization. */
				case 1:

					/* Substitution has not changed and literals are identical. */
					/* Indicate it and leave the loop of literals in  candidate */
					/* subsumed formula. */
					if ((Subst->buffer==NULL)||(Subst->substsize==1)
							||((Subst->substsize==kk)&&(IsEqual2(ptr1,ptr2)))){
						ssdata.nbofinst[ii]=-1;
						break;
					}

					/* Literals are not equal. Set subsuming data. */
					ssdata.pinst[ii][ssdata.nbofinst[ii]]=ptr2;
					ssdata.instnb[ii][ssdata.nbofinst[ii]]=jj;
					ssdata.nbofinst[ii]++;
					break;
			}

			/* If literals are equal leave the loop of literals in candidate */
			/* subsumed formula. */
			if (ssdata.nbofinst[ii]==-1){
				break;
			}
		}

		/* Return no subsumption if there are no candidate instance literals */
		/* for this literal. */
		if (ssdata.nbofinst[ii]==0){
			MYFREE(ssdata.pinst);
			return(0);
		}
	}

	/* Reset substitution to its previous state. */
	if (kk>0){
		Subst->substsize=kk;
		Subst->buffer[kk-1]=UNITEND;
	} else if (Subst->buffer!=NULL){
		Subst->substsize=1;
		Subst->buffer[0]=UNITEND;
	} else {
		Subst->substsize=0;
	}

	/* Call LiteralSubsuming() function. */
	ii=LiteralSubsuming(&binp2->formula[0],&binp2->formula[0],0,&ssdata,Subst,&sttime);
	if (ii==TIMEOUT){
		ii=0;
	}

	/* Free memory and return. */
	MYFREE(ssdata.pinst);
	return(ii);
} /* IsSubsumedByClause2 */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check literal subsuming)  OCJ
 *
 *    This function checks if there is a valid corresponding literal in
 *    candidate subsumed formula that can be subsumed by the passed literal
 *    in candidate subsuming formula that allows candidate subsuming formula
 *    to subsume candidate subsumed formula.
 *
 *    This function is called recursively.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    formula2: Pointer to formula in compiled binary form as it is in the
 *              formula field of binprefix structure. This formula will be
 *              checked for subsuming clause1.
 *    literal: Pointer to literal in candidate subsuming formula to be checked.
 *    litnb: Literal number of literal in candidate subsuming formula to be
 *           checked.
 *    pssdata: Pointer to structure with subsuming data.
 *    Subst: Pointer to a substitution structure. This structure is
 *           kept between calls to this function in order to save
 *           allocation and freeing operations of substitution buffer.
 *    sttime: Structure with the function starting time to check function
 *            timeouts. See IsSubsumedByClause() global comments for details.
 *
 *  RETURNS:
 *
 *    0 -> There is no valid corresponding literal that allows subsuming.
 *    1 -> There is a valid corresponding literal that allows subsuming.
 *    NOMEMORY -> Not enough memory.
 *    TIMEOUT -> Function timeout condition
 *
 *
 *--------------------------------------------------------------*/
int32_t LiteralSubsuming(uint8_t *formula2,uint8_t *literal,int32_t litnb,
		subsumdata *pssdata,subst *Subst,struct timeval *sttime){
	auto struct timeval endtime;                         /* Function current time */
	auto double timedif;                                 /* Function elapsed time */
	auto int32_t ii,jj,kk;                               /* Auxiliary */

	/* If no more literals then return valid subsumption. */
	if (literal[0]==UNITEND){
		return(1);
	}

	/* Literal is equal to some literal in candidate subsumed formula. */
	/* Make recursive call. */
	if (pssdata->nbofinst[litnb]==-1){
		return(LiteralSubsuming(formula2,NextItem(literal,OVERSUBTERMS),
				litnb+1,pssdata,Subst,sttime));

	/* Literal is not equal to any literal in candidate subsumed formula. */
	} else {

		/* Remember substitution status. */
		if (Subst->buffer!=NULL){
			kk=Subst->substsize;
		} else {
			kk=-1;
		}

		/* Loop through candidate instance literals for the given literal. */
		for (ii=0;ii<pssdata->nbofinst[litnb];ii++){

			/* Check valid generalization. */
			switch (Generalize(literal,pssdata->pinst[litnb][ii],Subst,ITEM2SEC)){

				/* Not enough memory. */
				case NOMEMORY:
					return(NOMEMORY);
					break;

				/* Valid generalization. Make recursive call. */
				case 1:
					jj=LiteralSubsuming(formula2,NextItem(literal,OVERSUBTERMS),
							litnb+1,pssdata,Subst,sttime);
					if (0!=jj){
						return(jj);
					}
					break;
			}

			/* Reset substitution to its previous state. */
			if (kk>0){
				Subst->substsize=kk;
				Subst->buffer[kk-1]=UNITEND;
			} else if (Subst->buffer!=NULL){
				Subst->substsize=1;
				Subst->buffer[0]=UNITEND;
			} else {
				Subst->substsize=0;
			}

			/* Check timeout. */
			gettimeofday(&endtime,NULL);
			timedif=endtime.tv_sec-sttime->tv_sec
					+(endtime.tv_usec-sttime->tv_usec)/1000000.0;
			if (timedif>0.0185){
				return(TIMEOUT);
			}
		}
	}

	/* Return no subsumption. */
	return(0);
} /* LiteralSubsuming */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Forward and backward simplifications)  OCJ
 *
 *    This function performs the forward simplifications of a clause.
 *    This function also checks for getting an empty clause without
 *    assertions. If empty clause without assertions is detected then
 *    it is added to the UNPROC queue.
 *
 *    The input clause must be unlinked form the owner KB.
 *
 *    This function just calls the appropriate functions that perform
 *    each simplification inference. Any required memory freeing is
 *    performed by this function or the functions called by it, even
 *    in the case of error and also for the memory allocated to *pclause
 *    by the caller.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    pclause: Address of pointer to cmprefix structure of clause.
 *    nkb: KB where the text clauses are placed.
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          FROMPASSIVE if the clause is coming from PASSIVE queue and
 *          therefore it only needs a subsumption check in FirstSimplify().
 *          NOBKSIMPLIFY if backward simplification must not be performed.
 *
 *  RETURNS:
 *
 *    0 -> Process performed and no empty clause without assertions inferred.
 *    1 -> Process performed and an empty clause without assertions was inferred.
 *    TIMEOUT -> Time has expired or other thread has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t FwdSimplify(cmprefix **pclause,int32_t nkb,int32_t type){
	auto int32_t ii;                                     /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;

	/* Debug. */
	ptr10=Decompile((*pclause)->part2.bin);
	MYFREE(ptr10);
	#endif

	#if DEMODPOSITION == 1
	/* Perform demodulations if there are equalities. */
	if ((UNITEQUALITY|NONUNITEQUALITY)&pbmflags){
		for (ii=1;ii!=0;){
			ii=FwdDemodulation(pclause,nkb,type);
			if ((ii==TIMEOUT)||(ii==NOMEMORY)){
				return(ii);
			}
		}
	}
	#endif

	/* Perform initial forward simplifications. If there is a NOMEMORY */
	/* error or if the clause is deleted then memory freeing is */
	/* performed by FirstSimplify() function, not here. */
	switch (ii=FirstSimplify(pclause,nkb,type)){

		/* Not enough memory or timeout. */
		case NOMEMORY:
		case TIMEOUT:
			return(ii);
			break;

		/* Clause has been deleted or subsumed. */
		case 2:
		case 3:
			*pclause=NULL;
			return(0);
			break;

		/* Clause has been locked. */
		case 4:
			return(0);
			break;
	}

	/* Check empty clause. This can be achieved by trivial equality resolution */
	/* and destructive equality resolution in FirstSimplify() function. */
	if ((*pclause)->part2.bin->formula[0]==UNITEND){

		/* Empty clause has not assertions. */
		if ((*pclause)->part2.bin->asserts==NULL){
			(*pclause)->flags=UNPROC;
			if (NOMEMORY==AddBinClause2KB(*pclause,nkb,SETFNUMBER)){
				FreeEquData((*pclause)->part2.bin);
				if ((*pclause)->part2.bin->asserts!=NULL){
					MYFREE((*pclause)->part2.bin->asserts);
				}
				MYFREE((*pclause)->part2.bin);
				MYFREE(*pclause);
				return(NOMEMORY);
			}
			if (NULL==(kbset[nkb]->frstprfnode=mallocmn(sizeof(proofnode)))){
				return(NOMEMORY);
			}
			kbset[nkb]->frstprfnode->type=ACLAUSE;
			kbset[nkb]->frstprfnode->ptr.Aclause=*pclause;
			kbset[nkb]->status=UNSATISFIABLE;
			return(1);
		}

		/* Empty clause has assertions. Return as an empty clause cannot */
		/* be further simplified. */
		return(0);
	}

	#if DEMODPOSITION == 2
	/* Perform demodulations if there are equalities. */
	if ((UNITEQUALITY|NONUNITEQUALITY)&pbmflags){
		for (ii=1;ii!=0;){
			ii=FwdDemodulation(pclause,nkb,type);
			if ((ii==TIMEOUT)||(ii==NOMEMORY)){
				return(ii);
			}
		}
	}
	#endif

	/* Perform additional forward simplification. If there is a NOMEMORY */
	/* error or if the clause is deleted then memory freeing is */
	/* performed by FwdSimplify() function, not here. */
	switch (ii=SecondSimplify(pclause,nkb,type)){
		case NOMEMORY:
		case TIMEOUT:
			return(ii);
			break;

	}

	/* Check empty clause without assertions. This can be achieved by subsumption */
	/* resolution in SecondSimplify() function. */
	if ((*pclause)->part2.bin->formula[0]==UNITEND){

		/* Empty clause has not assertions. */
		if ((*pclause)->part2.bin->asserts==NULL){
			(*pclause)->flags=UNPROC;
			if (NOMEMORY==AddBinClause2KB(*pclause,nkb,SETFNUMBER)){
				FreeEquData((*pclause)->part2.bin);
				if ((*pclause)->part2.bin->asserts!=NULL){
					MYFREE((*pclause)->part2.bin->asserts);
				}
				MYFREE((*pclause)->part2.bin);
				MYFREE(*pclause);
				return(NOMEMORY);
			}
			if (NULL==(kbset[nkb]->frstprfnode=mallocmn(sizeof(proofnode)))){
				return(NOMEMORY);
			}
			kbset[nkb]->frstprfnode->type=ACLAUSE;
			kbset[nkb]->frstprfnode->ptr.Aclause=*pclause;
			kbset[nkb]->status=UNSATISFIABLE;
			return(1);
		}

		/* Empty clause has assertions. Return as an empty clause cannot */
		/* be further simplified. */
		return(0);
	}

	#if DEMODPOSITION == 3
	/* Perform demodulations if there are equalities. */
	if ((UNITEQUALITY|NONUNITEQUALITY)&pbmflags){
		for (ii=1;ii!=0;){
			ii=FwdDemodulation(pclause,nkb,type);
			if ((ii==TIMEOUT)||(ii==NOMEMORY)){
				return(ii);
			}
		}
	}
	#endif

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile((*pclause)->part2.bin);
	MYFREE(ptr10);
	#endif

	return(0);
} /* FwdSimplify */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (First clause simplification)  OCJ
 *
 *    This function performs the following simplifications in a clause:
 *    - Simplifications that detect clause redundancy, leading to
 *      clause deletion: tautology detection and forward subsumption.
 *    - Simplifications that don't involve any other external clause:
 *      remove duplicate literals, equivalent factoring, trivial
 *      equality resolution and destructive equality resolution.
 *
 *    If the clause is deleted then its storage is freed. The clause
 *    storage is also freed in case of a NOMEMORY error condition.
 *
 *    This function just calls the appropriate functions that perform
 *    each simplification inference. Any required memory freeing is
 *    performed by this function, even in the case of error and also
 *    for the memory allocated to *pclause by the caller.
 *
 *    IMPORTANT: This function must never be called for main KB.
 *    if this changes in the future then the set of pkb variable
 *    must be changed accordingly.
 *
 *
 *  ARGUMENTS:
 *
 *    pclause: Address of pointer to cmprefix structure of clause.
 *    nkb: KB where the text clauses are placed.
 *    flags: PASSIVE if passive clauses are to be checked for
 *           simplifications. Active clauses are always checked.
 *           FROMPASSIVE if the clause is coming from PASSIVE queue and
 *           therefore it only needs a subsumption check in FirstSimplify().
 *
 *  RETURNS:
 *
 *    0 -> No simplifications were done.
 *    1 -> Some simplifications were done.
 *    2 -> Clause was deleted because it is a tautology.
 *    3 -> Clause was subsumed.
 *    4 -> Clause was locked because it was subsumed.
 *    TIMEOUT -> Time has expired or other thread has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t FirstSimplify(cmprefix **pclause,int32_t nkb,int32_t flags){
	auto cmprefix *smpclauses[4];                        /* Pointers to simplified clauses */
	auto int32_t smpsize;                                /* Number of elements in smpclauses set */
	auto kbase *pkb;                                     /* Address of pointer to KB */
	auto char *ptr10;                                    /* Auxiliary pointer */
	auto int32_t ii,jj;                                  /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr11;                                    /* Auxiliary pointer */
	#endif

	/* Get address of pointer to KB. This function is never called */
	/* for main KB. If this changes in the future then see for */
	/* instance how this is done in Initialize_KB() function. */
	pkb=kbset[nkb];

	/* Check if clause is a tautology. This is done here only if there */
	/* are not equalities. */
	if ((0==(flags&FROMPASSIVE))&&(0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))){
		if (IsTautology(*pclause)){
			pkb->prstats.tautologies++;
			FreeEquData((*pclause)->part2.bin);
			if ((*pclause)->part2.bin->asserts!=NULL){
				MYFREE((*pclause)->part2.bin->asserts);
			}
			MYFREE((*pclause)->part2.bin);
			MYFREE(*pclause);
			return(2);
		}
	}

	/* Initialize the set of simplified clauses. The last element with */
	/* index [smpsize-1] contains the last result of a simplification */
	/* in binary format. The other elements contain simplified clauses */
	/* in text format. */
	smpsize=1;
	smpclauses[0]=*pclause;

	/* Simplifications done only if clause is not coming from PASSIVE. */
	if (0==(flags&FROMPASSIVE)){

		/* Equivalent factoring is ON. Try to take an equivalent factor of clause. */
		if (kbset[nkb]->opts.equivfact){
			switch (ii=EquivalentFactor(*pclause,&smpclauses[1],nkb)){
				case NOMEMORY:
				case TIMEOUT:
					FreeEquData((*pclause)->part2.bin);
					if ((*pclause)->part2.bin->asserts!=NULL){
						MYFREE((*pclause)->part2.bin->asserts);
					}
					MYFREE((*pclause)->part2.bin);
					MYFREE(*pclause);
					return(ii);
					break;
				case 1:
					smpsize=2;
					pkb->prstats.equivfact++;
					break;
			}

		/* Equivalent factoring is OFF. Remove duplicate literals. */
		} else {
			switch (RemoveDuplicates(*pclause,&smpclauses[1],nkb,1)){
				case NOMEMORY:
					FreeEquData((*pclause)->part2.bin);
					if ((*pclause)->part2.bin->asserts!=NULL){
						MYFREE((*pclause)->part2.bin->asserts);
					}
					MYFREE((*pclause)->part2.bin);
					MYFREE(*pclause);
					return(NOMEMORY);
					break;
				case 1:
					smpsize=2;
					pkb->prstats.removedups++;
					break;
			}
		}

		/* Debug. */
		#ifdef DEBUGCODE
		ptr11=Decompile(smpclauses[smpsize-1]->part2.bin);
		MYFREE(ptr11);
		#endif

		/* There are equalities. */
		if ((UNITEQUALITY|NONUNITEQUALITY)&pbmflags){

			/* Trivial equality resolution. */
			switch (TrivEqResolution(smpclauses[smpsize-1],&smpclauses[smpsize])){

				/* Not enough memory. */
				case NOMEMORY:
					for (ii=0;ii<(smpsize-1);ii++){
						MYFREE(smpclauses[ii]->part2.text);
						MYFREE(smpclauses[ii]);
					}
					FreeEquData(smpclauses[smpsize-1]->part2.bin);
					if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
						MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
					}
					MYFREE(smpclauses[smpsize-1]->part2.bin);
					MYFREE(smpclauses[smpsize-1]);
					return(NOMEMORY);
					break;

				/* Clause was simplified. If the result is an empty clause */
				/* then add text clauses to KB, set pclause to simplified */
				/* clause and return. */
				case 1:
					smpsize++;
					pkb->prstats.triveqres++;
					if (smpclauses[smpsize-1]->part2.bin->formula[0]==UNITEND){
						for (ii=0;ii<(smpsize-1);ii++){
							AddTxtFormula2KB(smpclauses[ii],smpclauses[ii]->inference,nkb,SETFNUMBER);
						}
						*pclause=smpclauses[smpsize-1];
						return(1);
					}
					break;
			}

			/* Destructive equality resolution. */
			switch (DestrEqResolution(smpclauses[smpsize-1],&smpclauses[smpsize])){

				/* Not enough memory. */
				case NOMEMORY:
					for (ii=0;ii<(smpsize-1);ii++){
						MYFREE(smpclauses[ii]->part2.text);
						MYFREE(smpclauses[ii]);
					}
					FreeEquData(smpclauses[smpsize-1]->part2.bin);
					if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
						MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
					}
					MYFREE(smpclauses[smpsize-1]->part2.bin);
					MYFREE(smpclauses[smpsize-1]);
					return(NOMEMORY);
					break;

				/* Clause was simplified. If the result is an empty clause */
				/* then add text clauses to KB, set pclause to simplified */
				/* clause and return. */
				case 1:
					smpsize++;
					pkb->prstats.desteqres++;
					if (smpclauses[smpsize-1]->part2.bin->formula[0]==UNITEND){
						for (ii=0;ii<(smpsize-1);ii++){
							AddTxtFormula2KB(smpclauses[ii],smpclauses[ii]->inference,nkb,SETFNUMBER);
						}
						*pclause=smpclauses[smpsize-1];
						return(1);
					}
					break;
			}

			/* Check if clause is a tautology. This is done here because destructive */
			/* equality resolution may produce a tautology. For instance, de clause */
			/* ~X0=X1|~X2=X3|vapp(X0,X2)=vapp(X1,X3) produces the clause */
			/* vapp(X0,X2)=vapp(X0,X2) that is a tautology. */
			if (0==(flags&FROMPASSIVE)){
				if (IsTautology(smpclauses[smpsize-1])){
					pkb->prstats.tautologies++;
					for (ii=0;ii<(smpsize-1);ii++){
						MYFREE(smpclauses[ii]->part2.text);
						MYFREE(smpclauses[ii]);
					}
					FreeEquData(smpclauses[smpsize-1]->part2.bin);
					if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
						MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
					}
					MYFREE(smpclauses[smpsize-1]->part2.bin);
					MYFREE(smpclauses[smpsize-1]);
					return(2);
				}
			}
		}
	}

	/* Check clause subsumption. */
	switch (jj=CheckFwSubsuming(smpclauses[smpsize-1],nkb,flags)){

		/* Not enough memory. */
		case NOMEMORY:
		case TIMEOUT:
			for (ii=0;ii<(smpsize-1);ii++){
				MYFREE(smpclauses[ii]->part2.text);
				MYFREE(smpclauses[ii]);
			}
			FreeEquData(smpclauses[smpsize-1]->part2.bin);
			if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
				MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
			}
			MYFREE(smpclauses[smpsize-1]->part2.bin);
			MYFREE(smpclauses[smpsize-1]);
			return(jj);
			break;

		/* Clause is not subsumed. */
		case 0:

			/* Debug. */
			#ifdef DEBUGCODE
			#ifdef VERBOSE
			if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
				for (ii=1;ii<(smpsize-1);ii++){
					printf("First simplification result:\n%s\n",smpclauses[ii]->part2.text);
				}
			}
			#endif
			if (smpsize>1){
				ptr10=Decompile(smpclauses[smpsize-1]->part2.bin);
				#ifdef VERBOSE
				if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
					printf("First simplification result:\n%s\n",ptr10);
				}
				#endif
				MYFREE(ptr10);
			}
			#endif

			/* Add text clauses to KB and set pclause to simplified clause. */
			for (ii=0;ii<(smpsize-1);ii++){
				AddTxtFormula2KB(smpclauses[ii],smpclauses[ii]->inference,nkb,SETFNUMBER);
			}
			*pclause=smpclauses[smpsize-1];
			break;

		/* Clause is subsumed. */
		case 1:
			#if defined(DEBUGCODE) && defined(VERBOSE)
			if ((num_wkbs==1)&&(procctl.status==RUNNING)&&(verbfrom<=kbset[0]->nformulas)
					&&(verbto>=kbset[0]->nformulas)){
				printf("Inferred clause was subsumed\n");
			}
			#endif

			/* Count subsumption. */
			pkb->prstats.fwdsubsum++;

			/* Add text clauses to KB. This is always done because due to AVATAR */
			/* successive lock and unlock cycles a clause can be subsumed after */
			/* generating a simplified successor an so be part of the proof queue. */
			for (ii=0;ii<(smpsize-1);ii++){
				AddTxtFormula2KB(smpclauses[ii],smpclauses[ii]->inference,nkb,SETFNUMBER);
			}

			/* Clause is not locked. Convert it to text form. This is always done */
			/* because due to AVATAR successive lock and unlock cycles a clause can */
			/* be subsumed after it generates a simplified successor an so be part of */
			/* the proof queue. */
			if (0==(smpclauses[smpsize-1]->flags&LOCKED)){
				if (NULL==(ptr10=Decompile(smpclauses[smpsize-1]->part2.bin))){
					if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
						MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
					}
					MYFREE(smpclauses[smpsize-1]->part2.bin);
					MYFREE(smpclauses[smpsize-1]);
					return(NOMEMORY);
				}
				FreeEquData(smpclauses[smpsize-1]->part2.bin);
				if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
					MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
				}
				MYFREE(smpclauses[smpsize-1]->part2.bin);
				smpclauses[smpsize-1]->part2.text=ptr10;
				AddTxtFormula2KB(smpclauses[smpsize-1],smpclauses[smpsize-1]->inference,nkb,SETFNUMBER);
				return(3);

			/* Clause is locked. */
			} else {

				/* Set pclause to simplified clause. */
				*pclause=smpclauses[smpsize-1];

				/* Add clause to LOCKED queue and return. */
				/* If a NOMEMORY condition arises then delete it. */
				smpclauses[smpsize-1]->flags&=(~(UNPROC|PASSIVE|ACTIVE));
				smpclauses[smpsize-1]->flags|=LOCKED;
				if (NOMEMORY==AddBinClause2KB(smpclauses[smpsize-1],nkb,0)){
					MYFREE(smpclauses[smpsize-1]->part2.bin->lockasserts);
					if (smpclauses[smpsize-1]->part2.bin->asserts!=NULL){
						MYFREE(smpclauses[smpsize-1]->part2.bin->asserts);
					}
					MYFREE(smpclauses[smpsize-1]->part2.bin);
					MYFREE(smpclauses[smpsize-1]);
					return(NOMEMORY);
				}
				return(4);
			}
			break;
	}

	/* Return. */
	if (smpsize>1){
		return(1);
	}
	return(0);
} /* FirstSimplify */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Forward clause simplification)  OCJ
 *
 *    This function performs the forward simplifications in a clause
 *    that are not performed by FirstSimplify() function:
 *    - Subsumption resolution
 *    - Other possible simplifications to be added in the future
 *
 *    The clause storage is freed in case of a NOMEMORY error condition.
 *
 *    This function just calls the appropriate functions that perform
 *    each simplification inference. Any required memory freeing is
 *    performed by this function, even in the case of error and also
 *    for the memory allocated to *pclause by the caller.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    pclause: Address of pointer to cmprefix structure of clause.
 *    nkb: KB where the text clauses are placed.
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> No simplifications were done.
 *    1 -> Simplifications were done.
 *    TIMEOUT -> Time has expired or other thread has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t SecondSimplify(cmprefix **pclause,int32_t nkb,int32_t type){
	auto cmprefix *ptr1;                                 /* Auxiliary pointer */
	auto int32_t ii;                                     /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Auxiliary */
	#endif

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile((*pclause)->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Perform forward subsumption resolution. */
	switch (ii=FwSubsResltn(*pclause,&ptr1,nkb,type)){

		/* Not enough memory or timeout. */
		case NOMEMORY:
		case TIMEOUT:
			FreeEquData((*pclause)->part2.bin);
			if ((*pclause)->part2.bin->asserts!=NULL){
				MYFREE((*pclause)->part2.bin->asserts);
			}
			MYFREE((*pclause)->part2.bin);
			MYFREE(*pclause);
			return(ii);
			break;

		/* The clause was simplified. */
		case 1:

			/* If original clause is not locked then add it in text form to KB. */
			if (0==((*pclause)->flags&LOCKED)){
				AddTxtFormula2KB(*pclause,(*pclause)->inference,nkb,SETFNUMBER);

			/* If original clause is locked then add it to LOCKED queue. */
			} else {
				(*pclause)->flags&=(~(UNPROC|PASSIVE|ACTIVE));
				if (NOMEMORY==AddBinClause2KB(*pclause,nkb,SETFNUMBER)){
					MYFREE((*pclause)->part2.bin->lockasserts);
					FreeEquData((*pclause)->part2.bin);
					if ((*pclause)->part2.bin->asserts!=NULL){
						MYFREE((*pclause)->part2.bin->asserts);
					}
					MYFREE((*pclause)->part2.bin);
					MYFREE(*pclause);
					return(NOMEMORY);
				}
			}

			/* Get simplified clause and account for this inference. */
			*pclause=ptr1;
			kbset[nkb]->prstats.fwsubsresltns++;

			/* If an empty clause has been generated then return as an empty */
			/* clause cannot be further simplified. */
			if (ptr1->part2.bin->formula[0]==UNITEND){
				return(0);
			}
			break;
	}

	return(ii);
} /* SecondSimplify */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Backward simplification)  OCJ
 *
 *    This function performs backward simplifications by a clause.
 *
 *    Also, if the function called gets an empty clause without asertions
 *    (as may be the case of subsumption resolution) then it must add
 *    the empty clause to the KB as UNPROC, set the frstprfnode field
 *    of the KB and return with an return code 1 such that this
 *    circumstance is recognized by this function.
 *
 *    This function just calls the appropriate functions that perform
 *    each simplification inference.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause.
 *    nkb: KB where the text clauses are placed.
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Process performed and no empty clause without assertions generated.
 *    1 -> Process performed and an empty clause without assertions was generated.
 *    TIMEOUT -> Time has expired or other thread has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t BckSimplify(cmprefix *clause,int32_t nkb,int32_t type){
	auto int32_t ii;                                     /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Auxiliary */
	#endif

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	#endif

	/* WARNING: Make sure the clause parameters are correctly updated */
	/* as appropriate by the functions that generate them before adding */
	/* them to the KB. */

	/* Perform backward demodulation. */
	switch (ii=((UNITEQUALITY|NONUNITEQUALITY)&pbmflags?BckDemodulation(clause,nkb,type):0)){

		/* Not enough memory or timeout. */
		case NOMEMORY:
		case TIMEOUT:
			return(ii);
			break;
	}

	/* Perform backward subsumption. */
	switch (ii=BckSubsuming(clause,nkb,type)){

		/* Not enough memory or timeout. */
		case NOMEMORY:
		case TIMEOUT:
			return(ii);
			break;
	}

	/* Perform backward subsumption resolution. */
	switch (ii=BkSubsResltn(clause,nkb,type)){

		/* Not enough memory, timeout or empty clause without assertions generated. */
		case NOMEMORY:
		case TIMEOUT:
		case 1:
			return(ii);
			break;
	}

	return(0);
} /* BckSimplify */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Factoring for resolution inferences)  OCJ
 *
 *    This function performs the factoring required for resolution
 *    inferences. It factors the given literal with all unifiable
 *    selected literals in the formula of the binprefix structure.
 *    The given literal must also be in the  formula of the binprefix
 *    structure.
 *
 *    This function allocates memory for the result. It is the caller
 *    responsibility to free that memory.
 *
 *
 *  ARGUMENTS:
 *
 *    binp: Pointer to binprefix structure to be factored.
 *    literal: Pointer to literal in binp formula.
 *    result: Address of pointer where the address of the resulting
 *            binprefix structure will be placed.
 *    factlit: Address of pointer where the address of factored
 *             literal will be placed.
 *    Subst: Pointer to the initial substitution of the resolution inference.
 *           It will be upgraded to a compatible factoring substitution.
 *    flag: If not zero then only the literals following the given literal
 *          are checked for factoring, as it is assumed that the previous
 *          literals have been already checked for resolution and factored
 *          if possible with the given literal. Also it means that the
 *          binp clause is the primary parent and also the primary clause
 *          in the substitution. Therefore if marker is zero it means that
 *          the binp clause is the secondary clause in the substitution.
 *
 *  RETURNS:
 *
 *    0 if no factors were found.
 *    1 if factors were found.
 *    NOMEMORY if not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t ResolFactor(binprefix *binp,uint8_t *literal,binprefix **result,
		uint8_t **factlit,subst *Subst,int32_t flag){
	auto cmprefix auxcl;                                 /* Auxiliary cmprefix structure. */
	auto uint8_t *ptr1,*ptr2,*ptr4;                      /* Auxiliary pointers */
	auto binprefix *ptr3;                                /* Auxiliary pointer */
	auto int32_t ii,jj,kk,nn,ss,rr;                      /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Auxiliary */
	#endif

	/* Allocate storage for result and copy original clause. */
	nn=binpsize+2*binp->size;
	if (NULL==(auxcl.part2.bin=mallocmn(nn))){
		return(NOMEMORY);
	}
	memcpy(auxcl.part2.bin,binp,binpsize+binp->size);
	ptr2=&auxcl.part2.bin->formula[0]+(literal-&binp->formula[0]);
	ii=GetLiteralNbr(auxcl.part2.bin,ptr2);

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile(auxcl.part2.bin);
	MYFREE(ptr10);
	#endif

	/* Loop through literals. */
	rr=0;
	for (ptr1=(flag?NextItem(ptr2,OVERSUBTERMS):&auxcl.part2.bin->formula[0]),
			ptr4=(flag?NextItem(literal,OVERSUBTERMS):NULL),jj=(flag?ii+1:0);
			ptr1[0]!=UNITEND;ptr1=(kk?ptr1:NextItem(ptr1,OVERSUBTERMS)),
			ptr4=(flag?NextItem(ptr4,OVERSUBTERMS):NULL),jj=(kk?jj:jj+1)){
		kk=0;

		/* Literal is selected and is not the primary literal. */
		if ((ii!=jj)&&(SELECTED&ptr1[0])){

			/* Remember substitution status. */
			if (Subst->buffer!=NULL){
				ss=Subst->substsize;
			} else {
				ss=-1;
			}

			/* Check unification of terms. */
			switch (Unify(ptr2,ptr1,Subst,flag?0:ITEM1SEC|ITEM2SEC)){

				/* Not enough memory. */
				case NOMEMORY:
					if (auxcl.part2.bin->asserts!=NULL){
						MYFREE(auxcl.part2.bin->asserts);
					}
					MYFREE(auxcl.part2.bin);
					return(NOMEMORY);
					break;

				/* Literals unify. */
				case 1:

					/* Delete ptr1 literal. */
					DeleteLiteral(auxcl.part2.bin,ptr1);

					/* Update pointers and literal counters and indicate */
					/* that there was a literal deletion in this iteration. */
					/* Also mark the factored literal if appropriate. */
					if (ii>jj){
						ii--;
						ptr2=GetLiteralPtr(&auxcl,ii);
					}
					kk=rr=1;
					break;

				/* Literals don't unify. Reset substitution to its previous state. */
				default:
					if (ss>0){
						Subst->substsize=ss;
						Subst->buffer[ss-1]=UNITEND;
					} else if (Subst->buffer!=NULL){
						Subst->substsize=1;
						Subst->buffer[0]=UNITEND;
					} else {
						Subst->substsize=0;
					}
					break;
			}
		}
	}

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile(auxcl.part2.bin);
	MYFREE(ptr10);
	#endif

	/* Reallocate result to exact size, set factlit and return. */
	if (NULL==(ptr3=reallocmn(auxcl.part2.bin,binpsize+auxcl.part2.bin->size))){
		if (auxcl.part2.bin->asserts!=NULL){
			MYFREE(auxcl.part2.bin->asserts);
		}
		MYFREE(auxcl.part2.bin);
		return(NOMEMORY);
	}
	*result=ptr3;
	ptr3->equflg=0;
	*factlit=&ptr3->formula[0]+(ptr2-&auxcl.part2.bin->formula[0]);
	return(rr);
} /* ResolFactor */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform forward demodulation)  OCJ
 *
 *    This function tries to simplify the given clause in compiled or binary
 *    format with demodulation inferences by clauses in the indicated KB
 *    by calling Demodulate() function for every simplifier candidate clause
 *    in the requested KB. It must also be specified the type of clauses that
 *    will be considered, either active and/or passive clauses.
 *
 *    Candidate clauses are selected by inspecting every term in the
 *    argument clause and selecting clauses with a generalization term that
 *    is a maximum equality root term in a single literal clause.
 *
 *    The clause storage is freed in case of a NOMEMORY error condition.
 *
 *
 *  ARGUMENTS:
 *
 *    pclause: Address of pointer of clause to be simplified. The pointer to
 *             the simplified clause will be placed here on exit.
 *    nkb: KB index of KB that will be searched for candidate simplifying
 *         clauses
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Clause was not simplified.
 *    1 -> Clause was simplified.
 *    TIMEOUT -> Time has expired or other thread has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t FwdDemodulation(cmprefix **pclause,int32_t nkb,int32_t type){
	auto cmprefix *clause;                               /* Pointer to clause to be simplified */
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto dscstack *stack;                                /* Pointer to stack for sequential calls to QueryGeneric() */
	auto subst Subst;                                    /* Substitution for calls to Demodulate() */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr3;                      /* Auxiliary pointers */
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	auto int32_t ii,kk,oo;                               /* Auxiliary */

	/* Initialize clause pointer and substitution for Generalize() calls. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Renumber variables. This is necessary in order for SetEquData() */
	/* and CompareEquTerms() to generate the correct weight polynomials. */
	/* As the clause has been just inferred it may have shifted variable */
	/* numbers. */
	clause=*pclause;
	if (NOMEMORY==RenumberVars(clause->part2.bin)){
		FreeEquData(clause->part2.bin);
		if (clause->part2.bin->asserts!=NULL){
			MYFREE(clause->part2.bin->asserts);
		}
		MYFREE(clause->part2.bin);
		MYFREE(clause);
		return(NOMEMORY);
	}

	/* Loop through clause items. */
	stack=NULL;
	for (ptr1=NextItem(&clause->part2.bin->formula[0],IMMED);
			ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,IMMED)){

		/* Iterate if item is not a function. */
		if (0==(FUNCTION&ptr1[0])){
			continue;
		}

		/* Loop through ACTIVE and PASSIVE clauses as appropriate. */
		for (kk=ACTIVE;kk!=-1;kk=((kk==ACTIVE)&&(type&PASSIVE)?PASSIVE:-1)){

			/* Get pointer to top tree node. */
			switch (kk){
				case ACTIVE:
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr4[nkb];
					break;
				case PASSIVE:
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr5[nkb];
					break;
			}

			/* Loop through generalization candidates. */
			if (stack!=NULL){
				stack->used=0;
			}
			do {

				/* Check timeout and solution by other thread. */
				if (nkb!=num_wkbs){
					gettimeofday(&timestr,NULL);
					time=timestr.tv_sec-kbset[nkb]->time.tv_sec
							+(timestr.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0;
		 			if ((time>=kbset[nkb]->opts.timeout)||(procctl.status==UNSATISFIABLE)
		 					||(procctl.status==SATISFIABLE)){
						FreeEquData(clause->part2.bin);
						if (clause->part2.bin->asserts!=NULL){
							MYFREE(clause->part2.bin->asserts);
						}
						MYFREE(clause->part2.bin);
						MYFREE(clause);
						if (stack!=NULL){
							MYFREE(stack);
						}
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						return(TIMEOUT);
					}
		 			if ((procctl.status==NOMEMORY)||(procctl.status==TIMEOUT)){
						FreeEquData(clause->part2.bin);
						if (clause->part2.bin->asserts!=NULL){
							MYFREE(clause->part2.bin->asserts);
						}
						MYFREE(clause->part2.bin);
						MYFREE(clause);
						if (stack!=NULL){
							MYFREE(stack);
						}
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						return(procctl.status);
					}
				}

				/* Get a term in a candidate simplifying clause. */
				switch (ii=QueryGeneric(NextItem(ptr1,IMMED),&ptr2,&stack,0,ptr3)){

					/* Not enough memory. */
					case NOMEMORY:
						FreeEquData(clause->part2.bin);
						if (clause->part2.bin->asserts!=NULL){
							MYFREE(clause->part2.bin->asserts);
						}
						MYFREE(clause->part2.bin);
						MYFREE(clause);
						if (stack!=NULL){
							MYFREE(stack);
						}
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						return(NOMEMORY);
						break;

					/* We have a valid candidate. */
					case 1:

						/* Check generalization. */
						switch (Generalize(ptr2,ptr1,&Subst,ITEM2SEC|INITSUBST)){

							/* Not enough memory. */
							case NOMEMORY:
								FreeEquData(clause->part2.bin);
								if (clause->part2.bin->asserts!=NULL){
									MYFREE(clause->part2.bin->asserts);
								}
								MYFREE(clause->part2.bin);
								MYFREE(clause);
								if (stack!=NULL){
									MYFREE(stack);
								}
								if (Subst.buffer!=NULL){
									MYFREE(Subst.buffer);
								}
								return(NOMEMORY);
								break;

							/* Valid generalization. */
							case 1:

								/* Get pointer to candidate binprefix structure. */
								binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

								/* Debug. */
								#ifdef DEBUGCODE
								ptr10=Decompile(binp);
								MYFREE(ptr10);
								#endif

								/* Perform demodulation and check if it is successful. */
								if (1==(oo=Demodulate(clause,ptr1,binp,ptr2,pclause,&Subst,nkb))){

									/* Set locks of simplified clause and asserts of demodulation result. */
									if (NOMEMORY==GetLocksAsserts(binp,clause->part2.bin,(*pclause)->part2.bin)){
										FreeEquData(clause->part2.bin);
										if (clause->part2.bin->asserts!=NULL){
											MYFREE(clause->part2.bin->asserts);
										}
										MYFREE(clause->part2.bin);
										MYFREE(clause);
										if ((*pclause)->part2.bin->asserts!=NULL){
											MYFREE((*pclause)->part2.bin->asserts);
										}
										MYFREE((*pclause)->part2.bin);
										MYFREE((*pclause));
										if (stack!=NULL){
											MYFREE(stack);
										}
										if (Subst.buffer!=NULL){
											MYFREE(Subst.buffer);
										}
										return(NOMEMORY);
									}

									/* Set type of inference, flags and dist2conj in new clause. */
									(*pclause)->inference=FWDEMODULATION;
									(*pclause)->flags=UNPROC;
									(*pclause)->part2.bin->dist2conj=clause->part2.bin->dist2conj;

									/* If original clause is not locked then convert it to text form. */
									if (0==(LOCKED&clause->flags)){
										if (NULL==(ptr10=Decompile(clause->part2.bin))){
											FreeEquData(clause->part2.bin);
											if (clause->part2.bin->asserts!=NULL){
												MYFREE(clause->part2.bin->asserts);
											}
											MYFREE(clause->part2.bin);
											MYFREE(clause);
											if ((*pclause)->part2.bin->asserts!=NULL){
												MYFREE((*pclause)->part2.bin->asserts);
											}
											MYFREE((*pclause)->part2.bin);
											MYFREE((*pclause));
											if (stack!=NULL){
												MYFREE(stack);
											}
											if (Subst.buffer!=NULL){
												MYFREE(Subst.buffer);
											}
											return(NOMEMORY);
										}
										FreeEquData(clause->part2.bin);
										if (clause->part2.bin->asserts!=NULL){
											MYFREE(clause->part2.bin->asserts);
										}
										MYFREE(clause->part2.bin);
										clause->part2.text=ptr10;
									}

									/* Debug. */
									#ifdef DEBUGCODE
									ptr10=Decompile(binp);
									#ifdef VERBOSE
									if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
										printf("Forward demodulated from:\n%s\n",ptr10);
									}
									#endif
									MYFREE(ptr10);
									ptr10=Decompile((*pclause)->part2.bin);
									#ifdef VERBOSE
									if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
										printf("Result [%d]:\n%s\n",kbset[0]->nformulas,ptr10);
									}
									#endif
									MYFREE(ptr10);
									#endif
								}

								/* If clause was simplified or we got a NOMEMORY condition */
								/* then free memory and return. */
								if (oo!=0){
									switch (oo){

										/* Not enough memory or timeout. */
										case NOMEMORY:
										case TIMEOUT:
											FreeEquData(clause->part2.bin);
											if (clause->part2.bin->asserts!=NULL){
												MYFREE(clause->part2.bin->asserts);
											}
											MYFREE(clause->part2.bin);
											MYFREE(clause);
											return(oo);
											break;

										/* The clause was simplified. */
										case 1:

											/* If original clause is not locked then add it in */
											/* text form to KB. */
											if (0==(clause->flags&LOCKED)){
												AddTxtFormula2KB(clause,clause->inference,nkb,SETFNUMBER);

											/* If original clause is locked then add it to LOCKED queue. */
											} else {
												clause->flags&=(~(UNPROC|PASSIVE|ACTIVE));
												if (NOMEMORY==AddBinClause2KB(clause,nkb,SETFNUMBER)){
													MYFREE(clause->part2.bin->lockasserts);
													FreeEquData(clause->part2.bin);
													if (clause->part2.bin->asserts!=NULL){
														MYFREE(clause->part2.bin->asserts);
													}
													MYFREE(clause->part2.bin);
													MYFREE(clause);
													return(NOMEMORY);
												}
											}

											/* Account for this inference. */
											kbset[nkb]->prstats.fwdemodulations++;
											break;
									}
									if (stack!=NULL){
										MYFREE(stack);
									}
									if (Subst.buffer!=NULL){
										MYFREE(Subst.buffer);
									}
									return(oo);
								}
								break;
						}
						break;
				}
			} while(ii);
		}
	}

	/* Restore pointer to original clause, free memory and return clause not simplified. */
	*pclause=clause;
	if (stack!=NULL){
		MYFREE(stack);
	}
	if (Subst.buffer!=NULL){
		MYFREE(Subst.buffer);
	}
	return(0);
} /* FwdDemodulation */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform backward demodulation)  OCJ
 *
 *    This function tries to simplify the active and/or passive clauses
 *    in the indicated KB by the given clause in compiled or binary format
 *    with demodulation inferences by calling Demodulate() function for
 *    every candidate clause for simplification in the requested KB. It
 *    must also be specified the type of clauses that will be simplified,
 *    either active and/or passive clauses.
 *
 *    The given clause must be suitable for performing demodulation: It must
 *    be a single literal positive equality with a maximum equality root term.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in compiled binary form.
 *    nkb: KB index of KB that will be searched for clauses that are
 *         candidates to be simplified.
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Process performed and no empty clause without assertions generated.
 *    1 -> Process performed and an empty clause without assertions was generated.
 *         This function never returns 1 but this is documented here as a possible
 *         return code from functions called by BckSimplify() function.
 *    TIMEOUT -> Time has expired or other thread has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t BckDemodulation(cmprefix *clause,int32_t nkb,int32_t type){
	auto cmprefix *newcl;                                /* Pointer to simplified clause cmprefix structure */
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto dscstack *stack;                                /* Pointer to stack for sequential calls to QueryGeneric() */
	auto subst Subst;                                    /* Substitution for calls to Demodulate() */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *equality;                              /* Pointer to equality literal in clause */
	auto uint8_t *ptr1,*ptr2,*ptr3;                      /* Auxiliary pointers */
	auto int32_t ii,jj,kk,oo;                            /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Check if clause is suitable for performing demodulation from it. */
	if ((clause->part2.bin->literals>1)||(EQUALITY!=((EQUALITY|NEGATED)&clause->part2.bin->formula[0]))){
		return(0);
	} else {
		equality=&clause->part2.bin->formula[0];
	}

	/* Get pointers to equality root terms and reset MAXIMUM */
	/* and SELECTED flags in each of them. */
	ptr1=NextItem(equality,IMMED);
	ptr2=NextItem(ptr1,OVERSUBTERMS);
	ptr1[0]&=(~(MAXIMUM|SELECTED));
	ptr2[0]&=(~(MAXIMUM|SELECTED));

	/* Renumber variables. This is necessary in order for SetEquData() */
	/* and CompareEquTerms() to generate the correct weight polynomials. */
	/* As the clause has been just inferred it may have shifted variable */
	/* numbers. */
	if (NOMEMORY==RenumberVars(clause->part2.bin)){
		return(NOMEMORY);
	}

	/* Calculate equality ordering data and remember that */
	/* this was done for a potentially demodulating clause. */
	/* Equality ordering data may be already calculated because */
	/* a clause can go several times as unprocessed due to being */
	/* previously locked. */
	if (clause->part2.bin->equflg==0){
		if (NOMEMORY==SetEquData(clause->part2.bin,equality,ptr1,ptr2,nkb)){
			return(NOMEMORY);
		}
		clause->part2.bin->equflg=2;
	}

	/* Set the SELECTED and MAXIMUM flags for equality root terms */
	/* as appropriate. Maximal equality root terms are also */
	/* candidates for demodulating from them as they may become */
	/* maximum after the substitution. If both terms are equal */
	/* then demodulation is not possible as there is no possible */
	/* maximum term even after a substitution. */
	/* Release equality tree data if the root terms ordering */
	/* is completely defined. */
	switch (CompareEquTerms(clause->part2.bin,ptr1,ptr2,&((symbol *)&equality[1])->leftnode,
			&((symbol *)&equality[1])->rightnode,kbset[nkb]->opts.termord,nkb)){
		case NOMEMORY:
			return(NOMEMORY);
			break;
		case EQUAL:
			FreeEquNode(ptr1,((symbol *)&equality[1])->leftnode);
			FreeEquNode(ptr2,((symbol *)&equality[1])->rightnode);
			((symbol *)&equality[1])->leftnode=NULL;
			return(0);
			break;
		case FAILURE:
			ptr1[0]|=SELECTED;
			ptr2[0]|=SELECTED;
			return(0);
			break;
		case GREATER:
			ptr1[0]|=(SELECTED|MAXIMUM);
			FreeEquNode(ptr1,((symbol *)&equality[1])->leftnode);
			FreeEquNode(ptr2,((symbol *)&equality[1])->rightnode);
			((symbol *)&equality[1])->leftnode=NULL;
			break;
		case LESSER:
			ptr2[0]|=(SELECTED|MAXIMUM);
			FreeEquNode(ptr1,((symbol *)&equality[1])->leftnode);
			FreeEquNode(ptr2,((symbol *)&equality[1])->rightnode);
			((symbol *)&equality[1])->leftnode=NULL;
			break;
	}

	/* Initialize substitution for Generalize() calls. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Loop through equality root terms. */
	for (ptr1=NextItem(equality,IMMED),jj=0;jj<2;ptr1=NextItem(ptr1,OVERSUBTERMS),jj++){

		/* Iterate if the term is not selected (that is, maximal) */
		/* or it is a variable. Variables are discarded because if */
		/* they are also in the other root term then the variable */
		/* can never be maximum even after a substitution. Also, */
		/* if the variable is not part of the other root term the */
		/* equality makes no sense as it that can be unified with */
		/* everything). */
		if ((0==(SELECTED&ptr1[0]))||(VARIABLE&ptr1[0])){
			continue;
		}

		/* Loop through ACTIVE and PASSIVE clauses as appropriate. */
		stack=NULL;
		for (kk=ACTIVE;kk!=-1;kk=((kk==ACTIVE)&&(type&PASSIVE)?PASSIVE:-1)){

			/* Get pointer to top tree node. */
			switch (kk){
				case ACTIVE:
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr3[nkb];
					break;
				case PASSIVE:
					ptr3=(((symbol *)&ptr1[1])->symbol)->discr6[nkb];
					break;
			}

			/* Loop through candidates that are instances of maximum equality root term. */
			if (stack!=NULL){
				stack->used=0;
			}
			do {

				/* Check timeout and solution by other thread. */
				if (nkb!=num_wkbs){
					gettimeofday(&timestr,NULL);
					time=timestr.tv_sec-kbset[nkb]->time.tv_sec
							+(timestr.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0;
		 			if ((time>=kbset[nkb]->opts.timeout)||(procctl.status==UNSATISFIABLE)
		 					||(procctl.status==SATISFIABLE)){
						if (stack!=NULL){
							MYFREE(stack);
						}
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						return(TIMEOUT);
					}
		 			if ((procctl.status==NOMEMORY)||(procctl.status==TIMEOUT)){
						if (stack!=NULL){
							MYFREE(stack);
						}
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						return(procctl.status);
					}
				}

				/* Get a term in a candidate clause to be simplified. */
				switch (ii=QueryUI(NextItem(ptr1,IMMED),&ptr2,&stack,0,ptr3,INSTANCE)){

					/* Not enough memory. */
					case NOMEMORY:
						if (stack!=NULL){
							MYFREE(stack);
						}
						if (Subst.buffer!=NULL){
							MYFREE(Subst.buffer);
						}
						return(NOMEMORY);
						break;

					/* We have a valid candidate. */
					case 1:

						/* Get pointer to binprefix structure of clause to be simplified. */
						binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

						/* Clause is not in disposal queue therefore */
						/* it has not been already simplified. */
						if (0==(binp->clause->flags&INDISPOSALQUEUE)){

							/* Check generalization. */
							switch (Generalize(ptr1,ptr2,&Subst,ITEM2SEC|INITSUBST)){

								/* Not enough memory. */
								case NOMEMORY:
									if (stack!=NULL){
										MYFREE(stack);
									}
									if (Subst.buffer!=NULL){
										MYFREE(Subst.buffer);
									}
									return(NOMEMORY);
									break;

								/* Valid generalization. */
								case 1:

									/* Debug. */
									#ifdef DEBUGCODE
									ptr10=Decompile(binp);
									MYFREE(ptr10);
									#endif

									/* Perform demodulation. */
									if (NOMEMORY==(oo=Demodulate(binp->clause,ptr2,clause->part2.bin,
											ptr1,&newcl,&Subst,nkb))){
										if (stack!=NULL){
											MYFREE(stack);
										}
										if (Subst.buffer!=NULL){
											MYFREE(Subst.buffer);
										}
										return(NOMEMORY);
									}

									/* Successful simplification. */
									if (oo){

										/* Set distance to conjecture of new simplified clause. */
										newcl->part2.bin->dist2conj=binp->dist2conj;

										/* Set locks of simplified clause and asserts of demodulation result. */
										if (NOMEMORY==GetLocksAsserts(clause->part2.bin,binp,newcl->part2.bin)){
											if (newcl->part2.bin->asserts!=NULL){
												MYFREE(newcl->part2.bin->asserts);
											}
											MYFREE(newcl->part2.bin);
											MYFREE(newcl);
											if (stack!=NULL){
												MYFREE(stack);
											}
											if (Subst.buffer!=NULL){
												MYFREE(Subst.buffer);
											}
											return(NOMEMORY);
										}

										/* Debug. */
										#ifdef DEBUGCODE
										ptr10=Decompile(binp);
										#ifdef VERBOSE
										if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
											printf("Backward demodulating to:\n%s\n",ptr10);
										}
										#endif
										MYFREE(ptr10);
										ptr10=Decompile(newcl->part2.bin);
										#ifdef VERBOSE
										if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
											printf("Result [%d]:\n%s\n",kbset[0]->nformulas,ptr10);
										}
										#endif
										MYFREE(ptr10);
										#endif

										/* Add original simplified clause to disposal queue. This queue interconnects */
										/* clauses from other queues. The clause must not be removed from any queue */
										/* to which it is linked. */
										binp->clause->flags|=INDISPOSALQUEUE;
										binp->clause->prevdisp=kbset[nkb]->lastdisp;
										kbset[nkb]->lastdisp=binp->clause;
										kbset[nkb]->prstats.bkdemodulations++;

										/* Add new simplified clause to KB as UNPROC. A new number is not assigned. */
										/* It will be assigned when moving this clause to passive, so that it */
										/* has a higher number than its parents. */
										newcl->flags=UNPROC;
										newcl->inference=BKDEMODULATION;
										if (NOMEMORY==AddBinClause2KB(newcl,nkb,0)){
											if (newcl->part2.bin->asserts!=NULL){
												MYFREE(newcl->part2.bin->asserts);
											}
											MYFREE(newcl->part2.bin);
											MYFREE(newcl);
											if (stack!=NULL){
												MYFREE(stack);
											}
											if (Subst.buffer!=NULL){
												MYFREE(Subst.buffer);
											}
											return(NOMEMORY);
										}
									}
									break;
							}
						}
						break;
				}
			} while(ii);
		}
	}

	/* Free memory and return. */
	if (stack!=NULL){
		MYFREE(stack);
	}
	if (Subst.buffer!=NULL){
		MYFREE(Subst.buffer);
	}
	return(0);
} /* BckDemodulation */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Demodulate a clause)  OCJ
 *
 *    This function demodulates a clause in internal compiled binary
 *    format. The new clause is placed in new allocated storage and
 *    its parents are set, but no other data is included in the
 *    clause cmprefix.
 *
 *    It is the caller responsibility to link and/or unlink the clauses
 *    and set the other prefix fields.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause to be simplified
 *            by demodulation.
 *    term1: Pointer to term that will be replaced in clause.
 *    binp: Pointer to binprefix structure of clause from where
 *          demodulation is performed.
 *    term2: Pointer to the term in "from" clause that is a generalization
 *           of term1.
 *    pnewcl: Address of pointer to cmprefix structure of new clause to be
 *            allocated for the result of demodulation.
 *    Subst: Pointer to substitution to be applied to term2.
 *    nkb: KB index of KB that owns the process.
 *
 *  RETURNS:
 *
 *    0 -> The clause was not simplified because ordering constraints
 *         were not met.
 *    1 -> The clause was successfully simplified.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t Demodulate(cmprefix *clause,uint8_t *term1,binprefix *binp,
		uint8_t *term2,cmprefix **pnewcl,subst *Subst,int32_t nkb){
	auto uint8_t *equality;                              /* Pointer to equality literal in binp */
	auto binprefix *binp2;                               /* Auxiliary binary prefix pointer */
	auto uint8_t *ptr1,*ptr2;                            /* Auxiliary pointers */
	auto int32_t ii,jj,kk,nn;                            /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                     /* Auxiliary pointer */

	/* Debug. */
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	ptr10=Decompile(binp);
	MYFREE(ptr10);
	#endif

	/* Get pointer to equality from which demodulation is performed. */
	equality=&binp->formula[0];

	/* Set ptr1 to the other root term in the "from" equality. If Next item */
	/* of term2 is the end of clause then it is the second root term in the */
	/* "from" equality. Otherwise term2 is the first root term in the "from" */
	/* equality. Take into account that the "from" clause has only one literal. */
	ptr1=NextItem(term2,OVERSUBTERMS);
	if (ptr1[0]&UNITEND){
		ptr1=NextItem(equality,IMMED);
	}

	/* Check ordering constraint of root terms in equality */
	/* "from" literal. */
	switch (ii=CheckEqRootOrder(binp,equality,term2,ptr1,Subst,binp->maxvarnb+1,nkb,MAXIMUM)){
		case NOMEMORY:
		case 0:
			return(ii);
			break;
	}

	/* Allocate storage for inferred clause. */
	if (NULL==(*pnewcl=mallocmn(sizeof(cmprefix)))){
		return(NOMEMORY);
	}
	nn=binpsize+2*(clause->part2.bin->size+binp->size);
	if (NULL==(((*pnewcl)->part2.bin)=mallocmn(nn))){
		MYFREE(*pnewcl);
		return(NOMEMORY);
	}
	(*pnewcl)->part2.bin->clause=*pnewcl;
	(*pnewcl)->part2.bin->equflg=0;

	/* Copy formula of "to" parent clause up to and not including term1. */
	ii=term1-&clause->part2.bin->formula[0];
	memcpy(&(*pnewcl)->part2.bin->formula[0],&clause->part2.bin->formula[0],ii);

	/* Shift the numbers of variables of copied items. */
	kk=binp->maxvarnb+1;
	for (ptr2=NextItem(&(*pnewcl)->part2.bin->formula[0],IMMED);
			ptr2<&(*pnewcl)->part2.bin->formula[ii];
			ptr2=NextItem(ptr2,IMMED)){
		if (VARIABLE&ptr2[0]){
			*((uint16_t *)&ptr2[1])+=kk;
		}
	}

	/* Add the other equality root term of clause from */
	/* where demodulation is performed. */
	jj=NextItem(ptr1,OVERSUBTERMS)-ptr1;
	memcpy(&(*pnewcl)->part2.bin->formula[ii],ptr1,jj);
	ii+=jj;

	/* Add remaining items of clause being simplified */
	/* including the UNITEND marker byte. */
	ptr1=NextItem(term1,OVERSUBTERMS);
	jj=clause->part2.bin->size-(ptr1-&clause->part2.bin->formula[0]);
	memcpy(&(*pnewcl)->part2.bin->formula[ii],ptr1,jj);

	/* Shift the numbers of variables of last copied items. */
	for (ptr1=&(*pnewcl)->part2.bin->formula[ii];ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,IMMED)){
		if (VARIABLE&ptr1[0]){
			*((uint16_t *)&ptr1[1])+=kk;
		}
	}

	/* Apply substitution to inferred clause. Binary clause */
	/* size field is set by this function, so there is no need */
	/* to call UpdateSize() function here. In case of NOMEMORY */
	/* there is no need to free asserts as they are not set yet. */
	if (NOMEMORY==ApplySubst2Clause(*pnewcl,&nn,Subst,kk)){
		MYFREE((*pnewcl)->part2.bin);
		MYFREE(*pnewcl);
		return(NOMEMORY);
	}

	/* Reallocate the formula buffer to the exact length needed. */
	if (NULL==(binp2=reallocmn((*pnewcl)->part2.bin,binpsize+(*pnewcl)->part2.bin->size))){
		if ((*pnewcl)->part2.bin->asserts!=NULL){
			MYFREE((*pnewcl)->part2.bin->asserts);
		}
		MYFREE((*pnewcl)->part2.bin);
		MYFREE(*pnewcl);
		return(NOMEMORY);
	}
	(*pnewcl)->part2.bin=binp2;

	/* Debug. */
	#ifdef DEBUGCODE
	(*pnewcl)->part2.bin->asserts=NULL;
	ptr10=Decompile((*pnewcl)->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Set simplified clause cmprefix data. */
	(*pnewcl)->parent1=binp->clause;
	(*pnewcl)->parent2=clause;

	return(1);
} /* Demodulate */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform forward subsumption resolution)  OCJ
 *
 *    This function tries to simplify the given clause in compiled or binary
 *    format with subsumption resolution inferences by clauses in the indicated
 *    KB by calling SubsResltn() function for every simplifier candidate clause
 *    in the requested KB. It must also be specified the type of clauses that
 *    will be considered, either active and/or passive clauses.
 *
 *    Simplifier candidate clauses are selected from those that have a literal
 *    whose negation is a generalization of a literal in the clause to be
 *    simplified.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in compiled binary form.
 *    pnewcl: Address of pointer where the simplified clause will be placed.
 *    nkb: KB index of KB that will be searched for a subsuming clause
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Clause was not simplified.
 *    1 -> Clause was simplified.
 *    TIMEOUT -> Time has expired or other thread has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t FwSubsResltn(cmprefix *clause,cmprefix **pnewcl,int32_t nkb,int32_t type){
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto dscstack *stack;                                /* Pointer to stack for sequential calls to QueryGeneric() */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr3;                      /* Auxiliary pointers */
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	auto int32_t ii,kk;                                  /* Auxiliary */

	/* Loop through clause literals. */
	stack=NULL;
	for (ptr1=&clause->part2.bin->formula[0];ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS)){

		/* Loop through ACTIVE and PASSIVE clauses. ACTIVE clauses have */
		/* selected and non selected literals. */
		for (kk=ACTIVE;kk!=-1;
				kk=(kk==ACTIVE?ACTIVE|SELECTED:(kk==PASSIVE?-1:(type&PASSIVE?PASSIVE:-1)))){

			/* Get pointer to top tree node. */
			switch (kk){
				case ACTIVE:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr3[nkb];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr4[nkb];
					}
					break;
				case PASSIVE:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr5[nkb];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr6[nkb];
					}
					break;
				default:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr1[nkb];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr2[nkb];
					}
					break;
			}

			/* Loop through generalization candidates. */
			if (stack!=NULL){
				stack->used=0;
			}
			do {

				/* Check timeout and solution by other thread. */
				gettimeofday(&timestr,NULL);
				time=timestr.tv_sec-kbset[nkb]->time.tv_sec
						+(timestr.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0;
	 			if ((time>=kbset[nkb]->opts.timeout)||(procctl.status==UNSATISFIABLE)
	 					||(procctl.status==SATISFIABLE)){
					if (stack!=NULL){
						MYFREE(stack);
					}
					return(TIMEOUT);
				}
	 			if ((procctl.status==NOMEMORY)||(procctl.status==TIMEOUT)){
					if (stack!=NULL){
						MYFREE(stack);
					}
					return(procctl.status);
				}

				/* Get a candidate negative generalization literal. */
				switch (ii=QueryGeneric(NextItem(ptr1,IMMED),&ptr2,&stack,0,ptr3)){

					/* Not enough memory. */
					case NOMEMORY:
						if (stack!=NULL){
							MYFREE(stack);
						}
						return(NOMEMORY);
						break;

					/* We have a valid candidate. */
					case 1:

						/* Get pointer to candidate binprefix structure. */
						binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

						/* Debug. */
						#ifdef DEBUGCODE
						ptr10=Decompile(binp);
						MYFREE(ptr10);
						#endif

						/* Perform subsumption resolution and check if it is successful. */
						switch (SubsResltn(clause,ptr1,binp,ptr2,pnewcl)){

							/* Not enough memory. */
							case NOMEMORY:
								if (stack!=NULL){
									MYFREE(stack);
								}
								return(NOMEMORY);
								break;

							/* Successful simplification. */
							case 1:

								/* Set locks of simplified clause and asserts of demodulation result. */
								if (NOMEMORY==GetLocksAsserts(binp,clause->part2.bin,(*pnewcl)->part2.bin)){
									if ((*pnewcl)->part2.bin->asserts!=NULL){
										MYFREE((*pnewcl)->part2.bin->asserts);
									}
									MYFREE((*pnewcl)->part2.bin);
									MYFREE((*pnewcl));
									if (stack!=NULL){
										MYFREE(stack);
									}
									return(NOMEMORY);
								}

								/* Set type of inference, flags and dist2conj in new clause. */
								(*pnewcl)->inference=FWSUBSRESLTNS;
								(*pnewcl)->flags=UNPROC;
								(*pnewcl)->part2.bin->dist2conj=clause->part2.bin->dist2conj;

								/* If original clause is not locked then convert it to text form. */
								if (0==(LOCKED&clause->flags)){
									if (NULL==(ptr10=Decompile(clause->part2.bin))){
										if ((*pnewcl)->part2.bin->asserts!=NULL){
											MYFREE((*pnewcl)->part2.bin->asserts);
										}
										MYFREE((*pnewcl)->part2.bin);
										MYFREE((*pnewcl));
										if (stack!=NULL){
											MYFREE(stack);
										}
										return(NOMEMORY);
									}
									FreeEquData(clause->part2.bin);
									if (clause->part2.bin->asserts!=NULL){
										MYFREE(clause->part2.bin->asserts);
									}
									MYFREE(clause->part2.bin);
									clause->part2.text=ptr10;
								}

								/* Debug. */
								#ifdef DEBUGCODE
								ptr10=Decompile(binp);
								#ifdef VERBOSE
								if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
									printf("Subsumption resolution simplification by:\n%s\n",ptr10);
								}
								#endif
								MYFREE(ptr10);
								ptr10=Decompile((*pnewcl)->part2.bin);
								#ifdef VERBOSE
								if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
									printf("Result [%d]:\n%s\n",kbset[0]->nformulas,ptr10);
								}
								#endif
								MYFREE(ptr10);
								#endif
								if (stack!=NULL){
									MYFREE(stack);
								}
								return(1);
								break;
						}
						break;
				}
			} while(ii);
		}
	}

	/* Free memory and return unsuccessful simplification. */
	if (stack!=NULL){
		MYFREE(stack);
	}
	return(0);
} /* FwSubsResltn */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform backward subsumption resolution)  OCJ
 *
 *    This function tries to simplify clauses in KB by the given clause in
 *    compiled or binary format with subsumption resolution inferences by
 *    calling SubsResltn() function for every simplifier candidate clause
 *    in the requested KB. It must also be specified the type of clauses that
 *    will be considered, either active and/or passive clauses.
 *
 *    Clauses candidate to be simplified are selected from those that have
 *    a literal whose negation is an instance of a literal in the clause
 *    to be simplified.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in compiled binary form.
 *    nkb: KB index of KB that will be searched for a subsuming clause
 *    type: PASSIVE if passive clauses are to be checked for
 *          simplifications. Active clauses are always checked.
 *          Any other flag is ignored.
 *
 *  RETURNS:
 *
 *    0 -> Process performed and no empty clause without assertions generated.
 *    1 -> Process performed and an empty clause without assertions was generated.
 *    TIMEOUT -> Time has expired or other thread has solved the problem.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t BkSubsResltn(cmprefix *clause,int32_t nkb,int32_t type){
	auto cmprefix *newcl;                                /* Pointer to simplified clause cmprefix structure */
	auto binprefix *binp;                                /* Pointer to candidate binprefix structure */
	auto dscstack *stack;                                /* Pointer to stack for sequential calls to QueryGeneric() */
	auto double time;                                    /* Elapsed time in seconds */
	auto struct timeval timestr;                         /* Structure for time measurement */
	auto uint8_t *ptr1,*ptr2,*ptr3;                      /* Auxiliary pointers */
	auto int32_t ii,kk;                                  /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Loop through clause literals. */
	stack=NULL;
	for (ptr1=&clause->part2.bin->formula[0];ptr1[0]!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS)){

		/* Loop through ACTIVE and PASSIVE clauses. ACTIVE clauses have */
		/* selected and non selected literals. */
		for (kk=ACTIVE;kk!=-1;
				kk=(kk==ACTIVE?ACTIVE|SELECTED:(kk==PASSIVE?-1:(type&PASSIVE?PASSIVE:-1)))){

			/* Get pointer to top tree node. */
			switch (kk){
				case ACTIVE:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr3[nkb];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr4[nkb];
					}
					break;
				case PASSIVE:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr5[nkb];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr6[nkb];
					}
					break;
				default:
					if (ptr1[0]&NEGATED){
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr1[nkb];
					} else {
						ptr3=(((symbol *)&ptr1[1])->symbol)->discr2[nkb];
					}
					break;
			}

			/* Loop through candidate instances. */
			if (stack!=NULL){
				stack->used=0;
			}
			do {

				/* Check timeout and solution by other thread. */
				gettimeofday(&timestr,NULL);
				time=timestr.tv_sec-kbset[nkb]->time.tv_sec
						+(timestr.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0;
	 			if ((time>=kbset[nkb]->opts.timeout)||(procctl.status==UNSATISFIABLE)
	 					||(procctl.status==SATISFIABLE)){
					if (stack!=NULL){
						MYFREE(stack);
					}
					return(TIMEOUT);
				}
	 			if ((procctl.status==NOMEMORY)||(procctl.status==TIMEOUT)){
					if (stack!=NULL){
						MYFREE(stack);
					}
					return(procctl.status);
				}

				/* Get a candidate negative instance literal. */
				switch (ii=QueryUI(NextItem(ptr1,IMMED),&ptr2,&stack,0,ptr3,UNIFICATION)){

					/* Not enough memory. */
					case NOMEMORY:
						if (stack!=NULL){
							MYFREE(stack);
						}
						return(NOMEMORY);
						break;

					/* We have a valid candidate. */
					case 1:

						/* Get pointer to candidate binprefix structure. */
						binp=(binprefix *)(ptr2-(binpsize+((symbol *)&ptr2[1])->offset));

						/* Debug. */
						#ifdef DEBUGCODE
						ptr10=Decompile(binp);
						MYFREE(ptr10);
						#endif

						/* Clause is not in disposal queue therefore */
						/* it has not been already simplified. */
						if (0==(binp->clause->flags&INDISPOSALQUEUE)){

							/* Perform subsumption resolution and check if it is successful. */
							switch (SubsResltn(binp->clause,ptr2,clause->part2.bin,ptr1,&newcl)){

								/* Not enough memory. */
								case NOMEMORY:
									if (stack!=NULL){
										MYFREE(stack);
									}
									return(NOMEMORY);
									break;

								/* Successful simplification. */
								case 1:

									/* Set distance to conjecture of new simplified clause. */
									newcl->part2.bin->dist2conj=binp->dist2conj;

									/* Set locks of simplified clause and asserts of simplification result. */
									if (NOMEMORY==GetLocksAsserts(clause->part2.bin,binp,newcl->part2.bin)){
										if (newcl->part2.bin->asserts!=NULL){
											MYFREE(newcl->part2.bin->asserts);
										}
										MYFREE(newcl->part2.bin);
										MYFREE(newcl);
										if (stack!=NULL){
											MYFREE(stack);
										}
										return(NOMEMORY);
									}

									/* Debug. */
									#ifdef DEBUGCODE
									ptr10=Decompile(binp);
									#ifdef VERBOSE
									if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
										printf("Backward subsumption resolution simplification to:\n%s\n",ptr10);
									}
									#endif
									MYFREE(ptr10);
									ptr10=Decompile(newcl->part2.bin);
									#ifdef VERBOSE
									if ((num_wkbs==1)&&(verbfrom<=kbset[0]->nformulas)&&(verbto>=kbset[0]->nformulas)){
										printf("Result [%d]:\n%s\n",kbset[0]->nformulas,ptr10);
									}
									#endif
									MYFREE(ptr10);
									#endif

									/* Add original simplified clause to disposal queue for later */
									/* deletion or locking. This queue interconnects clauses from other */
									/* queues. The clause must not be removed from any queue to which */
									/* it is linked. */
									binp->clause->flags|=INDISPOSALQUEUE;
									binp->clause->prevdisp=kbset[nkb]->lastdisp;
									kbset[nkb]->lastdisp=binp->clause;
									kbset[nkb]->prstats.bksubsresltns++;

									/* Add new simplified clause to KB as UNPROC. */
									newcl->flags=UNPROC;
									newcl->inference=BKSUBSRESLTNS;
									if (NOMEMORY==AddBinClause2KB(newcl,nkb,0)){
										if (newcl->part2.bin->asserts!=NULL){
											MYFREE(newcl->part2.bin->asserts);
										}
										MYFREE(newcl->part2.bin);
										MYFREE(newcl);
										if (stack!=NULL){
											MYFREE(stack);
										}
										return(NOMEMORY);
									}

									/* Check empty clause without assertions. */
									if ((newcl->part2.bin->formula[0]==UNITEND)&&(newcl->part2.bin->asserts==NULL)){
										if (stack!=NULL){
											MYFREE(stack);
										}
										if (NULL==(kbset[nkb]->frstprfnode=mallocmn(sizeof(proofnode)))){
											return(NOMEMORY);
										}
										kbset[nkb]->frstprfnode->type=ACLAUSE;
										kbset[nkb]->frstprfnode->ptr.Aclause=newcl;
										return(1);
									}
									break;
							}
						}
						break;
				}
			} while(ii);
		}
	}

	/* Free memory and return unsuccessful simplification. */
	if (stack!=NULL){
		MYFREE(stack);
	}
	return(0);
} /* BkSubsResltn */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform subsumtion resolution inference)  OCJ
 *
 *    This function tries to perform subsumption resolution simplification
 *    of a clause by another clause. The simplified clause is placed in
 *    new allocated storage and its parents are set, but no other data is
 *    included in the clause cmprefix.
 *
 *    It is the caller responsibility to link and/or unlink the clauses
 *    and set the other prefix fields.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to cmprefix structure of clause to be simplified.
 *    literal1: Pointer to literal that will be deleted in clause.
 *    binp: Pointer to binprefix structure of clause that simplifies
 *          input clause.
 *    literal2: Pointer to the term whose negation must be a generalization
 *              of literal1.
 *    pnewcl: Address of pointer to cmprefix structure of new clause to be
 *            allocated for the clause that results from simplification.
 *
 *  RETURNS:
 *
 *    0 -> The clause was not simplified.
 *    1 -> The clause was successfully simplified.
 *    NOMEMORY -> Not enough memory.
 *
 *
 *--------------------------------------------------------------*/
int32_t SubsResltn(cmprefix *clause,uint8_t *literal1,binprefix *binp,
		uint8_t *literal2,cmprefix **pnewcl){
	auto binprefix *binp1,*binp2,*binp3;                 /* Auxiliary binary prefix pointers */
	auto subst Subst;                                    /* Auxiliary substitution */
	auto int32_t ii,nn;                                  /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                     /* Auxiliary pointer */

	/* Debug. */
	ptr10=Decompile(clause->part2.bin);
	MYFREE(ptr10);
	ptr10=Decompile(binp);
	MYFREE(ptr10);
	#endif

	/* Initialize substitution. */
	Subst.buffer=NULL;
	Subst.substsize=0;

	/* Check if literal2 is a generalization of literal1. */
	switch (ii=Generalize(literal2,literal1,&Subst,ITEM2SEC|NEGATED|INITSUBST)){
		case NOMEMORY:
		case 0:
			if (Subst.buffer!=NULL){
				MYFREE(Subst.buffer);
			}
			return(ii);
			break;
	}

	/* Candidate clause to be simplified has only one literal. */
	if (clause->part2.bin->literals==1){

		/* Free substitution memory. */
		if (Subst.buffer!=NULL){
			MYFREE(Subst.buffer);
		}

		/* Formula in binp has only one literal, build empty clause and return. */
		if (binp->literals==1){
			if (NULL==((*pnewcl)=mallocmn(sizeof(cmprefix)))){
				return(NOMEMORY);
			}
			if (NULL==((*pnewcl)->part2.bin=mallocmn(binpsize+1))){
				MYFREE(*pnewcl);
				return(NOMEMORY);
			}
			(*pnewcl)->part2.bin->clause=*pnewcl;
			(*pnewcl)->part2.bin->equflg=0;
			(*pnewcl)->parent1=clause;
			(*pnewcl)->parent2=binp->clause;
			(*pnewcl)->part2.bin->formula[0]=UNITEND;
			(*pnewcl)->part2.bin->literals=0;
			(*pnewcl)->part2.bin->size=1;
			return(1);

		/* Formula in binp has more than one literal, return no simplification. */
		} else {
			return(0);
		}
	}

	/* Allocate storage for auxiliary binprefix structures */
	/* and copy input clauses. */
	nn=clause->part2.bin->size+binpsize;
	if (NULL==(binp1=mallocmn(nn))){
		if (Subst.buffer!=NULL){
			MYFREE(Subst.buffer);
		}
		return(NOMEMORY);
	}
	memcpy(binp1,clause->part2.bin,nn);
	nn=binp->size+binpsize;
	if (NULL==(binp2=mallocmn(nn))){
		if (Subst.buffer!=NULL){
			MYFREE(Subst.buffer);
		}
		MYFREE(binp1);
		return(NOMEMORY);
	}
	memcpy(binp2,binp,nn);

	/* Delete literal1 and literal2 in auxiliary binprefix structures. */
	DeleteLiteral(binp1,((uint8_t *)binp1)+(literal1-(uint8_t *)clause->part2.bin));
	DeleteLiteral(binp2,((uint8_t *)binp2)+(literal2-(uint8_t *)binp));

	/* Check if binp2 subsumes binp1. */
	switch (ii=IsSubsumedByClause2(binp1,binp2,&Subst)){
		case NOMEMORY:
		case 0:
			if (Subst.buffer!=NULL){
				MYFREE(Subst.buffer);
			}
			MYFREE(binp1);
			MYFREE(binp2);
			return(ii);
			break;
	}

	/* Free substitution memory and binp2. */
	if (Subst.buffer!=NULL){
		MYFREE(Subst.buffer);
	}
	MYFREE(binp2);

	/* Reallocate the formula buffer to the exact length needed. */
	if (NULL==(binp3=reallocmn(binp1,binpsize+binp1->size))){
		MYFREE(binp1);
		return(NOMEMORY);
	}

	/* Allocate cmprefix structure for simplified clause. */
	if (NULL==((*pnewcl)=mallocmn(sizeof(cmprefix)))){
		MYFREE(binp3);
		return(NOMEMORY);
	}
	(*pnewcl)->part2.bin=binp3;
	binp3->equflg=0;
	binp3->clause=*pnewcl;

	/* Debug. */
	#ifdef DEBUGCODE
	ptr10=Decompile((*pnewcl)->part2.bin);
	MYFREE(ptr10);
	#endif

	/* Set simplified clause cmprefix data. */
	(*pnewcl)->parent1=clause;
	(*pnewcl)->parent2=binp->clause;

	return(1);
} /* SubsResltn */
